#ifndef __si_sha_abrgdev_h__fa579d1b_b6fa_4ea9_a9ce_a183389fc416
#define __si_sha_abrgdev_h__fa579d1b_b6fa_4ea9_a9ce_a183389fc416

/*! \file si_sha/abrgdev.h
 *  \brief Automation Bridge Device base class
 *
 *  This class serves as a base class for bridges on the JavaScript
 *  side.  The base class is able to handle much of the boiler-plate
 *  code of being a device.  Derived class bridges need to provide
 *  virtual functions for the following:
 *    - sc_automationBridgeDevice_processEventF
 *    - sc_automationBridgeDevice_processRequestF
 *    - sc_automationBridgeDevice_discoverF
 *
 *  The base will start the needed threads, and when events from the
 *  network or requests from JavaScript arrive, it will forward these
 *  to the derived class for processing.
 *  
 */

#include <si_std/queuet.h>
#include <si_sha/abrgcmn.h>
#include <si_sha/autodev.h>

#include <unistd.h>

typedef struct sc_automationDeviceMgr sc_automationDeviceMgr;


/** Macro declares sc_automationPacketPtrhQueue, a queue of automation
 *  packet pointer-handles. */
sm_queueDecl(sc_automationPacketPtrh);


/** \name Derived class function signatures.
 */
/*@{*/
typedef sc_status (*sc_automationBridgeDevice_newPacketF)(void* br,
                                                sc_automationPacketPtrh* pkt);
typedef sc_status (*sc_automationBridgeDevice_processEventF)(void* br,
                                                sc_automationPacket* evt);
typedef sc_status (*sc_automationBridgeDevice_processRequestF)(void* br,
                                                sc_automationRequest* req,
                                                sc_automationResponse* res);
typedef sc_status (*sc_automationBridgeDevice_discoverF)(void* br);
/*@}*/

/** \brief Table of functions that must be supplied by derived class
 ** bridges. 
 * 
 */
typedef struct sc_automationBridgeDevice_vtable {
  sc_automationBridgeDevice_processEventF   processEvent;
  sc_automationBridgeDevice_processRequestF processRequest;
  sc_automationBridgeDevice_newPacketF      newPacket;
  sc_automationBridgeDevice_discoverF       discover;
} sc_automationBridgeDevice_vtable;

/** \brief Base automation bridge device class.
 *
 *  This serves as a base class for bridges on the JavaScript
 *  side.  The base class is able to handle much of the boiler-plate
 *  code of being a device.  Derived class bridges need to provide
 *  virtual functions for the following:
 *    - sc_automationBridgeDevice_processEventF
 *    - sc_automationBridgeDevice_processRequestF
 *    - sc_automationBridgeDevice_discoverF
 *
 *  The base will start the needed threads, and when events from the
 *  network or requests from JavaScript arrive, it will forward these
 *  to the derived class for processing.
 */
struct sc_automationBridgeDevice {
  sc_automationDevice base_;      /**< Bridges are derived from Device base. */

  /* object stuff */
  const sc_automationBridgeDevice_vtable* vtbl_; /**< Derived v-table */
  void* instance_;                /**< Derived instance address. Passed to 
                                   *   derived v-table functions. */
  
  se_bridgeState          state_; /**< Bridge state (started, stopped, etc.) */

  pthread_mutex_t         mux_;   /**< Mutex for thread safety */
  pthread_t               svcThread_; /**< Thread handling events & requests */
  
  sc_automationPacketPtrh xchg_;   /**< Exchange point for packets between JS
                                    *   engine and service thread. */
  
  sc_automationDeviceMgr* pdm_;    /**< Reference to Automation Device Manager*/
  int                     fd_;     /**< Low-level bridge driver fd */
  int                     fdCtlC_; /**< Control fd for Client side */
  int                     fdCtlS_; /**< Control fd for Server side */
};

sc_status sc_automationBridgeDevice_construct(
                               sc_automationBridgeDevice* pbd,
                               sc_automationDeviceMgr*    pdm,
                               const TCHAR* name,
                               const TCHAR* type,
                               const TCHAR* addr,
                               const TCHAR* networkType,
                               const sc_automationBridgeDevice_vtable* vtbl,
                               void* instance);
/**< Constructs automation bridge device base class.
 *
 * This should only be called by derived bridge class constructors.
 *
 * \param[in,out] pbd Base Bridge device to construct (pointer to base
 *                    member).
 * \param[in] pdm     Reference to device manager.
 * \param[in] name    Name (identifier) of the bridge device in LASE.
 * \param[in] type    Aevice type of the bridge device in LASE.
 * \param[in] addr    Address of bridge device on the network.
 * \param[in] networkType Type of network the bridge connects to.
 * \param[in] vtbl    Derived class v-table.
 * \param[in] instance Derived class address (passed to all derived
 *                     class functions in v-table).
 */

sc_status sc_automationBridgeDevice_destruct(sc_automationBridgeDevice* pbd);
/**< Destructs Bridge Device base class.
 * \param[in,out] pbd Pointer to base member to destruct.
 */

sc_status sc_automationBridgeDevice_newPacket(sc_automationBridgeDevice* pbd,
                                              sc_automationPacketPtrh* pptrh);
/**< Syntactic sugar.  Calls derived class newPacket via v-table.
 *
 * Returns a new empty packet to use for network messages.
 *
 * \param[in,out] pbd   The bridge of interest.
 * \param[in]     pptrh The pointer-handle to which the new packet will
 *                      be assigned.
 */


sc_status sc_automationBridgeDevice_discover(sc_automationBridgeDevice* pbd);
/**< Syntactic sugar.  Calls derived class discover via v-table.
 *
 * All discovered devices are passed on for creation to the JavaScript
 * context automatically.
 * 
 * \param[in,out] pbd   The bridge of interest.
 */

sc_status sc_automationBridgeDevice_sendMessage(
                                           sc_automationBridgeDevice* pbd,
                                           sc_automationPacket*       msg,
                                           sc_automationPacketPtrh*   res);
/**< Sends the message in \p msg and returns the response in \p res.
 *
 * This internally forwards \p msg to the service thread, which in turn
 * calls the derived class processRequest function.
 *
 * \param[in,out] pbd The bridge to send the message through.
 * \param[in]     msg The message to send.
 * \param[out]    res The response is assigned to the this
 *                    pointer-handle.
 */


#endif /*  __si_sha_abrgdev_h__fa579d1b_b6fa_4ea9_a9ce_a183389fc416 */
