#ifndef __si_adevmgr_h__48ec342b_57a0_4648_a018_6a8574ce7bba
#define __si_adevmgr_h__48ec342b_57a0_4648_a018_6a8574ce7bba

/** \file si_sha/adevmgr.h
 *  \brief Automation Device Manager object
 *
 *  The Automation device manager keeps track of mappings between
 *  device addresses and instance identifiers in JavaScript.  It also
 *  maps device types to JavaScript class names.
 *
 *  This important function serves several critical purposes in
 *  CASE, including:
 *   - When an event comes in from the network, the src address must be
 *     mapped to an instance in JavaScript, so that the event can be
 *     reissued in JavaScript from the appropriate object.
 *   - When a new device is discovered on an automation network, the
 *     corresponding JavaScript class name must be found, and an
 *     instance identifier must be looked up or created, in order to
 *     create a new JS object instance of the appropriate type and
 *     identifier.
 */

#include <si_std/listt.h>
#include <si_sha/autodev.h>
#include <si_sha/autopkt.h>

typedef struct sc_lase sc_lase;

/** \brief A struct that facilitates string to string mappings.
 */
typedef struct sc_nameToName {
  const TCHAR* first;  /**< First string, typically the key. */
  const TCHAR* second; /**< Second string, typically the value. */
} sc_nameToName;

/** \brief Macro declares sc_nameToNameList, a list of nameToName
 * structs.
 */
sm_listDecl(sc_nameToName);

/** \brief Automation Device Manager
 *
 *  The Automation device manager keeps track of mappings between
 *  device addresses and instance identifiers in JavaScript.  It also
 *  maps device types to JavaScript class names.
 *
 *  This important function serves several critical purposes in
 *  CASE, including:
 *   - When an event comes in from the network, the src address must be
 *     mapped to an instance in JavaScript, so that the event can be
 *     reissued in JavaScript from the appropriate object.
 *   - When a new device is discovered on an automation network, the
 *     corresponding JavaScript class name must be found, and an
 *     instance identifier must be looked up or created, in order to
 *     create a new JS object instance of the appropriate type and
 *     identifier.
 *     
 */
typedef struct sc_automationDeviceMgr {

  /* local caches */
  sc_nameToNameList instcFromDevAddr; /**< List for instance lookups */
  sc_nameToNameList inameFromDevAddr; /**< Cache for instance name lookups */
  sc_nameToNameList classFromDevType; /**< Cache for class name lookups */
  sc_lase*          jsEngine; /**< JavaScript environment */
  struct sqlite3*   db;       /**< SQLite3 Database for long-term storage */
} sc_automationDeviceMgr;
   
sc_status sc_automationDeviceMgr_construct(sc_automationDeviceMgr* pom,
                                           sc_lase* jscx);
/**< Constructs an automation device manager for the given JavaScript
 *   environment.
 *
 * \param[in,out] pom   The object manager to construct.
 * \param[in]     jscx  The JavaScript environment to manage.
 */

sc_status sc_automationDeviceMgr_destruct(sc_automationDeviceMgr* pom);
/**<  Destructs the given device manager.
 *
 * \param[in,out] pom   The object manager to destruct.
 */

sc_status sc_automationDeviceMgr_getInstanceNameFromDeviceAddress(
                                      sc_automationDeviceMgr* pom,
                                      const TCHAR* deviceAddr,
                                      const TCHAR** instanceName);
/**< Returns the JavaScript object identifier (name) for the given device
 *   address.
 *
 * \note The string returned in \p instanceName should not be freed
 *       as it is stored in the cache.
 *
 * \param[in,out] pom           The object manager to use.
 * \param[in]     deviceAddr    The device address to find a name for.
 * \param[out]    instanceName  The returned instance identifier, or NULL.
 * 
 */

sc_status sc_automationDeviceMgr_getClassFromDeviceType(
                                      sc_automationDeviceMgr* pom,
                                      const TCHAR* deviceType,
                                      const TCHAR** className);
/**< Returns the JavaScript class name for the given device type.
 *
 * \note The string returned in \p className should not be freed
 *       as it is stored in the cache.
 *
 * \param[in,out] pom         The object manager to use.
 * \param[in]     deviceType  The device address to find a name for.
 * \param[out]    className   The returned class name, or NULL.
 * 
 */

sc_status sc_automationDeviceMgr_getInstanceFromDeviceAddress(
                                      sc_automationDeviceMgr* pom,
                                      const TCHAR* deviceAddr,
                                      const TCHAR** instanceName);
/**< Returns the JavaScript object identifier (name) for the device with
 *   the given address, if an object already exists for it, or NULL otherwise.
 *
 * The difference between this and getInstanceNameFromDeviceAddress is
 * that this function will return NULL for a named device that does not
 * have an object instance already created for it in the current
 * JavaScript context, whereas the other will always return the name for
 * the device if was ever given one.
 *
 * \note The string returned in \p instanceName should not be freed
 *       as it is stored in the cache.
 *
 * \param[in,out] pom           The object manager to use.
 * \param[in]     deviceAddr    The device address to find a name for.
 * \param[out]    instanceName  The returned instance identifier, or NULL.
 * 
 */

sc_status sc_automationDeviceMgr_createDevice(sc_automationDeviceMgr* pom,
                                              const TCHAR* bridgeName,
                                              sc_automationDevice* pd);
/**<  Creates a JavaScript object for device \p pd which is attached to
 * bridge \p bridgeName.
 *
 * \note A script request will be submitted to create the device in the
 * JavaScript context.
 *
 * \param[in,out] pom           The device manager to use.
 * \param[in]     bridgeName    The JS identifier of the bridge the
 *                              device is connected to.
 * \param[in]     pd            The device to create an instance for.
 * 
 */

sc_status sc_automationDeviceMgr_relayMessage(sc_automationDeviceMgr* pom,
                                              const TCHAR* deviceAddress,
                                              sc_automationPacket* pkt);
/**< Relays the message in \p pkt to the device object in JavaScript
 * whose address matches \p deviceAddress.
 *
 * If the device with the given address does not yet have an associated
 * JavaScript object instance, the device manager attempts to create one. If
 * the device has not been seen yet or the type is unknown, the device
 * manager will not be able to create an instance and the event will be
 * dropped.
 *
 * \param[in,out] pom           The device manager to use.
 * \param[in]     deviceAddress The device to relay the message to.
 * \param[in]     pkt           The message to relay.
 */

sc_status sc_automationDeviceMgr_saveArchive(sc_automationDeviceMgr* pom,
                                             const TCHAR* ident,
                                             const TCHAR* archive);
/**< Stores the given archive for the given device in the device
 * database.
 *
 * This information is restored to this device the next time it is
 * discovered and a new JavaScript object instance is created for it
 * (for example, if LASE is restarted).
 *
 * \param[in,out] pom      The device manager to use. 
 * \param[in]     ident    The identifier of the device for which the 
 *                         archive is saved.
 * \param[in]     archive  The JSON string to store as an archive.
 *
 * \note The archive can be properly restored to correct device even if
 * the device identifier subsequently changes via 
 * sc_automationDeviceMgr_renameDevice.
 * 
 */

sc_status sc_automationDeviceMgr_saveLink(sc_automationDeviceMgr* pom,
                                          const TCHAR* srcIdent,
                                          const TCHAR* srcEvent,
                                          const TCHAR* dstIdent,
                                          const TCHAR* dstMethd);
/**< An event link is stored between devices \p srcIdent and \p
 * dstIdent.
 *
 * The link will be restored the next time the devices are
 * re-created.
 * 
 * \param[in,out] pom      The device manager to use. 
 * \param[in]     srcIdent The identifier of the device sourcing the
 *                         event.
 * \param[in]     srcEvent The name of the event linking src and dst.
 * \param[in]     dstIdent The identifier of the device that subscribed
 *                         to the event.
 * \param[in]     dstMethd The name of the method to be called on the
 *                         dst device when the event occurs.
 *
 * \note The link can be properly restored between the correct devices 
 * even if one or both identifiers subsequently changes via 
 * sc_automationDeviceMgr_renameDevice.
 * 
 */

sc_status sc_automationDeviceMgr_clearLinks(sc_automationDeviceMgr* pom,
                                            const TCHAR* srcIdent);
/**< Clears any archived event links for which \p srcIdent was the event
 * source.
 *
 * This is required before \p srcIdent is saved, so that any previously stored
 * links are forgotten.
 *
 * \param[in,out] pom      The device manager to use.
 * \param[in]     srcIdent The device to clear event links for.
 * 
 */

sc_status sc_automationDeviceMgr_renameDevice(sc_automationDeviceMgr* pom,
                                              const TCHAR* curName,
                                              const TCHAR* newName);
/**< Renames the device with identifier \p curName to \p newName.
 *
 * This renaming takes effect immediately, and is permanent.  The next
 * time the device is newly discovered (for example, on LASE restart)
 * it will retain this new name.
 *
 * \param[in,out] pom      The device manager to use.
 * \param[in]     curName  The current name of the device.
 * \param[in]     newName  The new name of the device.
 * 
 */

#endif /* __si_adevmgr_h__48ec342b_57a0_4648_a018_6a8574ce7bba */
