#ifndef __si_sha_autobrg_h__83e6f8a1_6266_40d6_931e_b98bbf61c3d7
#define __si_sha_autobrg_h__83e6f8a1_6266_40d6_931e_b98bbf61c3d7

/** \file si_sha/autobrg.h
 *  \brief Automation bridge controller base class.
 *
 * The automation bridge base class makes it easier to be a bridge
 * controller in LASE by implementing many of the common boiler plate
 * things nearly all bridges have to do, such as spin a thread to
 * listen for activity on the network, as well as from the script side.
 * 
 */
#include <si_sha/abrgcmn.h>
#include <unistd.h>
#include <pthread.h>


typedef struct sc_automationBridgeBase sc_automationBridgeBase;


/** \brief Represents a connection to an automation bridge controller.
 */
typedef struct sc_automationBridgeConnection
{
  sc_rep _rep;           /**< Connection may be pointer-handled */
  
  int fd_;               /**< Connection fd. */
  int dieW_;             /**< die fd. */
  int dieR_;             /**< die fd (thread side). */
  pthread_t thread_;     /**< Connection thread. */
  s_bool open_;          /**< True if the connection is still open. */
    
  sc_automationBridgeBase* brg_; /**< Bridge we are connected to. */
} sc_automationBridgeConnection;

/** \brief Macro declares sc_automationBridgeConnectionPtrh, a
 *  pointer-handle to a bridge connection.
 */
sm_ptrhDecl(sc_automationBridgeConnection);

/** \brief Macro declares sc_automationBridgeConnectionPtrhList, a
 *  list of pointer-handles to a bridge connection.
 */
sm_listDecl(sc_automationBridgeConnectionPtrh);

sc_status sc_automationBridgeConnection_new(sc_automationBridgeBase* pb, int fd,
                                       sc_automationBridgeConnectionPtrh* conn);
/**< Create a new connection object with the given socket.
 *
 * \param[in,out] pb    Bridge base device.
 * \param[in]     fd    Socket that the connection is on.
 * \param[out]    conn  The pointer-handle the new connection is
 *                      assigned to.
 */

sc_status sc_automationBridgeConnection_open(sc_automationBridgeConnection* pc);
/**< Creates a new thread to handle the connection.
 *
 * \param[in,out] pc The connection to handle.
 */

sc_status sc_automationBridgeConnection_close(
                                           sc_automationBridgeConnection* pc);
/**< Signals the connection thread to close the connection and exit.
 *
 * This waits for the connection thread to exit.
 *
 * \param[in,out] pc The connection to close.
 */


/** Function signature for derived class serviceDevice function */
typedef sc_status (*sc_automationBridge_serviceDeviceF) (void* br,
                                                      sc_automationPacket* pkt);

/** Function signature for derived class processRequest function */
typedef sc_status (*sc_automationBridge_processRequestF)(void* br,
                                              sc_automationBridgeConnection* pc,
                                              sc_automationRequest* req,
                                              sc_automationResponse* res);

/** \brief V-table for derived class functions. */
typedef struct sc_automationBridge_vtable {
  sc_automationBridge_serviceDeviceF    serviceDevice;
  sc_automationBridge_processRequestF   processRequest;
} sc_automationBridge_vtable;

/** \brief Automation Bridge Base class
 *
 * The automation bridge base class makes it easier to be a bridge
 * controller in LASE by implementing many of the common boiler plate
 * things nearly all bridges have to do, such as spin a thread to
 * listen for activity on the network, as well as from the script side.
 */
struct sc_automationBridgeBase {

  const TCHAR* name_;          /**< Name of the exposed device */
  const TCHAR* basePath_;      /**< Base path of system. */
  
  const sc_automationBridge_vtable* vtbl_;  /**< Derived class v-table */
  void* instance_;             /**< Derived object instance address. */

  se_bridgeState   state_;     /**< State of bridge controller (started, 
                                    stopped, etc)*/
  pthread_mutex_t  mux_;       /**< Mutex to protect connection list and 
                                    device traffic (only one connection at
                                    a time can talk to the bridge). */
  pthread_t        svcThread_; /**< Thread to service requests incoming 
                                    from network and script side */
      
  int fdSvc_;  /**< Service file descriptor (for connections)  */
  int fdDev_;  /**< Device file descriptor (e.g., serial port) */
  int fdCtlW_; /**< Control write file descriptor */
  int fdCtlR_; /**< Control read file descriptor (thread side) */

  sc_automationPacketPool pool_;     /**< packet pool for network originated 
                                          traffic */
  sc_automationPacketPtrhList ooo_;  /**< out-of-order packets received from 
                                          the network. */
  sc_automationBridgeConnectionPtrhList connections_; /**< active connections*/
};


sc_status sc_automationBridgeBase_construct(
                                sc_automationBridgeBase*          pb,
                                const TCHAR* name,
                                const TCHAR* basePath,
                                s_uint32 pktSize, s_uint32 pktPoolSize,
                                const sc_automationBridge_vtable* vtbl,
                                void* instance);
/**<  Constructs an automation bridge controller base class.
 *
 * \param[in,out] pb           The bridge base member to construct.
 * \param[in]     name         The exposed name of the bridge
 * \param[in]     basePath     The base path of the system.
 * \param[in]     pktSize      The max size of packets on the network.
 * \param[in]     pktPoolSize  The initial size of packet pool.
 * \param[in]     vtbl         The derived class v-table.
 * \param[in]     instance     The address of the derived class instance.
 *
 * \note When calling derived class functions via the v-table, the
 *       derived instance address is given as 'this' (first parameter).
 *       
 */

sc_status sc_automationBridgeBase_start(sc_automationBridgeBase* pb);
/**< Starts the bridge controller service thread.
 * 
 * \param[in,out] pb           The bridge base instance.
 */

sc_status sc_automationBridgeBase_stop (sc_automationBridgeBase* pb);
/**< Stops the bridge controller service thread.
 * 
 * \param[in,out] pb           The bridge base instance.
 */

sc_status sc_automationBridgeBase_destruct(sc_automationBridgeBase* pb);
/**< Destructs the bridge controller base.
 * \param[in,out] pb           The bridge base member address.
 */


#endif /* __si_sha_autobrg_h__83e6f8a1_6266_40d6_931e_b98bbf61c3d7 */
