#ifndef __si_sha_autopkt_h__00911341_b086_4066_8b09_3d6bf85f104f
#define __si_sha_autopkt_h__00911341_b086_4066_8b09_3d6bf85f104f

/** \file si_sha/autopkt.h
 *  \brief General Packet data container.
 *
 *  This packet object is a general purpose data container for
 *  the various types of automation networks.  It functions like a
 *  stack (LIFO), and access to the object and the data it contains is 
 *  available from both JavaScript and C.  
 */

#include <si_sys/defs.h>
#include <si_std/ptrh.h>
#include <si_std/listt.h>

#include <pthread.h>

/** \brief Packet data container object.
 *
 *  This packet object is a general purpose data container for
 *  the various types of automation networks.  It functions like a
 *  stack (LIFO), and access to the object and the data it contains is 
 *  available from both JavaScript and C.  
 */
typedef struct sc_automationPacket {
  sc_rep _rep;            /**< Packets are typically pointer-handled. */
  struct {
    s_uint8* buffer_;     /**< Byte storage for packet data. */
    s_uint8* top_;        /**< Current position of the 'top' of the stack. */
    s_int32  capacity_;   /**< Total capacity of the packet object. */
    s_int32  size_;       /**< Amount of data currently in packet. */
    s_bool   own_;        /**< Should packet release buffer on destruct. */
  } priv_;                /**< All members in priv_ are considered private.*/
} sc_automationPacket;

sc_status sc_automationPacket_construct(sc_automationPacket* ppkt,
                                        s_int32 capacity,
                                        s_uint8* pbuffer);
/**<  Constructs a new packet with the given capacity and buffer.
 *
 * If \p pbuffer is NULL then a new buffer is allocated with the given
 * capacity in the constructor, and the own_ member is set to true.
 * Otherwise, \p pbuffer is used directly, assumed to have \p capacity,
 * and own_ is set to false.  On destruct, if own_ is true, then the
 * buffer_ member is freed.
 *
 * \param[in,out] ppkt     The packet object to construct.
 * \param[in]     capacity The capacity the packet will have.
 * \param[in]     pbuffer  If not NULL, this buffer is used directly.
 *                         Otherwise, a buffer of size \p capacity is
 *                         allocated in the constructor.
 *                         
 */

sc_status sc_automationPacket_serializeToFd(sc_automationPacket* pkt, int fd);
/**< Serializes the given packet to the file descriptor \p fd.
 * 
 * \param[in,out] pkt  The packet of interest.
 * \param[in]     fd   The file descriptor to serialize to.
 * \see sc_automationPacket_deserializeFromFd for pulling a packet out
 *      of a file descriptor.
 */
sc_status sc_automationPacket_deserializeFromFd(sc_automationPacket* pkt, int fd);
/**< De-serializes the given packet from the file descriptor \p fd.
 * 
 * \param[in,out] pkt  The packet of interest.
 * \param[in]     fd   The file descriptor to serialize from.
 * \see sc_automationPacket_serializeToFd for pushing a packet into a
 *      file descriptor.
 */

sc_status sc_automationPacket_tell(sc_automationPacket* pkt, s_uint32* ppos);
/**< Returns the position of the packet in \p ppos.
 *
 * This position can be returned to by using the setPos method.
 *
 * \param[in,out] pkt  The packet of interest.
 * \param[out]    ppos The current position of the packet.
 */

sc_status sc_automationPacket_setPos(sc_automationPacket* pkt, s_uint32 pos);
/**< Sets the given packet to a previously found position.
 *
 * Note that many positions would not result in a usable packet.  To
 * get a useable position, call sc_automationPacket_tell when the
 * packet is in a good state.
 *
 * \param[in,out] pkt  The packet of interest.
 * \param[in]     pos  The position to move to.
 */


sc_status sc_automationPacket_clear(sc_automationPacket* pkt);
/**< Empties the given packet. Resets top and sets size to 0.
 * 
 * \param[in,out] pkt  The packet to clear.
 */

/** \name Functions for moving data into and out of packets.
 *
 *  These functions push/pop byte or integer data into/out of the
 *  packet.
 */
/*@{*/
/* push/pop U/Int8 */
sc_status sc_automationPacket_pushUInt8(sc_automationPacket* pkt, s_uint8 val);
sc_status sc_automationPacket_popUInt8(sc_automationPacket* pkt, s_uint8* val);
sc_status sc_automationPacket_pushInt8(sc_automationPacket* pkt, s_int8 val);
sc_status sc_automationPacket_popInt8(sc_automationPacket* pkt, s_int8* val);
/* push/pop U/Int16 */
sc_status sc_automationPacket_pushUInt16(sc_automationPacket* pkt, s_uint16 v);
sc_status sc_automationPacket_popUInt16(sc_automationPacket* pkt, s_uint16* v);
sc_status sc_automationPacket_pushInt16(sc_automationPacket* pkt, s_int16 v);
sc_status sc_automationPacket_popInt16(sc_automationPacket* pkt, s_int16* v);
/* push/pop U/Int32 */
sc_status sc_automationPacket_pushUInt32(sc_automationPacket* pkt, s_uint32 v);
sc_status sc_automationPacket_popUInt32(sc_automationPacket* pkt, s_uint32* v);
sc_status sc_automationPacket_pushInt32(sc_automationPacket* pkt, s_int32 v);
sc_status sc_automationPacket_popInt32(sc_automationPacket* pkt, s_int32* v);
/* push/pop U/Int64 */
sc_status sc_automationPacket_pushUInt64(sc_automationPacket* pkt, s_uint64 v);
sc_status sc_automationPacket_popUInt64(sc_automationPacket* pkt, s_uint64* v);
sc_status sc_automationPacket_pushInt64(sc_automationPacket* pkt, s_int64 v);
sc_status sc_automationPacket_popInt64(sc_automationPacket* pkt, s_int64* v);

/* push/pop binary */
sc_status sc_automationPacket_pushBinary(sc_automationPacket* ppkt,
                                         const s_uint8* pdata,
                                         s_int32 nbytes);
sc_status sc_automationPacket_popBinary(sc_automationPacket* ppkt,
                                        s_uint8* pdata,
                                        s_int32 nbytes);
/*@}*/

/** \brief Declares sc_automationPacketPtrh, a packet pointer-handle.
 */
sm_ptrhDecl(sc_automationPacket);\
   
/** \brief Declares sc_automationPacketPtrhList, a list of packet
 *         pointer-handles.
 */
sm_listDecl(sc_automationPacketPtrh);

/** \brief A pool of packets for constant time allocation and release.
 * 
 */
typedef struct sc_automationPacketPool {
  pthread_mutex_t             mux_;          /**< Mutex for pool thread safety.*/
  sc_automationPacketPtrhList pool_;         /**< List of free packets. */
  s_int32                     packetSize_;   /**< Size of packets to pool. */
  s_int32                     currCapacity_; /**< Number of contained packets */
  s_int32                     maxCapacity_;  /**< Max number of packets to 
                                              *   allow via growth. */
  sc_blockAllocator           al_;           /**< Block allocator for growable 
                                              * constant time block
                                              * allocation/release. */
} sc_automationPacketPool;

sc_status sc_automationPacketPool_construct(sc_automationPacketPool* ppool,
                                            s_int32 pktSize,
                                            s_int32 initialCapacity,
                                            s_int32 maxCapacity);
/**< Constructs a packet pool with the given \p initialCapacity of
 *   packets (each with \p pktSize bytes) growable up to \p maxCapacity.
 *
 *   Packet pools are typically used by bridges to minimize packet
 *   allocation latency.  With a pool, the allocation is done in
 *   amortized constant time.  When all references to an allocated packet
 *   have been released, the packet is automatically returned to the
 *   free list in the pool it came from.
 *
 * \param[in,out] ppool       The packet pool to construct.
 * \param[in]     pktSize     The size of packets in the pool.
 * \param[in]     initialCapacity The number of packets to ready on
 *                            construction.
 * \param[in]     maxCapacity The maximum number of packets to allow
 *                            the pool to grow to.  Can be used to
 *                            simulate low memory conditions.
 */

sc_status sc_automationPacketPool_destruct(sc_automationPacketPool* ppool);
/**< Destructs the given packet pool.
 * 
 * \param[in,out] ppool The packet pool to destruct.
 * \note All packets must be returned to the pool before destruction.
 */

sc_status sc_automationPacketPool_allocate(sc_automationPacketPool* ppool,
                                           sc_automationPacketPtrh* pptrh);
/**<  Allocates a packet from the packet pool.
 * 
 * \param[in,out] ppool The packet pool to allocate from.
 * \param[out]    pptrh The pointer-handle to assign the
 *                      allocated packet to.
 */

sc_status sc_automationPacketPool_free(sc_automationPacketPool* ppool,
                                       sc_automationPacketPtrh* pptrh);
/**< Releases a packet back to the packet pool.
 * 
 * \param[in,out] ppool The pool to return the packet to.
 * \param[in]    pptrh The pointer-handle to the packet being returned.
 */

#endif /* __si_sha_autopkt_h__00911341_b086_4066_8b09_3d6bf85f104f */
