#ifndef __si_sha_instbdev_h__307cbb05_231b_4e1a_b64e_6074532a9bc8
#define __si_sha_instbdev_h__307cbb05_231b_4e1a_b64e_6074532a9bc8

/** \file si_sha/instbdev.h
 *  \brief Insteon Bridge device
 *
 *  The Insteon bridge device derives from the base automation
 *  device, and implements a number of Insteon specific functions for
 *  use by the JavaScript device.
 *  
 */

#include <si_sha/abrgdev.h>


/** \brief The Insteon bridge device class.
 *
 *  The Insteon bridge device derives from the base automation
 *  device, and implements a number of Insteon specific functions for
 *  use by the JavaScript device.
 */
typedef struct sc_insteonBridgeDevice {  
  sc_automationBridgeDevice base_;  /**< base class */
  sc_automationPacketPool   pool_;  /**< Insteon packet pool */
} sc_insteonBridgeDevice;

sc_status sc_insteonBridgeDevice_construct(sc_insteonBridgeDevice* brg,
                                           sc_automationDeviceMgr* pdm,
                                           const TCHAR* name,
                                           const TCHAR* basePath);
/**< Constructs the Insteon bridge device \p brg with the given \p name.
 *
 * \param[in,out] brg       The insteon bridge device to construct.
 * \param[in]     pdm       The device manager for name resolution.
 * \param[in]     name      The name of the insteon bridge.
 * \param[in]     basePath  The basePath of the system.
 */

sc_status sc_insteonBridgeDevice_destruct(sc_insteonBridgeDevice* brg);
/**< Destructs the Insteon bridge \p brg.
 */

sc_status sc_insteonBridgeDevice_newPacket(sc_insteonBridgeDevice* pb,
                                           sc_automationPacketPtrh* pptrh);
/**< Returns a new packet object from the Insteon device pool.
 *
 * \param[in,out]  brg   The bridge device to use.
 * \param[out]    pptrh  The pointer-handle where the newly allolcated
 *                       packet will be assigned.
 *                       
 * \note The packet will be automatically returned to this pool when
 *       all references to it are released.
 */

sc_status sc_insteonBridgeDevice_sendStdMessage(sc_insteonBridgeDevice* pb,
                                                const TCHAR* address,
                                                s_uint8 cmd1, s_uint8 cmd2);
/**< Sends an Insteon standard message to the bridge controller.
 *
 * \param[in,out]  brg      The bridge device to use.
 * \param[in]      address  The address of the device to send the
 *                          message to.
 * \param[in]      cmd1     The first byte of the standard message.
 * \param[in]      cmd2     The second byte of the standard message.
 *
 * \note The cmd1 and cmd2 arguments for standard messages are defined
 *       by Insteon in the document 'Insteon Command Tables'.
 * \see http://www.insteon.net/pdf/INSTEON_Command_Tables_20070925a.pdf
 *      for details about possible values of cmd1 and cmd2 and their
 *      semantics.
 */

sc_status sc_insteonBridgeDevice_sendStdWaitAck(sc_insteonBridgeDevice* pb,
                                                const TCHAR* address,
                                                s_uint8 cmd1, s_uint8 cmd2,
                                                sc_automationPacketPtrh* rptrh);
/**< Sends a standard message for which a data response message is
 *   expected.
 *
 * Some messages are a query for data.  The response comes back as a
 * full message.
 *
 * \param[in,out]  brg      The bridge device to use.
 * \param[in]      address  The address of the device to send the
 *                          message to.
 * \param[in]      cmd1     The first byte of the standard message.
 * \param[in]      cmd2     The second byte of the standard message.
 * \param[out]     rptrh    The response message.
 *
 * \see sc_insteonBridgeDevice_sendStdMessage.
 */

sc_status sc_insteonBridgeDevice_processEvent(sc_insteonBridgeDevice* brg,
                                              sc_automationPacket* pkt);
/**< Processes the network event in \p pkt, relaying it to the appropriate
 *   JavaScript object via the device manager.
 *
 * \param[in,out]  brg  The bridge device to use.
 * \param[in]      pkt  The event packet to relay.
 *
 * \note Derived class override.
 */

sc_status sc_insteonBridgeDevice_processRequest(sc_insteonBridgeDevice* brg,
                                                sc_automationRequest* req,
                                                sc_automationResponse* res);
/**< Processes the given request \p req and returns a response in \p res.
 *
 * Processes the request by serializing it to the insteon bridge
 * controller/driver.  
 *
 * \param[in,out]  brg  The bridge device to use.
 * \param[in]      req  The request to send.
 * \param[out]     res  The response to the request.
 *
 * \note Derived class override.
 * 
 */

sc_status sc_insteonBridgeDevice_resetDeviceDB(sc_insteonBridgeDevice* pb);
/**< Resets (clears) the internal device database of the Insteon bridge
 *   via its controller.
 *
 * \note After calling this, the Insteon network will have to be
 *       manually re-linked.
 */

#endif /* __si_sha_instbdev_h__307cbb05_231b_4e1a_b64e_6074532a9bc8 */
