#ifndef __si_sha_instbrg_h__51d162f8_d78f_4f57_848a_b2dd96c0f02b
#define __si_sha_instbrg_h__51d162f8_d78f_4f57_848a_b2dd96c0f02b


/** \file   si_sha/instbrg.h
 *  \brief  Low-level Insteon bridge driver.
 */

#include <si_sys/serial.h>
#include <si_sha/autobrg.h>
#include <si_sha/insteon.h>

/** \brief Insteon Bridge driver
 *
 *  This object is the low level bridge driver for the Insteon
 *  device network.  The C member functions of this object encode the
 *  binary serial messages to the controller, and follow the bridge <->
 *  host protocol.  This object currently works with the Insteon USB
 *  PowerLinc Modem (Dual Band) #2413U.  This device is reported on the
 *  network as Insteon device type 03-15.
 */
typedef struct sc_insteonBridge {
  sc_automationBridgeBase base_;        /**< Automation bridge base class. */
  sc_serialPort           port_;        /**< The serial port the bridge is connected to. */
  s_uint32                address_;     /**< The bridge network address. */
  s_uint8                 category_;    /**< The bridge device type. (MSB) */
  s_uint8                 subcategory_; /**< The bridge device type. (LSB) */
  sc_automationPacketPtrhList ooo_;     /**< List of out-of-order packets from network. */
} sc_insteonBridge;



sc_status sc_insteonBridge_construct(sc_insteonBridge* pb,
                                     const TCHAR* name,
                                     const TCHAR* devName,
                                     const TCHAR* basePath);
/**< Constructs an insteon bridge controller on the serial port with
 * the special filename \p devName.
 *
 * \param[in,out] pb    The bridge object to construct.
 * \param[in]    name   The name of the bridge.  Used to name the Unix
 *                      socket its interface is provided on.
 * \param[in]  devName  The name of the 'special' file for the serial
 *                      port the bridge is attached to.
 * \param[in] basePath  The base path for the application.
 */

sc_status sc_insteonBridge_destruct(sc_insteonBridge* pb);
/**< Destructs the given insteon bridge.  */

sc_status sc_insteonBridge_getVersion(sc_insteonBridge* pb,
                                      sc_insteonBridgeVersion* pv);
/**< Gets the version of the PLM bridge.
 * \param[in,out] pb The bridge to query.
 * \param[out]    pv The version struct to fill in.
 */

sc_status sc_insteonBridge_resetLinks(sc_insteonBridge* pb);
/**< Resets (clears) the link database of the bridge.
 *
 * This erases all linked devices from the bridge EERAM.  After this
 * call, the Insteon network must be rebuilt by manually re-linking
 * devices to the bridge.
 *
 * \param[in,out] pb The bridge to reset.
 * 
 */

sc_status sc_insteonBridge_getFirstLinkEntry(sc_insteonBridge* pb,
                                             sc_insteonLinkEntry* ple);
/**< Gets the first entry in the PLM link database.
 * 
 * \param[in,out] pb  The bridge to query.
 * \param[out]    ple The link entry struct to fill in.
 * \note If the returned link entry has isEOL set to true, then the
 *       record should be treated as an end-of-list marker, and the
 *       record data should be ignored.
 */

sc_status sc_insteonBridge_getNextLink(sc_insteonBridge* pb,
                                       sc_insteonLinkEntry* ple);
/**< Gets the next entry in the PLM link database.
 * 
 * \param[in,out] pb  The bridge to query.
 * \param[out]    ple The link entry struct to fill in.
 * \note If the returned link entry has isEOL set to true, then the
 *       record should be treated as an end-of-list marker, and the
 *       record data should be ignored.
 */

sc_status sc_insteonBridge_getConfiguration(sc_insteonBridge* pb, s_uint8* pcfg);
/**< Gets the PLM bridge configuration byte.
 *
 * \param[in,out] pb    The bridge to query.
 * \param[out]    pcfg  The configuration byte.
 *
 * \see se_insteonConfigurationFlags for a description of the meanings
 *      of the different bits in the configuration byte.
 */

sc_status sc_insteonBridge_setConfiguration(sc_insteonBridge* pb, s_uint8 cfg);
/**< Sets the bridge configuration byte.
 * 
 * \param[in,out] pb   The bridge to query.
 * \param[in]     cfg  The configuration byte.
 * 
 * \see se_insteonConfigurationFlags for a description of the meanings
 *      of the different bits in the configuration byte.
 */

sc_status sc_insteonBridge_sendGroupMessage(sc_insteonBridge* pb,
                                            s_uint8 groupId,
                                            s_uint8 cmd1, s_uint8 cmd2);
/**< Send a message to the group identified by \p groupId.
 *
 * A group-cast message with the given cmd1 and cmd2 bytes is sent out.
 *
 * \param[in,out] pb       The bridge to send the message through.
 * \param[in]     groupId  The group to send the message to.
 * \param[in]     cmd1     The first command byte of the message.
 * \param[in]     cmd2     The second command byte of the message.
 *
 * \see http://www.insteon.net/pdf/INSTEON_Command_Tables_20070925a.pdf
 *      for details about possible values of cmd1 and cmd2 and their
 *      semantics.
 */

sc_status sc_insteonBridge_sendStdMessage(sc_insteonBridge* pb,
                                          s_uint32 toAddress,
                                          s_uint8 cmd1, s_uint8 cmd2);
/**< Sends a 'standard' Insteon message to \p toAddress on the network.
 *
 * Nearly all the most common messages on the Insteon network are
 * standard messages.  For example, an 'On' message is typically
 * cmd1=0x11 and cmd2=0-0xff (on-level). 'Off' is typically cmd1=0x13 
 * and cmd2=0x01. All standard and extended commands are documented by
 * Insteon in the document 'Insteon Command Tables'.
 * 
 * \param[in,out] pb         The bridge to send the message through.
 * \param[in]     toAddress  The address of the device to send the
 *                           message to.
 * \param[in]     cmd1       The first command byte of the message.
 * \param[in]     cmd2       The second command byte of the message.
 * 
 * \see http://www.insteon.net/pdf/INSTEON_Command_Tables_20070925a.pdf
 *      for details about possible values of cmd1 and cmd2 and their
 *      semantics.
 */


sc_status sc_insteonBridge_sendExtMessage(sc_insteonBridge* pb,
                                          s_uint32 toAddress,
                                          s_uint8 cmd1, s_uint8 cmd2,
                                          s_uint8* data, s_int32 len);
/**< Sends an 'extended' Insteon message to \p toAddress on the network.
 * 
 * All standard and extended commands are documented by Insteon in the document
 * 'Insteon Command Tables'.
 * 
 * \param[in,out] pb         The bridge to send the message through.
 * \param[in]     toAddress  The address of the device to send the
 *                           message to.
 * \param[in]     cmd1       The first command byte of the message.
 * \param[in]     cmd2       The second command byte of the message.
 * \param[in]     data       The 14 bytes of extended payload data.
 * \param[in]     len        The length of the given data.  The
 *                           remaining bytes will be zero padded.
 *                           
 * \see http://www.insteon.net/pdf/INSTEON_Command_Tables_20070925a.pdf
 *      for details about possible values of cmd1 and cmd2 and their
 *      semantics.
 */

#endif /* __si_sha_instbrg_h__51d162f8_d78f_4f57_848a_b2dd96c0f02b */
