#ifndef __si_sha_lase_h__2bf6de29_af3b_4a35_803b_7262e61900f3
#define __si_sha_lase_h__2bf6de29_af3b_4a35_803b_7262e61900f3

/** \file si_sha/lase.h
 *  \brief LASE class
 *
 * The LASE class is the center of the automation scripting
 * environment.  It holds the JavaScript runtime, context, and global
 * objects, the device manager for name and type mapping and maintains
 * the thread that runs the JavaScript engine, and the script queue
 * that feeds it.
 *
 * The LASE object also initializes the system by reading the 'bridges'
 * table in the database, and creating instances of any bridges it
 * finds.  It also installs a number of helper function into the global
 * JavaScript namespace:
 *   - print(var): Converts var to string and prints it to the console.
 *   - save(): Save the current state of all devices in the context.
 *             This includes any user-defined properties and event links.
 *   - rename(device, newName): Renames device 'newName'
 */

#include <si_std/queuet.h>
#include <si_sha/sreqpool.h>
#include <si_sha/adevmgr.h>
#include <pthread.h>
#include <jsapi.h>

/** \brief Enum describing start/stop state of LASE.
 */
typedef enum se_laseState {
  se_laseStateStopping,
  se_laseStateStopped,
  se_laseStateStarting,  
  se_laseStateStarted
} se_laseState;

/** Macro declares sc_scriptRequestPtrhQueue, a queue of script request
 * pointer-handles.
 */
sm_queueDecl(sc_scriptRequestPtrh);

/** \brief The LASE class.
 * 
 * The LASE class is the center of the automation scripting
 * environment.  It holds the JavaScript runtime, context, and global
 * objects, the device manager for name and type mapping and maintains
 * the thread that runs the JavaScript engine, and the script queue
 * that feeds it.
 *
 * The LASE object also initializes the system by reading the 'bridges'
 * table in the database, and creating instances of any bridges it
 * finds.  It also installs a number of helper function into the global
 * JavaScript namespace:
 *   - print(var): Converts var to string and prints it to the console.
 *   - save(): Save the current state of all devices in the context.
 *            This includes any user-defined properties and event links.
 *   - rename(device, newName): Renames device 'newName'
 */
struct sc_lase {
  sc_automationDeviceMgr      dmgr;     /**< Device manager */
  sc_string                   basePath; /**< Base path of lase data */
  
  JSRuntime*                  jsrt;     /**< JavaScript runtime */
  JSContext*                  jscx;     /**< JavaScript context */
  JSObject *                  glob;     /**< JavaScript global object */

  sc_scriptRequestPool        pool;     /**< Pool of script requests for 
                                             constant time allocation/release */
  sc_scriptRequestPtrhQueue   reqs;     /**< Queue (FIFO) of outstanding script
                                             requests */
  pthread_t                   svcThread;/**< Service thread (executes all 
                                             script requests). */
  int                         ctlFdC;   /**< Control socket. */
  int                         ctlFdS;   /**< Control socket (svcThread side). */
  se_laseState                state;    /**< Start/stop state of LASE. */
  pthread_mutex_t             mux;      /**< Protects request queue, etc. */  
};

sc_status sc_lase_construct(sc_lase* pl, s_uint32 envSize,
                            const TCHAR* basePath);
/**< Construct the LASE system with the given JavaScript environment
 *   size, with LASE data based at \p basePath.
 *
 * \param[in,out] pl       The LASE object to construct.
 * \param[in]     envSize  The JavaScript environment size.
 * \param[in]     basePath The base path of the lase data directory
 *                         structure.
 * \note In a standard installed system, basePath will be
 * /var/lib/lase. On a development system, it could be the src/lase 
 * directory of your source tree.
 *
 * \note To avoid excessive garbage collection, try setting \p envSize
 * to 4MB or more.
 */

sc_status sc_lase_destruct(sc_lase* pl);
/**< Destruct the LASE object \p pl.
 */

sc_status sc_lase_start(sc_lase* pl);
/**< Start the lase system \p pl. The system must be started before it will
 *   process any script requests.
 */
sc_status sc_lase_stop(sc_lase* pl);
/**< Stop the lase system \p pl. Any outstanding script requests will
 *   be ignored.
 */

sc_status sc_lase_allocRequest(sc_lase* pl, sc_scriptRequestPtrh* ptrh);
/**< Allocate a script request object for later submission. 
 *
 * \param[in,out] pl    The LASE system to allocate from.
 * \param[in]     ptrh  The pointer-handle to which the allocated
 *                      request will be assigned.
 * \note This runs in constant time.
 */

sc_status sc_lase_submitRequest(sc_lase* pl, sc_scriptRequestPtrh* ptrh);
/**< Submit the script request referenced by \p ptrh for execution in
 *   \p pl.
 *
 * This submits the script for execution, but does *not* wait for it to
 * complete.
 *
 * \param[in,out] pl    The LASE system in which the script will be executed.
 * \param[in]     ptrh  A reference to the script to submit for execution.
 *
 * \note You are guaranteed that scripts will be executed in the order
 *       they are submitted.
 * 
 * \see sc_lase_executeRequest
 * 
 */

sc_status sc_lase_executeRequest(sc_lase* pl, sc_scriptRequestPtrh* ptrh);
/**< Executes the script request referenced by \p ptrh in LASE \p pl.
 *
 * This executes the script synchronously.  That is, it submits the script
 * request and *does* wait for it to complete.
 *
 * \param[in,out] pl    The LASE system in which the script is executed.
 * \param[in]     ptrh  A reference to the script to execute.
 *
 * \note You are guaranteed that scripts will be executed in the order
 *       they are submitted.
 * 
 * \see sc_lase_submitRequest to submit the request without having to
 *      wait.
 * 
 */

#endif /* __si_sha_lase_h__2bf6de29_af3b_4a35_803b_7262e61900f3 */
