#ifndef __si_sha_sreqpool_h__bef99ef7_3764_4c6e_bfdd_819e790fd643
#define __si_sha_sreqpool_h__bef99ef7_3764_4c6e_bfdd_819e790fd643

/** \file si_sha/sreqpool.h
 *  \brief A script request object and pool.
 *
 *  Script request objects are used to feed JavaScript commands to the 
 *  JavaScript context.
 *  
 */

#include <si_sys/pthrdsem.h>
#include <si_sys/blkalloc.h>
#include <si_std/string.h>
#include <si_std/ptrh.h>
#include <si_std/listt.h>
#include <si_sha/autopkt.h>

/** \brief Enumeration of flag bits for script requests.
 *
 * Requests can have JSON response formatting turned on/off.
 * Requests can be packets relayed from the network or scripts.
 */
typedef enum se_scriptRequestFlags {
  se_scriptRequest_ReturnJSON   = 1,
  se_scriptRequest_RelayMessage = 2
} se_scriptRequestFlags;

/** \brief A script request object.
 *
 *  Script requests are submitted to a queue that is serviced by the
 *  JavaScript engine thread that owns the JavaScript context.
 *  Requests can either be to relay a packet from a network to a
 *  particular object, or to execute a given set of script commands.
 */
typedef struct sc_scriptRequest {
  sc_rep     _rep;          /**< Requests are pointer-handled. */
  s_uint32   flags;         /**< Requests flags. */

  /* for normal script requests */
  sc_string  scriptIn;      /**< JavaScript to execute. */
  sc_string  scriptOut;     /**< Script result. */

  /* for relaying packets to script */
  sc_automationPacket* pkt; /**< Packet relay request */
  const TCHAR*         obj; /**< Object identifier to relay to. */
  
  sc_sem     complete;      /**< Signals a waiting client that a particular 
                             *   request has completed. */
  sc_status  scriptStatus;  /**< Exit status of the script. */
} sc_scriptRequest;

sc_status sc_scriptRequest_clear(sc_scriptRequest* psr);
/**< Clear the script request object, resetting it to newly constructed
 *   state, without changing capacity.
 * \param[in,out] psr Script request object to clear.
 */

/** \brief Macro declares sc_scriptRequestPtrh: script request
 *         pointer-handle objects.
 */
sm_ptrhDecl(sc_scriptRequest);

/** \brief Macro declares sc_scriptRequestPtrhList: a list of script request
 *         pointer-handle objects.
 */
sm_listDecl(sc_scriptRequestPtrh);

/** \brief A pool of script request objects.
 *
 * Used to allocate script request objects in constant time.
 */
typedef struct sc_scriptRequestPool {
  pthread_mutex_t            mux_;          /**< Mutex for thread safety. */
  sc_scriptRequestPtrhList   pool_;         /**< Free list of requests. */
  s_int32                    currCapacity_; /**< Number of requests in pool.*/
  s_int32                    maxCapacity_;  /**< Max requests allowed in pool.*/
  sc_blockAllocator          al_;           /**< Block allocator for growable, 
                                             *   constant time block 
                                             *   allocation/release */
} sc_scriptRequestPool;


sc_status sc_scriptRequestPool_construct(sc_scriptRequestPool* ppool,
                                         s_int32 initialCapacity,
                                         s_int32 maxCapacity);
/**< Constructs a script request pool with the given \p initialCapacity of
 *   requests growable up to \p maxCapacity.
 *   
 * \param[in,out] ppool       The script request pool to construct.
 * \param[in]     initialCapacity The number of requests to ready on
 *                            construction.
 * \param[in]     maxCapacity The maximum number of requests to allow
 *                            the pool to grow to.  
 */

sc_status sc_scriptRequestPool_destruct(sc_scriptRequestPool* ppool);
/**< Destructs the given \p sc_scriptRequestPool.
 * 
 * \param[in,out] ppool The script request pool to destruct.
 * \note All requests must be returned to the pool before destruction.
 */

sc_status sc_scriptRequestPool_allocate(sc_scriptRequestPool* ppool,
                                        sc_scriptRequestPtrh* pptrh);
/**<  Allocates a request from the script request pool.
 * 
 * \param[in,out] ppool The pool to allocate from.
 * \param[out]    pptrh The pointer-handle to assign the
 *                      allocated request to.
 */

sc_status sc_scriptRequestPool_free(sc_scriptRequestPool* ppool,
                                    sc_scriptRequestPtrh* pptrh);
/**< Releases a request back to the \p sc_scriptRequestPool.
 * 
 * \param[in,out] ppool The pool to return the request to.
 * \param[in]     pptrh The pointer-handle to the request being returned.
 */


#endif /* __si_sha_sreqpool_h__bef99ef7_3764_4c6e_bfdd_819e790fd643 */