#ifndef __si_sha_zigbdev_h__861d580b_5f07_4139_aedd_9f19632dbf73
#define __si_sha_zigbdev_h__861d580b_5f07_4139_aedd_9f19632dbf73

/** \file si_sha/zigbdev.h
 *  \brief ZigBee Bridge device
 *
 *  The ZigBee bridge device derives from the base automation
 *  device, and relays messages between the JavaScript and the low-level
 *  ZigBee bridge controller.
 */

#include <si_sha/abrgdev.h>

/** \brief The ZigBee bridge device class.
 *
 *  The ZigBee bridge device derives from the base automation
 *  device, and relays messages between the JavaScript and the low-level
 *  ZigBee bridge controller.
 */
typedef struct sc_zigbeeBridgeDevice {
  sc_automationBridgeDevice base_;   /**< base class */
  sc_automationPacketPool   pool_;   /**< ZigBee packet pool */
} sc_zigbeeBridgeDevice;

sc_status sc_zigbeeBridgeDevice_construct(sc_zigbeeBridgeDevice* brg,
                                          sc_automationDeviceMgr* pdm,
                                          const TCHAR* name,
                                          const TCHAR* basePath,
                                          s_uint8 channel,
                                          s_uint16 panId);
/**< Constructs the ZigBee bridge device \p brg with the given \p name.
 *
 * Upon construction, the bridge should create a network on \p channel,
 * with \p panId, if such a network doesn't already exist.
 *
 * The frequency channel can be one of the 16 channels 11-26 inclusive.
 * The panId can be any 16 bit value and is user assigned. It only
 * serves to distinguish networks if there happen to be multiple 
 * overlapping networks in a local area.
 * 
 * \param[in,out] brg       The zigbee bridge device to construct.
 * \param[in]     pdm       The device manager for name resolution.
 * \param[in]     name      The name of the zigbee bridge.
 * \param[in]     basePath  The basePath of the system.
 * \param[in]     channel   The frequency channel (11-26) for the network.
 * \param[in]     panId     The user assigned 16 bit panId.
 *
 * \note To change the channel and panId assigned at start time, use the
 *       utility lase-bridge-cfg.pl with the --args command line parameter.
 * \note On secure ZigBee networks, each device must know the channel
 *       and panId parameters of the network.
 */

sc_status sc_zigbeeBridgeDevice_destruct(sc_zigbeeBridgeDevice* brg);
/**< Destructs ZigBee bridge device \p brg.
 */

sc_status sc_zigbeeBridgeDevice_newPacket(sc_zigbeeBridgeDevice* brg,
                                          sc_automationPacketPtrh* pptrh);
/**< Returns a new packet object from the ZigBee device pool.
 *
 * \param[in,out]  brg   The bridge device to use.
 * \param[out]    pptrh  The pointer-handle where the newly allolcated
 *                       packet will be assigned.
 *                       
 * \note The packet will be automatically returned to this pool when
 *       all references to it are released.
 */

sc_status sc_zigbeeBridgeDevice_processEvent(sc_zigbeeBridgeDevice* brg,
                                             sc_automationPacket* pkt);
/**< Processes the network event in \p pkt, relaying it to the appropriate
 *   JavaScript object via the device manager.
 *
 * \param[in,out]  brg  The bridge device to use.
 * \param[in]      pkt  The event packet to relay.
 *
 * \note Derived class override.
 */

sc_status sc_zigbeeBridgeDevice_processRequest(sc_zigbeeBridgeDevice* brg,
                                               sc_automationRequest* req,
                                               sc_automationResponse* res);
/**< Processes the given request \p req and returns a response in \p res.
 *
 * Processes the request by serializing it to the ZigBee bridge
 * controller/driver.  
 *
 * \param[in,out]  brg  The bridge device to use.
 * \param[in]      req  The request to send.
 * \param[out]     res  The response to the request.
 *
 * \note Derived class override.
 */

sc_status sc_zigbeeBridgeDevice_discover(sc_zigbeeBridgeDevice* brg);
/**< Discovers devices that have joined the ZigBee network.
 *
 * \note Not currently implemented.
 */

#endif /* __si_sha_zigbdev_h__861d580b_5f07_4139_aedd_9f19632dbf73 */
