#ifndef __si_sha_zigbrg_h__760f62ef_027d_40ff_9763_ede80a65e02d
#define __si_sha_zigbrg_h__760f62ef_027d_40ff_9763_ede80a65e02d

/** \file   si_sha/zigbrg.h
 * \brief  Low-level ZigBee bridge driver.
 * 
 *  This object is the low level bridge driver for the ZigBee
 *  device network.  The C member functions of this object encode the
 *  binary serial messages to the controller, and follow the bridge <->
 *  host protocol.  This object currently works with the Atmel Raven
 *  USB stick, with specially modified firmware. It has been assigned
 *  the device type 02-00-01.
 */

#include <si_sys/serial.h>
#include <si_sha/autobrg.h>
#include <si_sha/zigbee.h>

/** \brief ZigBee Bridge driver
 *
 *  This object is the low level bridge driver for the ZigBee
 *  device network.  The C member functions of this object encode the
 *  binary serial messages to the controller, and follow the bridge <->
 *  host protocol.  This object currently works with the Atmel Raven
 *  USB stick, with specially modified firmware. It has been assigned
 *  the device type 02-00-01.
 */
typedef struct sc_zigbeeBridge {
  sc_automationBridgeBase base_; /**< Automation bridge base class. */
  sc_serialPort port_;   /**< The serial port the bridge is connected to. */

  s_uint64 address_;     /**< The bridge network address. */
  s_uint32 type_;        /**< The bridge device type. */
  s_uint32 version_;     /**< The bridge software version */
  
  sc_automationPacketPtrhList ooo_; /**< List of out-of-order packets from 
                                         the network. */
} sc_zigbeeBridge;

sc_status sc_zigbeeBridge_construct(sc_zigbeeBridge* pzb,
                                    const TCHAR* name,
                                    const TCHAR* devName,
                                    const TCHAR* basePath);
/**< Constructs a ZigBee bridge driver on the serial port with
 * the special filename \p devName.
 *
 * \param[in,out] pzb   The bridge object to construct.
 * \param[in]    name   The name of the bridge.  Used to name the Unix
 *                      socket its interface is provided on.
 * \param[in]  devName  The name of the 'special' file for the serial
 *                      port the bridge is attached to.
 * \param[in] basePath  The base path for the application.
 */

sc_status sc_zigbeeBridge_destruct(sc_zigbeeBridge* pb);
/**< Destructs the given insteon bridge \p pb.
 */

sc_status sc_zigbeeBridge_setMode(sc_zigbeeBridge* pzb,
                                  s_uint8 mode);
/**< Sets the operational mode of the bridge controller.
 * 
 * \param[in,out] pzb   The ZigBee bridge.
 * \param[in]     mode  The operating mode of the bridge. Can be either
 *                      0 (sniffer) or 3 (network).
 *
 * \note When used in LASE, the bridge is always set to network mode.
 */


sc_status sc_zigbeeBridge_formNetwork(sc_zigbeeBridge* pzb,
                                      s_uint8 channel, s_uint16 panId);
/**< Forms the ZigBee network on the specified \p channel and with the
 * specified \p panId.
 *
 * The frequency channel can be one of the 16 channels 11-26 inclusive.
 * The panId can be any 16 bit value and is user assigned. It only
 * serves to distinguish networks if there happen to be multiple 
 * overlapping networks in a local area.
 * 
 * \param[in,out] pzb       The ZigBee bridge.
 * \param[in]     channel   The frequency channel (11-26) for the network.
 * \param[in]     panId     The user assigned 16 bit panId.
 *
 * \note To change the channel and panId assigned at start time, use the
 *       utility lase-bridge-cfg.pl with the --args command line parameter.
 * \note On secure ZigBee networks, each device must know the channel
 *       and panId parameters of the network.
 */


sc_status sc_zigbeeBridge_permitJoining(sc_zigbeeBridge* pzb,
                                        s_uint8 permitted);
/**< Sets the bridge to permit new devices to join the network.
 * 
 * Typically left on when configuring the network, and turned off
 * sometime later.
 *
 * \param[in,out] pzb         The ZigBee bridge.
 * \param[in]     permitted   Non-zero to permit joining, zero otherwise.
 */

sc_status sc_zigbeeBridge_setLeds(sc_zigbeeBridge* pzb, s_uint8 lmask);
/**< Sets the state of the 4 LEDs available on the Raven USB stick.
 *
 * This can be a useful debugging tool.
 *
 * \param[in,out] pzb     The ZigBee bridge.
 * \param[in]     lmask   The LED mask, can have any value from 0-15.
 * 
 */


#endif /* __si_sha_zigbrg_h__760f62ef_027d_40ff_9763_ede80a65e02d */

