#ifndef __si_sys_list_h__2e82ce77_7819_45a9_b682_5e4a386054d7
#define __si_sys_list_h__2e82ce77_7819_45a9_b682_5e4a386054d7

#include <si_sys/defs.h>
#include <si_sys/blkalloc.h>

/*! \file si_std/list.h
 *  \brief Generic list container. Not meant to be used directly.
 *  
 *  \see listt.h for an implementation of typesafe lists.
 * 
 */

typedef struct sc_listNodeAllocator sc_listNodeAllocator;

/*! \brief Generic list node structure
 *
 * Contains list control structures.
 */
typedef struct sc_listNode
{
  struct sc_listNode*   next; /**< Next node */
  struct sc_listNode*   prev; /**< Prev node */
  void*                 ownr; /**< Reserved */
} sc_listNode;

/*! \brief Generic list object. Not meant to be used directly.
 *
 * \see listt.h for type-safe lists.
 */
typedef struct sc_list {
  sc_listNode* head; /**< The head of the list */
  sc_listNode  end;  /**< The end node of the list */
  s_int32      nodeCount; /**< The count of nodes in the list */
    
  sc_listNodeAllocator* nodeAllocator; /**< allocator for list nodes */
} sc_list;


/*! \name List item function signatures. */
/*@{*/
typedef sc_status (*sc_listNode_constructCopyf)(void* pitem, void* pothr);
/**< Item copy constructor.
 * \param[in] pitem  The item to construct.
 * \param[in] pothr  The item to copy.
 */
typedef sc_status (*sc_listNode_assignf)(void* pitem, void* pothr);
/**< Item assignment operator.
 * \param[in] pitem  The item to assign to.
 * \param[in] pothr  The value to assign.
 */
typedef sc_status (*sc_listNode_destructf)(void* pitem);
/**< Item destructor.
 * \param[in] pitem  The item to destruct.
 */

/*@}*/

/*! \brief List node descriptor
 *
 *  Holds enough information about list items to maintain the list.
 */
typedef struct sc_listNodeDescriptor
{
  s_int32                       nodeSize;

  sc_listNode_constructCopyf    constructCopy;
  sc_listNode_assignf           assign;
  sc_listNode_destructf         destruct;

} sc_listNodeDescriptor;

/*! \brief List iterator
 */
typedef struct sc_listIterator
{
  struct { /* don't look at this ... */
    sc_listNode* node;
  } priv_;
} sc_listIterator;


sc_status sc_listIterator_init(sc_listIterator* pit);
/**< Init the iterator to NULL. In this state isEnd returns seTrue. */
sc_status sc_listIterator_next(sc_listIterator* pit);
/**< Move the iterator to the next element. */
sc_status sc_listIterator_prev(sc_listIterator* pit);
/**< Move the iterator to the previous element. */


sc_status sc_list_construct(sc_list* pl,
                            const sc_listNodeDescriptor* ndesc,
                            s_int32 initialCapacity, s_int32 maxCapacity);
/**< Generic list container constructor.
 * \param[in,out] pl    The list object to construct.
 * \param[in]     ndesc The descriptor for the list items.
 * \param[in]     initialCapacity The allocator will make sure there is
 *                enough storage for initialCapacity items to be added without
 *                allocator growth.
 * \param[in]     maxCapacity The maximum number of items the list is 
 *                allowed to hold.  If no particular maximum is
 *                desired, set this INT_MAX.
 */

sc_status sc_list_destruct(sc_list* pl);
/**< Generic list destructor.
 * \param[in,out] pl The list object to destruct.
 */

sc_status sc_list_getBegin(sc_list* pl, sc_listIterator* pit);
/**< Returns an iterator positioned at the first element of the list.
 *
 * If the list is empty, this will be the end of the list.
 * 
 * \param[in,out] pl   The list.
 * \param[in,out] pit  The iterator to position.
 */

sc_status sc_list_getEnd(sc_list* pl, sc_listIterator* pit);
/**< Returns an iterator positioned past the last element of the list.
 *
 * \param[in,out] pl   The list.
 * \param[in,out] pit  The iterator to position.
 */

sc_status sc_list_getIsEmpty(sc_list* pl, s_bool* pIsEmpty);
/**< Returns seTrue if the list is empty and seFalse otherwise.
 * \param[in,out] pl        The list to test.
 * \param[out]    pIsEmpty  On exit, set to seTrue if the list is
 *                           empty, seFalse otherwise.
 */

sc_status sc_list_clear(sc_list* pl);
/**< Erases all items from the list.
 * \param[in,out] pl The list to clear.
 */

sc_status sc_list_insert(sc_list* pl, sc_listIterator* it, void* nt);
/**< Inserts new item nt into the given list, after the element pointed
 * to by 'it'.
 *
 * \param[in,out] pl   The list to modify.
 * \param[in]     it   The element after which the insert should occur.
 * \param[in]     nt   The new item to insert.
 */

sc_status sc_list_erase(sc_list* pl, sc_listIterator* it);
/**< Erase the element 'it' from the given list.
 *
 * The element is removed from the list and freed.
 * 
 * \param[in,out] pl  The list to remove from.
 * \param[in]     it  The list element to remove and free.
 */

sc_status sc_list_remove(sc_list* pl, sc_listIterator* it);
/**< Remove the element 'it' from the given list.
 *
 * The element is removed from the list.
 * 
 * \param[in,out] pl  The list to remove from.
 * \param[in]     it  The list element to remove.
 */


sc_status sc_list_pushBack(sc_list* pl, void* pnt);
/**< Pushes a new item onto the end of the list.
 * \param[in,out] pl  The list to modify.
 * \param[in]     pnt The item to add.
 *
 * \note Items are added by value. If you want the list to refer to
 * particular objects, consider making a list of pointer handles.
 */

sc_status sc_list_pushFront(sc_list* pl, void* pnt);
/**< Pushes a new item onto the front of the list.
 * \param[in,out] pl  The list to modify.
 * \param[in]     pnt The item to add.
 *
 * \note Items are added by value. If you want the list to refer to
 * particular objects, consider making a list of pointer handles.
 */

sc_status sc_list_popBack(sc_list* pl, void* pnt);
/**< Pops an item from the end of the list.
 *
 * An error is returned if the list is empty.
 * 
 * \param[in,out] pl  The list to modify.
 * \param[in]     pnt The item to which the popped value will be
 *                    assigned.
 *
 * \note Items are popped by value. If you want the list to refer to
 * particular objects, consider making a list of pointer handles.
 */

sc_status sc_list_popFront(sc_list* pl, void* pnt);
/**< Pops an item from the front of the list.
 * 
 * An error is returned if the list is empty.
 * 
 * \param[in,out] pl  The list to modify.
 * \param[in]     pnt The item to which the popped value will be
 *                    assigned.
 *
 * \note Items are popped by value. If you want the list to refer to
 * particular objects, consider making a list of pointer handles.
 * 
 */

sc_status sc_list_front(sc_list* pl, void** ppnt);
/**< Returns a pointer to the item at the front of the list.
 * 
 * An error is returned if the list is empty.
 * 
 * \param[in,out] pl    The list
 * \param[out]    ppnt  The item pointer to modify.
 */

sc_status sc_list_back(sc_list* pl, void** ppnt);
/**< Returns a pointer to the last item in the list.
 *
 * An error is returned if the list is empty.
 * 
 * \param[in,out] pl    The list
 * \param[out]    ppnt  The item pointer to modify.
 */

sc_status sc_list_getSize(sc_list* pl, s_uint32* sz);
/**< Returns the number of elements in the list.
 *
 * \param[in,out] pl    The list
 * \param[out]    sz    The size of the list
 */


#endif /* __si_sys_list_h__2e82ce77_7819_45a9_b682_5e4a386054d7 */
