#ifndef __si_std_listt_h__eb634e8a_92f0_4ae9_8246_7289de8737a8
#define __si_std_listt_h__eb634e8a_92f0_4ae9_8246_7289de8737a8

/*! \file si_std/listt.h
 *  \brief A template file that generates strongly typed lists of user
 *  defined types.
 *
 *  This file contains the macros sm_listDecl and sm_listImpl for
 *  declaring and implementing list containers of user defined types.
 *  For example, sm_listDecl(ut_myRecord); would declare the types:
 *   - ut_myRecordList
 *   - ut_myRecordListIterator
 *
 *  In the file implementing ut_myRecord, one would simply add the line:
 *   sm_listImpl(ut_myRecord);
 *  to implement the list for this type.
 *  This is a process nearly identical to that used with C++ templates, but is
 *  usable from C code, and thus portable to small systems.
 *  For more documentation on list members, see list.h    
 *  
 */

#include <si_std/list.h>


/*! \brief This macro is an implementation detail and should be ignored
 *  by users of the list facilities here.
 */
#define sm_nodePtr(it)\
    (it)->priv_.node

/*! \brief Declares a strongly typed list iterator for user defined
 *         type T.
 *
 *  These functions are declared inline as static functions to allow
 *  the optimizer to remove the function call.  In particular, _get,
 *  _next, _prev and _isEnd will typically be used in loops containing
 *  potentially large numbers of elements, so this can be important.
 *  For the same reason, there is little error checking in release mode.
 */
#define sm_listIteratorDecl(T)\
    typedef sc_listIterator T##ListIterator;\
    static void T##ListIterator_init(T##ListIterator* I)\
    { sm_nodePtr(I) = NULL; }\
    static T* T##ListIterator_get(T##ListIterator* I)\
    { return (T*)(sm_nodePtr(I)+1); }\
    static void T##ListIterator_next(T##ListIterator* I)\
    {  sm_nodePtr(I) = sm_nodePtr(I)->next; }\
    static void T##ListIterator_prev(T##ListIterator* I)\
    {  sm_nodePtr(I) = sm_nodePtr(I)->prev; }\
    static s_bool T##ListIterator_isEnd(T##ListIterator* I)\
    { return ( (sm_nodePtr(I) == NULL) || (sm_nodePtr(I)->next == NULL) ) \
               ? seTrue : seFalse; }

/*! \brief A macro declaring a list containing items of the given type T.
 *  This macro declares strongly typed wrappers of all the list member
 *  functions given in list.h.
 *
 *  \note This declaration assumes the list type will be used in a
 *  statically linked fashion.
 */
#define sm_listDecl(T)\
    sm_listIteratorDecl(T);\
    typedef sc_list T##List;\
    typedef s_int32 (*T##ListNode_compare)(T* pLHS, T* pRHS);\
    sc_status T##List_construct(T##List* pl,\
                                s_int32 initialCapacity, s_int32 maxCapacity);\
    sc_status T##List_destruct(T##List* pl);\
    sc_status T##List_assign(T##List* pl, T##List* prhs);\
    \
    sc_status T##List_getBegin(T##List* pl, T##ListIterator* pit);\
    sc_status T##List_getEnd(T##List* pl, T##ListIterator* pit);\
    sc_status T##List_getIsEmpty(T##List* pl, s_bool* pb);\
    \
    sc_status T##List_clear(T##List* pl);\
    sc_status T##List_insert(T##List* pl, T##ListIterator* it, T* nt);\
    sc_status T##List_erase(T##List* pl, T##ListIterator* it);\
    sc_status T##List_remove(T##List* pl, T##ListIterator* it);\
    \
    sc_status T##List_pushBack(T##List* pl, T* pnt);\
    sc_status T##List_pushFront(T##List* pl, T* pnt);\
    sc_status T##List_popBack(T##List* pl, T* pnt);\
    sc_status T##List_popFront(T##List* pl, T* pnt);\
    \
    sc_status T##List_front(T##List* pl, T** pnt);\
    sc_status T##List_back(T##List* pl, T** pnt);\
    \
    sc_status T##List_getSize(T##List* pl, s_uint32* sz);\
    

/*! \brief A macro implementing a list template for the given type T.
 *  This includes all the member functions declared with the macro
 *  sm_listDecl.
 *
 *  \note This macro assumes that you have implemented the following
 *  functions for your type T:
 *  
 *  - sc_status T_constructCopy(T*, const T*);
 *  - sc_status T_assign(T*, const T*);
 *  - sc_status T_destruct(T*);
 *
 *  and that the list will be used in a statically linked fashion.
 */
#define sm_listImpl(T)\
    sc_status T##List_construct(T##List* pl,\
                                s_int32 icap, s_int32 mcap)\
    { \
      static const sc_listNodeDescriptor lv = { sizeof(T), \
      (sc_listNode_constructCopyf)T##_constructCopy,\
      (sc_listNode_assignf)T##_assign,\
      (sc_listNode_destructf)T##_destruct }; \
      return sc_list_construct(pl,&lv,icap,mcap); \
    }\
    sc_status T##List_destruct(T##List* pl)\
    { return sc_list_destruct(pl); }\
    sc_status T##List_getBegin(T##List* pl, T##ListIterator* pit)\
    { return sc_list_getBegin(pl,pit); }\
    sc_status T##List_getEnd(T##List* pl, T##ListIterator* pit)\
    { return sc_list_getEnd(pl,pit); }\
    sc_status T##List_getIsEmpty(T##List* pl, s_bool* pb)\
    { return sc_list_getIsEmpty(pl,pb); }\
    \
    sc_status T##List_insert(T##List* pl, T##ListIterator* it, T* nt)\
    { return sc_list_insert(pl,it,nt); }\
    sc_status T##List_erase(T##List* pl, T##ListIterator* it)\
    { return sc_list_erase(pl, it); }\
    sc_status T##List_remove(T##List* pl, T##ListIterator* it)\
    { return sc_list_remove(pl,it); }\
    sc_status T##List_clear(T##List* pl)\
    { return sc_list_clear(pl); }\
    \
    sc_status T##List_pushBack(T##List* pl, T* pnt)\
    { return sc_list_pushBack(pl,pnt); }\
    sc_status T##List_pushFront(T##List* pl, T* pnt)\
    { return sc_list_pushFront(pl,pnt); }\
    \
    sc_status T##List_popBack(T##List* pl, T* pnt)\
    { return sc_list_popBack(pl,pnt); }\
    sc_status T##List_popFront(T##List* pl, T* pnt)\
    { return sc_list_popFront(pl,pnt); }\
    \
    sc_status T##List_front(T##List* pl, T** ppnt)\
    { return sc_list_front(pl,(void**)ppnt); }\
    sc_status T##List_back(T##List* pl, T** ppnt)\
    { return sc_list_back(pl,(void**)ppnt); }\
    \
    sc_status T##List_getSize(T##List* pl, s_uint32* sz)\
    { return sc_list_getSize(pl,sz); }\




#endif /* __si_std_listt_h__eb634e8a_92f0_4ae9_8246_7289de8737a8 */
