#ifndef __si_std_queue_h__3b94b7d2_2a35_42a4_bd1a_4617fa9f7d2e
#define __si_std_queue_h__3b94b7d2_2a35_42a4_bd1a_4617fa9f7d2e

#include <si_sys/defs.h>
#include <si_sys/blkalloc.h>

/*! \file si_std/queue.h
 *  \brief Queue (FIFO) data structure.
 *
 * \note This generic implementation is not meant to be used directly.
 * \see queuet.h for typesafe queues.
 */

typedef struct sc_queueNode sc_queueNode;

/*! \brief A queueNode data structure.
 *
 *  Queues as implemented here are singly linked.
 */
struct sc_queueNode {
  sc_queueNode* next;
};

typedef sc_status (*sc_queueNode_constructCopy_f)(void*, void*);
/**< Copy constructor function signature for queue items.  */
typedef sc_status (*sc_queueNode_assign_f)(void*, void*);
/**< Assignment function signature for queue items.  */
typedef sc_status (*sc_queueNode_destruct_f)(void*);
/**< Destruct function signature for queue items.  */


/*! \brief A queue node descriptor.
 *
 *  Holds enough information about queue nodes to maintain the queue
 *  and its items.
 */
typedef struct sc_queueNodeDesc {
  s_uint32                       nodeSize; /**< sizeof item + sizeof sc_queueNode*/
  sc_queueNode_constructCopy_f   constructCopy; 
  sc_queueNode_assign_f          assign;
  sc_queueNode_destruct_f        destruct;  
} sc_queueNodeDesc;

/*! \brief Generic Queue object.
 *  \note Not meant to be used directly. \see sm_queueDecl.
 */
typedef struct sc_queue {

  sc_blockAllocator       blka; /**< allocator for queueNodes */
  const sc_queueNodeDesc* desc; /**< node descriptor */
  sc_queueNode*           head; /**< head of the queue (front)*/
  sc_queueNode*           tail; /**< tail of the queue (back) */

} sc_queue;

sc_status sc_queue_construct(sc_queue* q,
                             const sc_queueNodeDesc* desc,
                             s_int32 initialCapacity);
/**< Constructor for the generic queue object.
 * \param[in,out] q The queue object to construct.
 * \param[in] desc The descriptor for the queue items.
 * \param[in] initialCapacity The allocator will make sure there is
 * enough storage for initialCapacity items to be pushed without
 * allocator growth.
 */

sc_status sc_queue_pushBack(sc_queue* q, void* pnt);
/**< Push item pnt onto the back of the queue q.
 * \note Items are pushed by value.  To push a reference to a
 * particular object, make a queue of pointer-handle objects.
 */

sc_status sc_queue_popFront(sc_queue* q, void* pnt);
/**< Pop item pnt from the front of the queue q.
 * \param[in,out] q Queue to pop from.
 * \param[in,out] pnt Item to assign the value to.
 *
 * \note An error is returned if the queue is empty.
 * \note Items are popped by value.  To queue references to 
 * particular objects, make a queue of pointer-handles.
 */

sc_status sc_queue_destruct(sc_queue* q);
/**< Destruct the queue q. */


#endif /* __si_std_queue_h__3b94b7d2_2a35_42a4_bd1a_4617fa9f7d2e */
