#ifndef __si_std_string_h__ff6c1067_65b8_4fe7_8704_86e99cb0d45b
#define __si_std_string_h__ff6c1067_65b8_4fe7_8704_86e99cb0d45b

/*! \file si_std/string.h
 *  \brief A string class and utility functions.
 */

#include <si_sys/defs.h>


/*! \brief A dynamic string class.   
 *
 *  A dynamically growable string.  The string has both a length member
 *  and is NULL terminated.  The growth factor is 3/2. That is,
 *  whenever the string needs to grow, it allocates storage that is the
 *  required amount * 1.5.  This is done so that any small extra bits
 *  that are appended don't each cause an additional allocation/copy,
 *  at the expense of extra memory usage.  Thus, it is efficient to
 *  append to this string may times to form much a much larger string.
 *  The actual capacity available is kept at capacity_.
 */
typedef struct sc_string {
  TCHAR*    str_;       /**< The string buffer. */
  s_uint32  capacity_;  /**< The size of the string buffer */
  s_uint32  length_;    /**< The length of the string data at str_*/
} sc_string;

sc_status sc_string_construct(sc_string* pstr);
/**< Constructs an empty string of length 0.
 * Constructs an empty string.  No memory allocation is done in this
 * function.  The str_ member is set to NULL.
 */

sc_status sc_string_constructFrom(sc_string* pstr, const TCHAR* src);
/**< Constructs a new string that is a duplicate of src.
 * Constructs an empty string.  No memory allocation is done in this
 * function.  The str_ member is set to NULL.
 * \param[in,out] pstr The string to construct.
 * \param[in] src The source string to copy to the newly constructed object.
 */

sc_status sc_string_destruct(sc_string* pstr);
/**< destructs the string, freeing any allocated memory.
 */

sc_status sc_string_assign(sc_string* pstr, const TCHAR* src);
/**< Copies src to pstr, growing pstr if necessary.
 * \param[in,out] pstr The string to assign to.
 * \param[in] src The string value to copy to pstr.
 */

sc_status sc_string_assignChars(sc_string* pstr, const TCHAR* src,
                                s_uint32 nchars);
/**< Copies at most the first nchars characters of src to pstr, growing
 * pstr if necessary.
 *
 * If NULL is encountered before nchars, the copy is terminated.
 * 
 * \param[in,out] pstr The string to assign to.
 * \param[in] src The string value to copy to pstr.
 * \param[in] nchars The maximum number of characters to copy from src.
 */

sc_status sc_string_append(sc_string* pstr, const TCHAR* src);
/**< Appends the string src to pstr, growing pstr if necessary.
 * \param[in,out] pstr The string being appended to.
 * \param[in] src The string to append to pstr.
 * \note No allocation is done if src fits within pstr->capacity_.
 */

sc_status sc_string_appendChars(sc_string* pstr, const TCHAR* src,
                                s_uint32 nchars);
/**< Appends a maximum of nchars at src to pstr, growing pstr if necessary.
 * \param[in,out] pstr The string being appended to.
 * \param[in] src The string to append to pstr.
 * \param[in] nchars The maximum number of characters to append.
 * \note If the null character is encountered before nchars are copied,
 * the copy is terminated early, and success is returned.
 */

sc_status sc_string_prepend(sc_string* pstr, const TCHAR* src);
/**< Appends the string src to pstr, growing pstr if necessary.
 * \param[in,out] pstr The string being appended to.
 * \param[in] src The string to append to pstr.
 * \note No allocation is done if src fits within pstr->capacity_.
 */

sc_status sc_string_prependChars(sc_string* pstr, const TCHAR* src,
                                 s_uint32 nchars);
/**< Prepends a maximum of nchars at src to the start of pstr, growing
 *   pstr if necessary.
 * \param[in,out] pstr The string being pre-pended to.
 * \param[in] src The string to prepend to pstr.
 * \param[in] nchars The maximum number of characters to prepend.
 * \note If the null character is encountered before nchars are copied,
 * the copy is terminated early, and success is returned.
 */

sc_status sc_string_detach(sc_string* pstr, TCHAR** ppdst);
/**< Detached the given string's str_ storage, and returns it in ppdst.
 *
 * The string pstr is cleared to zero length and capacity, and its
 * current storage is returned in ppdst.  The caller becomes
 * responsible for freeing the storage, which can be done with free().
 * 
 * \param[in,out] pstr The string having its storage detached.
 * \param[out] ppdst The location of the string storage in pstr.
 */

sc_status sc_string_toLower(sc_string* pstr);
/**< Converts this string to lowercase (in place).
 *
 * \param[in,out] pstr The string to convert to lowercase.
 * 
 */

sc_status sc_string_trimRight(sc_string* pstr);
/**< Removes any whitespace from the end of the string.
 *
 * Eliminates whitespace characters from the end of the string.  Whitespace
 * includes space, tab, newline, and carriage-return. This is done in
 * place.
 * 
 * \param[in,out] pstr The string to trim.
 */

sc_status sc_string_free(sc_string* pstr);
/**< Sets the length and capacity of the string to zero, and frees any
 * memory used by the string.
 *
 * This function clears the string and frees any associated memory.  On
 * exit, the state of the string is the same as after running the
 * sc_string_construct function.
 * 
 * \see sc_string_construct, and sc_string_clear.
 * \param[in,out] pstr The string to free.
 */

sc_status sc_string_clear(sc_string* pstr);
/**< Sets the length of the string to 0 without changing the capacity.
 *
 * Sets length to 0, and str_ to "", without freeing any memory used by
 * the string.  Thus, capacity_ remains unchanged.
 * \see sc_string_free
 * \param[in,out] pstr The string to clear.
 */

sc_status sc_string_setMinCapacity(sc_string* pstr, s_uint32 numChars);
/**< Sets the minimum storage capacity of the string to numChars.
 * Sets the storage size of the str_ buffer to be a minimum of
 * numChars, growing the string if necessary.
 * \param[in,out] pstr The string whose capacity may be adjusted.
 * \param[in] numChars The minimum number of characters that may be
 * stored at str_.
 */




sc_status sc_string_allocCopy(TCHAR** dst, const TCHAR* src);
/**< This is a version of 'strdup' that uses sc_status error codes.
 * \param[out] dst pointer to a returned string pointer.
 * \param[in]  src pointer to a string to copy.
 */

int iswhitespace(TCHAR c);
/**< returns non-zero if c is a space, tab, newline, or carriage-return
 * character, zero otherwise.
 * \param[in] c Character to test.
 * \return non-zero if c is a whitespace character, zero if not.
 */

const TCHAR* str_advw(const TCHAR* psrc);
/**< Advances the given pointer psrc past any whitespace.
 * \param[in] psrc The pointer to advance.
 * \return psrc advanced past whitespace text.
 * 
 */

const TCHAR* str_advq(const TCHAR* psrc);
/**< Advances the given pointer psrc past any quoted
 *   text.
 * \param[in] psrc The pointer to advance.
 * \return psrc advanced past quoted text.
 */

const TCHAR* str_advl(const TCHAR* psrc);
/**< Advances the given pointer psrc to the next line.
 * \param[in] psrc The pointer to advance.
 * \return psrc advanced to the first character of the next line, or
 * NULL if no newline is found.
 */


int _stricmp(const char* s1, const char* s2);
/**< Case independent string comparison.
 * \param s1 The pointer to the first string.
 * \param s2 The pointer to the second string.
 *
 * \return non-zero if the strings do not match independent of case,
 * zero otherwise.
 */

#endif /* __si_std_string_h__ff6c1067_65b8_4fe7_8704_86e99cb0d45b */
