#ifndef __si_sys_blkalloc_h__138f5d42_147c_463e_899c_6d8a28944353
#define __si_sys_blkalloc_h__138f5d42_147c_463e_899c_6d8a28944353

/*! \file si_sys/blkalloc.h
 *  \brief Fixed size block allocator.
 *
 *  An object that can allocate and deallocate fixed sized blocks in
 *  constant time.  The storage (in blocks) can be fixed or growable at
 *  the caller's option.
 */

#include <stdlib.h>

typedef struct sc_bufferDesc sc_bufferDesc;
typedef struct sc_blockDesc sc_blockDesc;
struct sc_bufferDesc {
  sc_bufferDesc* next;
};
struct sc_blockDesc {
  sc_blockDesc* next;  
};

/*! \brief A fixed size block allocator object.
 *
 *  An object that can allocate and deallocate fixed sized blocks in
 *  constant time.  The storage (in blocks) can be fixed or growable at
 *  the caller's option.
 */
typedef struct sc_blockAllocator
{
  s_int32 nblocks;        /**< Number of blocks currently allocated. */
  s_int32 maxCapacity;    /**< Max number of blocks to allocate. */
  s_int32 blockSize;      /**< Size of blocks in bytes. */

  sc_blockDesc*  blocks;  /**< Free block list. */
  sc_bufferDesc* buffers; /**< List of buffers to free on destruction. */

} sc_blockAllocator;

sc_status sc_blockAllocator_construct(sc_blockAllocator* pba,
                                      s_int32 blockSize, 
                                      s_int32 initialCapacity,
                                      s_int32 maxCapacity);
/**< Constructs an allocator of initialCapacity blocks (growable 
 * up to maxCapacity), where each block contains blockSize bytes.
 * 
 * \param[in,out] pba             The block allocator object to construct.
 * \param[in]     blockSize       The size of each block in bytes.
 * \param[in]     initialCapacity The initial number of blocks to ready.
 * \param[in]     maxCapacity     The max number of blocks to allocate.
 */

sc_status sc_blockAllocator_destruct(sc_blockAllocator* pba);
/**< Destructs the given block allocator.
 *
 * \note: All allocated blocks must be freed before the destructor is called.
 */

sc_status sc_blockAllocator_allocate(sc_blockAllocator* pba,
                                     void** ppblock);
/**< Allocates a single block. May grow the block pool if necessary, up
 * to maxCapacity blocks.
 *
 * \param[in,out] pba      The allocator object to allocate from.
 * \param[out]    ppblock  The allocated block returned at *ppblock.
 * \note This function operates in constant time.
 */

sc_status sc_blockAllocator_free(sc_blockAllocator* pba,
                                 void* pblock);
/**< Frees a single block, returning it to the block pool.
 * 
 * \param[in,out] pba     The allocator object the block came from.
 * \param[in]     pblock  The block to return to the free block pool.
 * 
 * \note This function operates in constant time.
 */


#endif /* __si_sys_blkalloc_h__138f5d42_147c_463e_899c_6d8a28944353 */


