#ifndef __si_sys_h_pthrdsem_h__5c15a1dd_1951_4cc8_948b_b91cd51e709b
#define __si_sys_h_pthrdsem_h__5c15a1dd_1951_4cc8_948b_b91cd51e709b

/*! \file si_sys/pthrdsem.h
 *  \brief A portable counting semaphore for pthreads.
 *
 * Lightweight, portable simple counting semaphore for use with
 * pthreads.  This exists because Mac OS X does not support anonymous
 * semaphores (sem_init).  
 *  
 */

#include <pthread.h>
#include <si_sys/defs.h>

/*! \brief A portable counting semaphore for pthreads.
 *
 * Lightweight, portable simple counting semaphore for use with
 * pthreads.  This exists because Mac OS X does not support anonymous
 * semaphores (sem_init).  
 *
 * \note If a inter-process semaphore is desired, use posix semaphores, they 
 * are portable across linux implementations and OS X.
 * 
 */
typedef struct sc_sem {
  pthread_mutex_t m;   /**< Mutex protecting condition variable c. */
  pthread_cond_t  c;   /**< Condition variable for blocking when count is 0.*/
  s_int32         count; /**< The current semaphore count. */
  s_int32         max; /**< The maximum semaphore count. */
} sc_sem;

sc_status sc_sem_construct(sc_sem* s, s_int32 initialCount, s_int32 maxCount);
/**< Constructs semaphore s with a count of initialCount, and a maximum
 * of maxCount.
 *
 * \note To create the semphore initially unavailble, set initialCount
 * to 0.
 * \note To create a binary semaphore set maxCount to 1.
 */

sc_status sc_sem_wait(sc_sem* s);
/**< Attempts to acquire semaphore s, and blocks the calling thread if 
 * the semaphore is not available (count is 0).
 * 
 * \param[in,out] s The semaphore to wait for (acquire).
 */

sc_status sc_sem_trywait(sc_sem* s, s_bool* gotit);
/**< Attempts to acquire semaphore s, but does not block if the 
 * semaphore count is 0, returning immediately instead. If the semaphore 
 * has been acquired, *gotit is set to seTrue, and seFalse otherwise.
 * 
 * \param[in,out] s      The semaphore to try to acquire.
 * \param[out]    gotit  On exit, seTrue if the semaphore was acquired,
 *                       seFalse otherwise.
 */

sc_status sc_sem_post(sc_sem* s);
/**< Releases the semaphore s, incrementing its count, and potentially
 * unblocking a thread waiting in sc_sem_wait.
 * 
 * \param[in,out] s The semaphore to post (release).
 */

sc_status sc_sem_destruct(sc_sem* s);
/**< Destructs the semaphore s.
 */

#endif
