#ifndef __si_web_httpserv_h__65934665_d2e9_4f6b_81d5_3dfcd6197554
#define __si_web_httpserv_h__65934665_d2e9_4f6b_81d5_3dfcd6197554

/** \file si_web/httpserv.h
 *  \brief A tiny portable web server
 *
 *  This tiny web server supports HTTP 1.0 and serves up documents
 *  added to its internal document list via sc_httpServer_addDocument.
 *  The documents are created dynamically by a generator callback 
 *  function.
 */

#include <si_std/string.h>
#include <si_std/listt.h>
#include <pthread.h>

/** \name Forward declarations
 */
/*@{*/
typedef struct sc_httpDocument  sc_httpDocument;
typedef struct sc_httpRequest   sc_httpRequest;
typedef struct sc_httpResponse  sc_httpResponse;
typedef struct sc_httpUpload    sc_httpUpload;
/*@}*/

/** \brief Macro declares sc_httpDocumentList, a list of http
 *         documents.
 */
sm_listDecl(sc_httpDocument);

/*! \brief HTTP server object.
 *
 *  This tiny web server supports HTTP 1.0 and serves up documents
 *  added to its internal document list via sc_httpServer_addDocument.
 *  The documents are generally dynamically created by a generator
 *  callback function.
 */
typedef struct sc_httpServer
{
  int                 listenSock;    /**< Socket on which connections are accepted */
  int                 die[2];        /**< Pipes to signaling server thread to die */
  pthread_t           serverThread;  /**< Server thread */
  s_int32             started;       /**< Started bool */
  s_uint16            port;          /**< Port we are listening on */

  sc_httpDocumentList documents;     /**< List of installed documents */

} sc_httpServer;


sc_status sc_httpServer_construct(sc_httpServer* psrv, s_uint16 port);
/**< \brief Construct a new webserver on the given \p port.
 *
 * \param[in,out] psrv  The server object to construct.
 * \param[in]     port  The port for the server to listen on.
 * 
 * \note To start the server, you must call sc_httpServer_start.
 * \see sc_httpServer_start
 */

sc_status sc_httpServer_destruct(sc_httpServer* pserver);
/**< \brief Destructs the server \p pserver.
 */


sc_status sc_httpServer_addDocument(sc_httpServer* psrv,
                                    sc_httpDocument* pdoc);
/**< \brief Adds the document \p pdoc to the server.
 *
 * The server may only serve documents that have been explicitly added
 * using this function.
 * 
 * \param[in,out] psrv  The server object.
 * \param[in]     pdoc  The document object to add.
 */

sc_status sc_httpServer_start(sc_httpServer* psrv);
/**< \brief Starts the HTTP server on the port given at construction.
 *
 * \param[in,out] psrv  The server object to start.
 * 
 * \see To add documents to server, call http_server_addDocument.
 */

sc_status sc_httpServer_stop(sc_httpServer* psrv);
/**< \brief Stops the HTTP server \p psrv.
 * 
 * \param[in,out] psrv  The server object to stop.
 */

sc_status sc_httpResponse_write(sc_httpResponse* presp, const char* pstr);
/**< \brief Write the specified string as output HTML.
 * 
 * \param[in,out] presp    The HTTP response object.
 * \param[in]     pstr     The string to write as response output.
 * 
 * \note This function is intended to be used to return a web page
 *       from an http document callback.
 */

sc_status sc_httpResponse_writeBinary(sc_httpResponse* presp, 
                                      const s_uint8* pbuff,
                                      s_uint32 buffSize);
/**< \brief Write the specified data as binary output.
 * 
 * \param[in,out] presp     The HTTP response object.
 * \param[in]     pbuff     The buffer to write as response output.
 * \param[in]     buffSize  The number of bytes of pbuff to write.
 * 
 * \note This function is intended to be used to return data
 *       from an http document callback.  This version is especially
 *       useful when returning an image file (e.g. .jpg).  
 */

/** \brief Enumeration of HTTP request types.
 */
typedef enum sc_httpRequestType {
  sc_httpRequestType_GET  = 1,
  sc_httpRequestType_PUT  = 2,
  sc_httpRequestType_POST = 3,
  sc_httpRequestType_INV  = 0
} sc_httpRequestType;

/** \brief Enumeration of form argument types.
 */
typedef enum sc_httpRequest_argType {
  sc_httpRequest_argType_unknown = -1,
  sc_httpRequest_argType_normal = 0,
  sc_httpRequest_argType_upload = 1,
} sc_httpRequest_argType;

const char* sc_httpRequest_getBody(const sc_httpRequest* preq);
/**< \brief Returns the body of the request in case the request was a
 *          PUT or POST
 *
 * Could be used, for example, to read a SOAP envelope in a request.
 *
 * \param[in] preq     The HTTP request object.
 * \return Pointer to the body of the request, if present.
 */

sc_httpRequestType sc_httpRequest_getRequestType(const sc_httpRequest* preq);
/**< \brief Returns the type of the request: GET, PUT, or POST
 * 
 * \param[in] preq     The HTTP request object.
 * \return the type of the request: GET, PUT, or POST
 */


int sc_httpRequest_getArgCount(const sc_httpRequest* preq);
/**< \brief Returns the number of arguments posted with request \p preq.
 * 
 * \param[in] preq     The HTTP request object.
 * \return The number of arguments in request \p preq.
 */

const char* sc_httpRequest_getArgName(const sc_httpRequest* preq, 
                                      s_uint32 aidx);
/**< \brief Returns the name of the argument at \p aidx.
 * 
 * \param[in] preq     The HTTP request object.
 * \param[in] aidx     The index of the argument whose name will be
 *                     returned.
 *                     
 * \return The name of the argument at \p aidx.
 * 
 * \note If idx >= argCount, this function returns NULL.
 */

sc_httpRequest_argType sc_httpRequest_getArgType(const sc_httpRequest* preq, 
                                                 const char* argName);
/**< \brief Returns the type of the argument named \p argName.
 * 
 * \param[in] preq     The HTTP request object.
 * \param[in] argName  The name of the argument whose type will be
 *                     returned.
 * \return The type of the argument named \p argName.
 *                     
 * \note If argName is not a known argument, this function returns NULL.
 */

int sc_httpRequest_getArgTypeByIndex(const sc_httpRequest* preq, 
                                     s_uint32 aidx);
/**< \brief Returns the type of the argument at idx.
 * 
 * \param[in] preq     The HTTP request object.
 * \param[in] aidx     The index of the argument whose type will be
 *                     returned.
 * 
 * \note If idx >= argCount, this function returns NULL.
 */

const char* sc_httpRequest_getArgValue(const sc_httpRequest* preq, 
                                       const char* argName);
/**< \brief Returns the value of the argument named argName.
 * 
 * \param[in] preq     The HTTP request object.
 * \param[in] argName  The name of the argument to get.
 * 
 * \return The value of the specified argument as a string.
 * 
 * \note If argName is not a known argument, this function returns NULL.
 * \note This function is for use with "normal" arguments as returned by
 *       argType.  If the named argument is an "upload" argument, this 
 *       function returns NULL.
 */

const char* sc_httpRequest_getArgValueByIndex(const sc_httpRequest* preq,
                                              s_uint32 aidx);
/**< \brief Returns the value of the argument at idx.
 *
 * \param[in] preq     The HTTP request object.
 * \param[in] aidx     The index of the argument to get.
 *
 * \return The value of the specified argument as a string.
 * 
 * \note If idx >= argCount, this function returns NULL.
 * \note This function is for use with "normal" arguments as returned by
 *       argType.  If the named argument is an "upload" argument, this 
 *       function returns NULL.
 */

const sc_httpUpload*
   sc_httpRequest_getArgUploadValue(const sc_httpRequest* preq,
                                    const char* argName);
/**< \brief Returns the value of the upload argument \p argName.
 *
 * \param[in] preq     The HTTP request object.
 * \param[in] argName  The name of the upload argument.
 *
 * \return The sc_httpUpload object with the name \p argName.
 * 
 * \note If argName is not a known argument, this function returns NULL.
 * \note This function is for use with "upload" arguments as returned by
 *       argType.  If the named argument is a "normal" argument, this 
 *       function returns NULL.
 * \note Information about the upload (name, filename, content type, etc) 
 *       can be determined using the sc_httpUpload functions.
 */

const sc_httpUpload*
   sc_httpRequest_getArgUploadValueByIndex(const sc_httpRequest* preq,
                                           s_uint32 aidx);
/**< \brief Returns the value of the upload argument at \p aidx.
 *
 * \param[in] preq     The HTTP request object.
 * \param[in] aidx     The index of the upload argument.
 * 
 * \return The sc_httpUpload object with at index \p aidx.

 * \note If idx >= argCount, this function returns NULL.
 * \note This function is for use with "upload" arguments as returned by
 *       argType.  If the argument specified is a "normal" argument, this 
 *       function returns NULL.
 * \note Information about the upload (name, filename, content type, etc) 
 *       can be determined using the sc_httpUpload functions.
 */



s_uint32 sc_httpUpload_getContentSize(const sc_httpUpload* upl);
/**< \brief Returns the size of uploaded content.
 * \return The size in bytes of the uploaded content.
 */

const char* sc_httpUpload_getContentType(const sc_httpUpload* upl);
/**< \brief Returns the content type of the uploaded content (mime type)
 * \return The MIME type of the uploaded content.
 */

const char* sc_httpUpload_getContentFilename(const sc_httpUpload* upl);
/**< \brief Returns the filename of the uploaded content as posted
 * \return The 'filename' parameter of the uploaded content.
 */

sc_status sc_httpUpload_getContent(const sc_httpUpload* upl,
                                   s_uint8* pOutBuf);
/**< \brief Copies the uploaded content to pOutBuf, which must be large 
 *          enough to accept it.
 * 
 * \see sc_httpUpload_getContentSize for how to determine the size of the
 *      uploaded content.
 */

/** \brief Signature of document generator callback function. */
typedef int (*sc_httpDocumentCallback)(const sc_httpRequest* req,
                                       sc_httpResponse* resp);

/** \brief HTTP Document object */
struct sc_httpDocument {
  char* name;          /**< name of the document as it will be served. */
  sc_httpDocumentCallback callback; /**< Callback function to dynamically 
                                         create the document. */
};

sc_status sc_httpDocument_construct(sc_httpDocument* doc, const char* name,
                                    sc_httpDocumentCallback callback);
/**< \brief Construct a new document \p doc with the given \p name and \p
 * callback generator function.
 *
 * \param[in,out] doc      The document to construct.
 * \param[in]     name     The name of the document as served.
 * \param[in]     callback The callback function that generates the
 *                         document dynamically when requested.
 */
sc_status sc_httpDocument_constructCopy(sc_httpDocument* doc,
                                        sc_httpDocument* oth);
/**< \brief Constructs document \p doc, copying members of \p oth.
 * 
 * \param[in,out] doc   The document to construct.
 * \param[in]     oth   The document to copy.
 *
 * \note This function is needed for sc_httpDocumentList.
 */
sc_status sc_httpDocument_assign(sc_httpDocument* doc,
                                 sc_httpDocument* rhs);
/**< \brief Assigns document \p doc, the value of \p rhs.
 * 
 * \param[in,out] doc   The document to assign to.
 * \param[in]     rhs   The value to assign.
 *
 * \note This function is needed for sc_httpDocumentList.
 */

sc_status sc_httpDocument_destruct(sc_httpDocument* doc);
/**< \brief Destructs document \p doc.
 */

/** \brief Macro that can create a static document object from a C
 *         array of document data.
 */
#define sm_HTTP_DOC_FROM_BUFFER(filename, buffername) \
    int dump_##buffername (const sc_httpRequest* preq, sc_httpResponse* presp) \
    { \
      return sc_httpResponse_writeBinary(presp, buffername, sizeof(buffername)); \
    }\
    sc_httpDocument http_doc_##buffername = { filename, dump_##buffername }


#endif /* __si_web_httpserv_h__65934665_d2e9_4f6b_81d5_3dfcd6197554 */

