#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include <si_web/httpserv.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>

/* stuff specific to unix */
#define SOCKET int
#define closesocket close
#define INVALID_SOCKET -1
#define sockets_open()  /**/
#define sockets_close() /**/


struct sc_httpUpload
{
  sc_string name;
  sc_string content_type;
  sc_string filename;
  s_uint32  content_size;
  s_uint8*  pcontent;
  sc_httpUpload* next;
};

struct sc_httpRequest 
{
  s_uint32  content_length;
  s_int32   requestType;
  sc_string body;

  sc_string user_agent;
  char      remote_host[20];

  sc_string document;
  
  s_uint32  form_arg_count;
  sc_string form_args;
  
  s_uint32  query_arg_count;
  sc_string query_args;
  
  sc_string boundary;
  
  s_uint32  form_encoding;
  s_uint32  upload_arg_count;
  sc_httpUpload* pupload;
}; 

struct sc_httpResponse 
{
  sc_string document;
};



#define FORM_ENCODING_URL 0
#define FORM_ENCODING_MULTIPART 1

sc_status sc_httpResponse_construct(sc_httpResponse* presp);
sc_status sc_httpResponse_destruct(sc_httpResponse* presp);
sc_status sc_httpResponse_clear(sc_httpResponse* presp);

sc_status sc_httpRequest_construct(sc_httpRequest* preq);
sc_status sc_httpRequest_destruct(sc_httpRequest* preq);
sc_status sc_httpRequest_clear(sc_httpRequest* preq);

sc_status sc_httpRequest_parseFromHeader(sc_httpRequest* preq, 
                                   const char* header, 
                                   const char** phend);

int sc_httpRequest_parseArgs(sc_httpRequest*);

sc_status sc_httpResponse_writeHeader(sc_httpResponse*,
                                      const sc_httpRequest*, int);

sc_status sc_httpUpload_construct(sc_httpUpload* pupl);
sc_status sc_httpUpload_destruct(sc_httpUpload* pupl);
sc_status sc_httpUpload_append(sc_httpUpload* ptupl,
                               sc_httpUpload* pnupl);



static const char* cont_type[] = {
  "text/html",
  "text/plain",
  "text/xml",
  "image/gif",
  "image/jpeg",
  "*/*"
  // lots of others possible!
};   

const char* http_content_type(const char* ext) 
{
  if(ext == 0) return cont_type[0];
  if(*ext == '.') ext++;
  if(_stricmp(ext, "htm")==0) 
    return cont_type[0];
  if(_stricmp(ext, "html")==0) 
    return cont_type[0];
  if(_stricmp(ext, "txt")==0) 
    return cont_type[1];
  if(_stricmp(ext, "xml")==0) 
    return cont_type[2];
  if(_stricmp(ext, "gif")==0) 
    return cont_type[3];
  if(_stricmp(ext, "jpg")==0) 
    return cont_type[4];
  if(_stricmp(ext, "jpeg")==0) 
    return cont_type[4];
  return cont_type[4];
}

const char * http_status(int status)
{
    switch (status) {
      case 200: return "OK";
      case 201: return "Created";
      case 202: return "Accepted";
      case 203: return "Partial information";
      case 204: return "No response";
      case 301: return "Moved";
      case 302: return "Found";
      case 303: return "Method";
      case 304: return "Not modified";
      case 400: return "Bad request";
      case 401: return "Unauthorized";
      case 402: return "Payment required";
      case 403: return "Forbidden";
      case 404: return "Not Found";
      case 500: return "Internal error";
      case 501: return "Not implemented";
      default:  return "Internal Error (Unknown)";
    }
}

sc_status sc_httpServer_construct(sc_httpServer* psrv, s_uint16 port)
{
  sc_status status = sc_status_success;

  status = sc_httpDocumentList_construct(&psrv->documents, 8, INT_MAX);
  sm_status_returnIfFailed(status);

  psrv->port = port;
  psrv->started = 0;
  psrv->listenSock = INVALID_SOCKET;

  pipe(psrv->die);
  
  sm_status_return(status);
}

sc_status sc_httpServer_destruct(sc_httpServer* psrv)
{
  sc_status status = sc_status_success;
  if(psrv->started != 0)
    sm_status_returnError(status, sc_status_failure);

  status = sc_httpDocumentList_clear(&psrv->documents);
  sm_status_returnIfFailed(status);
  status = sc_httpDocumentList_destruct(&psrv->documents);
  sm_status_returnIfFailed(status);

  close(psrv->die[0]);
  close(psrv->die[1]);  
  
  sm_status_return(status);
}

sc_status sc_httpServer_findDocument(sc_httpServer* pserver, 
                                     const char* name,
                                     sc_httpDocument** ppdoc)
{
  sc_status status = sc_status_success;
  sc_httpDocument* pdoc=0;
  sc_httpDocumentListIterator it;
  s_uint32 idx = 0;
  
  if(!pserver || !name || !ppdoc)
    sm_status_returnError(status, sc_err_nullArgument);

  status = sc_httpDocumentList_getBegin(&pserver->documents, &it);
  sm_status_returnIfFailed(status);
  
  for(; !sc_httpDocumentListIterator_isEnd(&it);
         sc_httpDocumentListIterator_next(&it) )
  {
    pdoc = sc_httpDocumentListIterator_get(&it);
    if(_stricmp(name, pdoc->name)==0) {
      *ppdoc = pdoc;
      sm_status_return(status);
    }
  }  
  *ppdoc = NULL;
  sm_status_return(status);
}

void* http_listen(void* pargs)
{
  SOCKET ws = 0;
  struct sockaddr_in sain = {AF_INET}, csain = {AF_INET};
  sc_httpServer* pws = (sc_httpServer*)(pargs);
  sc_httpDocument *pdoc = 0;
  int rv = 0, nfds;
  size_t len=0;
  socklen_t sz = sizeof(sain);
  sc_string s;
  char buf[128];
  sc_httpRequest req;
  sc_httpResponse resp;
  const char * phend=0;
  fd_set rfds;
  
  sc_string_construct(&s);
  sc_httpRequest_construct(&req);
  sc_httpResponse_construct(&resp);
  
  sockets_open();
  
  pws->listenSock = socket(AF_INET, SOCK_STREAM, 0);
  
  sain.sin_port = htons(pws->port);
  sain.sin_addr.s_addr = htonl(INADDR_ANY);  

  rv = bind(pws->listenSock, (struct sockaddr*)&sain, sz);   
  if(rv == -1) {    
    printf("bind returned error: %d\n", rv);
    return 0;
  }
  
  rv = listen(pws->listenSock, 5);
  if(rv == -1) {
    printf("listen returned error: %d\n", rv);
    return 0;
  }

  nfds = 1 + ((pws->listenSock > pws->die[0]) ? pws->listenSock : pws->die[0]);

  while(1) 
  {
    FD_ZERO(&rfds);
    FD_SET(pws->listenSock, &rfds);
    FD_SET(pws->die[0], &rfds);
    
    rv = select(nfds, &rfds, NULL, NULL, NULL);
    if(rv < 0)
      break; /* error */
    if(rv == 0)
      continue; /* timeout */
    if(FD_ISSET(pws->die[0], &rfds))
      break; /* shutdown */
    
    ws = accept(pws->listenSock, (struct sockaddr*)&csain, &sz); 
    if(ws != INVALID_SOCKET) {  
    
#ifdef _DEBUG
      s_uint32 ca = csain.sin_addr.s_addr;
      printf("connection accepted from: ");
      printf("%u.", ca & 0x000000ff );
      ca >>= 8;
      printf("%u.", ca & 0x000000ff );
      ca >>= 8;
      printf("%u.", ca & 0x000000ff );
      ca >>= 8;
      printf("%u\n", ca & 0x000000ff );     
#endif
          
      sc_string_clear(&s);
      while( (rv = recv(ws, buf, sizeof(buf), 0)) != 0) {
        sc_string_appendChars(&s, buf, rv);
        if(strstr(s.str_, "\n\r\n"))
          break; /* end of header */
      }
      
#ifdef _DEBUG     
      printf("Header received: \n");
      printf(s.str_);         
#endif      

      sc_httpRequest_clear(&req);
      if(sc_httpRequest_parseFromHeader(&req, s.str_, &phend) == 0)
      {        
        sc_httpResponse_clear(&resp);
        
        /* load remaining posted data from socket */
        if(req.content_length != 0) {
          len = strlen(phend);
          sc_string_appendChars(&req.body, phend, (s_uint32)len);
          sc_string_appendChars(&req.form_args, phend, (s_uint32)len);
          while(len < req.content_length) {
            rv = recv(ws, buf, sizeof(buf), 0);
            sc_string_appendChars(&req.body, buf, rv);
            sc_string_appendChars(&req.form_args, buf, rv);
            len += rv;
          }
          if(req.form_args.str_[req.form_args.length_-1] == '\n') {
            req.form_args.str_[req.form_args.length_-1] = '\0';
            req.form_args.length_--;
          }
          if(req.form_args.str_[req.form_args.length_-1] == '\r') {
            req.form_args.str_[req.form_args.length_-1] = '\0';
            req.form_args.length_--;
          }
        }
#ifdef _DEBUG
        if(req.form_args.length_ > 0)
          printf("%s\n", req.form_args.str_);
#endif        
        sc_httpRequest_parseArgs(&req);
        
        /* locate requested document */
        
        sc_httpServer_findDocument(pws, req.document.str_, &pdoc);
        if(pdoc) {      
          sc_httpResponse_writeHeader(&resp, &req, 200);
          if(pdoc->callback) {
            (*pdoc->callback)(&req, &resp);
          } else {
            /* read document in 'name' and write out */
          }
          
          {
            const char* pbuf = resp.document.str_;
            s_uint32 bytes_rem = resp.document.length_;
            while(bytes_rem) {
              rv = send(ws, (void*)pbuf, bytes_rem, 0);
                        bytes_rem -= rv;
                        pbuf += rv;
            }
          }
          close(ws);
        } else {
          sc_httpResponse_writeHeader(&resp, &req, 404);

          {
            const char* pbuf = resp.document.str_;
            s_uint32 bytes_rem = resp.document.length_;          
            while(bytes_rem) {
              rv = send(ws, (void*)pbuf, bytes_rem, 0);
              bytes_rem -= rv;
              pbuf += rv;
            }
          }
#ifdef _DEBUG     
          printf("404 sent:\n%s", resp.document.str_);
#endif      
          
          close(ws);
        }
      }
    } else break;
  }
  
  printf("HTTP server exiting...\n");
  
  sockets_close();
  return 0;
}
 
sc_status sc_httpServer_start(sc_httpServer* psrv)
{
  if(++psrv->started == 1) {
    pthread_create(&psrv->serverThread, NULL, http_listen, psrv);
  }
  return sc_status_success;
}

sc_status sc_httpServer_stop(sc_httpServer* psrv)
{
  char ctrlD = 0x04;
  
  if(--psrv->started == 0) {
    /* signal server thread to exit */
    write(psrv->die[1], &ctrlD, 1);
    /* wait for the server thread to exit. */
    pthread_join(psrv->serverThread, NULL);
    
    close(psrv->listenSock);
    psrv->listenSock = INVALID_SOCKET;
    psrv->serverThread = (pthread_t)-1;
  }
  return sc_status_success;
}

sc_status sc_httpUpload_construct(sc_httpUpload* pupl)
{
  pupl->next = 0;
  pupl->content_size = 0;
  sc_string_construct(&pupl->content_type);
  sc_string_construct(&pupl->name);
  sc_string_construct(&pupl->filename);
  pupl->pcontent = 0;
  
  return sc_status_success;
}
sc_status sc_httpUpload_destruct(sc_httpUpload* pupl)
{
  if(pupl->next) 
    sc_httpUpload_destruct(pupl->next);
  pupl->content_size = 0;
  sc_string_destruct(&pupl->content_type);
  sc_string_destruct(&pupl->name);
  sc_string_destruct(&pupl->filename);
  if(pupl->pcontent)
    free(pupl->pcontent);
  pupl->pcontent = 0;
  return sc_status_success;
}
sc_status sc_httpUpload_append(sc_httpUpload* ptupl, sc_httpUpload* pnupl)
{
  sc_httpUpload* pupl = ptupl;
    while(pupl->next) pupl = pupl->next;
  pupl->next = pnupl;
  return sc_status_success;
}

sc_status sc_httpRequest_construct(sc_httpRequest* preq)
{
  preq->upload_arg_count=0;
  preq->query_arg_count = 0;
  preq->form_arg_count = 0;
  preq->requestType = sc_httpRequestType_INV;
  preq->content_length=0;
  preq->form_encoding = FORM_ENCODING_URL;
  memset(preq->remote_host, 0, sizeof(preq->remote_host));
  sc_string_construct(&preq->query_args);
  sc_string_construct(&preq->form_args);
  sc_string_construct(&preq->document);
  sc_string_construct(&preq->user_agent);
  sc_string_construct(&preq->boundary);
  sc_string_construct(&preq->body);
  preq->pupload = 0;
  return sc_status_success;
}
sc_status sc_httpRequest_destruct(sc_httpRequest* preq)
{
  s_uint32 uidx=0;
  sc_string_destruct(&preq->query_args);
  sc_string_destruct(&preq->form_args);
  sc_string_destruct(&preq->document);
  sc_string_destruct(&preq->user_agent);
  sc_string_destruct(&preq->boundary);  
  sc_string_destruct(&preq->body);  
  if(preq->pupload)
    sc_httpUpload_destruct(preq->pupload);
  return sc_status_success;
}
int sc_httpRequest_clear(sc_httpRequest* preq)
{
  s_uint32 uidx = 0;
  preq->form_encoding = FORM_ENCODING_URL;
  preq->query_arg_count = 0;
  preq->form_arg_count  = 0;
  preq->requestType = sc_httpRequestType_INV;
  preq->content_length = 0;
  preq->remote_host[0] = '\0';
  
  sc_string_clear(&preq->query_args);
  sc_string_clear(&preq->form_args);
  sc_string_clear(&preq->document);
  sc_string_clear(&preq->user_agent);
  sc_string_clear(&preq->body);

  // get rid of all upload related stuff  
  sc_string_free(&preq->boundary);    
  if(preq->pupload)
    sc_httpUpload_destruct(preq->pupload);
  preq->pupload = 0;
  preq->upload_arg_count = 0;
  return 0;
}

const char* strstrb(const char* pbuf, int bufsize, const char* ppat, int patsize)
{
  int sidx=0, cidx=0, laststartidx = bufsize - patsize;
  
  if(laststartidx < 0)
    return 0;
  
  while(sidx <= laststartidx) {
    for(cidx = 0; cidx < patsize; cidx++) 
      if(pbuf[sidx + cidx] != ppat[cidx]) break;    
    if(cidx == patsize)
      return pbuf + sidx;
    sidx++;
  }
  
  return 0;
}

int sc_httpRequest_parseFromHeader(sc_httpRequest* preq, 
                                   const char* header, 
                                   const char** phend)
{
  s_uint32 sz;
  char ch;
  const char* phdr = str_advw(header);
  const char* ptmp = phdr+1; 

  // parse HTTP requestType
  while(!iswhitespace(*ptmp)) 
    ptmp++;
  if(strncmp(phdr, "GET", 3) ==0)
    preq->requestType = sc_httpRequestType_GET;
  else if(strncmp(phdr, "POST", 4) == 0) 
    preq->requestType = sc_httpRequestType_POST;
  else if(strncmp(phdr, "PUT", 3) == 0)
    preq->requestType = sc_httpRequestType_PUT;
  else
    preq->requestType = sc_httpRequestType_INV; /* return error */
  
  // parse document 
  phdr = str_advw(ptmp);
  if(phdr[0] == '/') phdr++; // assume no directory support
  ptmp = phdr;
  ch = *ptmp;
  while(ch != '?' && !iswhitespace(ch)) 
    ch = *++ptmp;
  sz = (s_uint32)(ptmp-phdr);
  sc_string_assignChars(&preq->document, phdr, sz);

  // tack on query string args if present
  if(ch == '?') {
    phdr = ++ptmp;
    ch = *ptmp;
    while(!iswhitespace(ch)) 
      ch = *++ptmp;
    sc_string_appendChars(&preq->query_args, phdr, (s_uint32)(ptmp-phdr));
  }

  // if this is a post request, read the content-type
  if(preq->requestType == sc_httpRequestType_POST) {
    phdr = strstr(ptmp, "Content-Type:");
    phdr = str_advw(phdr+13);
    ptmp = phdr;
    if(strncmp(phdr, "multipart/form-data", 19) == 0) {
    preq->form_encoding = FORM_ENCODING_MULTIPART;
    sc_string_assign(&preq->boundary, "--");
    // read multipart boundary signature
    ptmp = strstr(phdr+19,"boundary=");
    if(!ptmp) return -1;
    phdr = str_advw(ptmp+9);
    ptmp = phdr;
    while(!iswhitespace(*ptmp)) ptmp++;
        sc_string_appendChars(&preq->boundary, phdr, (s_uint32)(ptmp-phdr));
    } else // otherwise, assume standard www URL encoding.
      preq->form_encoding = FORM_ENCODING_URL;
  } else 
    preq->form_encoding = FORM_ENCODING_URL;    

  // parse user agent
  phdr = strstr(header, "User-Agent:");
  phdr = str_advw(phdr+11);
  ptmp = phdr;
  while((*ptmp) != '\r') 
    ptmp++;
  sz = (s_uint32)(ptmp-phdr);
  sc_string_assignChars(&preq->user_agent, phdr, sz);
    
  // parse host tag
  phdr = strstr(header, "Host:");
  phdr = str_advw(phdr+5);
  ptmp = phdr+1;
  while(!iswhitespace(*ptmp)) 
    ptmp++;
  if( (sz = (s_uint32)(ptmp-phdr)) < sizeof(preq->remote_host)-1) {
    strncpy(preq->remote_host, phdr, sz);
    preq->remote_host[sz] = '\0';
  } else { return -1; }

  // parse content length
  phdr = strstr(header, "Content-Length:");
  if(phdr) {
    phdr = str_advw(phdr+15);
    ptmp = phdr+1;
    while(!iswhitespace(*ptmp)) 
      ptmp++;
    preq->content_length = atoi(phdr);
  }

  // parse connection type
  //phdr = strstr(ptmp, "Connection:");
  //phdr = str_advw(phdr+11);
  //if(strnicmp(phdr, "Keep-Alive", 10)==0)
  //  r.conn_ = eKeepAlive;
  //else 
  //  r.conn_ = eDiscard;

  *phend = strstr(ptmp, "\r\n\r\n") + 4;
  return 0;
}

int parse_args_url_encoded(const sc_string* inargs, sc_string* outargs)
{
  char asc[3] = {0,0,0};
  const char *pargr = 0;
  char *pargw = 0;
  char ch;
  int arg_count;
  
  if (!inargs->str_) return 0;  
  arg_count = 0;
  pargr = inargs->str_;
  pargw = outargs->str_;
  // replace all + with ' ' and then all %XX with ascii equiv
  ch = *pargr;
  while(ch != '\0') 
  {
    if(ch == '&') {     
      *pargw++ = '\0';
    } else if(ch == '=') {
      *pargw++ = '\0';
      arg_count++;
    } else if(ch == '%') {
      asc[0] = *++pargr;
      asc[1] = *++pargr;      
      *pargw++ = (char)strtoul(asc,0,16);
    } else if(ch == '+') {
      *pargw++ = ' ';
    } else {
      *pargw++ = ch;
    }
    ch = *++pargr;
  }
  *pargw = '\0';
  return arg_count;
}

sc_status parse_args_multipart_encoded(sc_httpRequest* preq)
{
  sc_httpUpload *pupl=0;
  sc_string fargs;
  const char* ptmp=0, *pdata=0, *pend=0, *phdr=preq->form_args.str_;
  int blen = preq->boundary.length_;
  int alen = preq->form_args.length_;
  
  sc_string_construct(&fargs);
  phdr = strstr(phdr, preq->boundary.str_);
  if(phdr == 0) return -1;
  phdr += blen;

  while(phdr) 
  {
    pdata = strstr(phdr, "\r\n\r\n");
    pdata += 4;

    pupl = (sc_httpUpload*)malloc(sizeof(sc_httpUpload));
    sc_httpUpload_construct(pupl);

    // pull out name of field
    ptmp = strstr(phdr, "name=");
    ptmp += 5;
    if(*ptmp == '\"' ) { 
      pend = ++ptmp;  // remove starting quote
      while(*pend != '\"') pend++;
        sc_string_appendChars(&pupl->name, ptmp, (s_uint32)(pend-ptmp));   
    } else {
      pend = ptmp;
      while (!iswhitespace(*pend)) pend++;
      sc_string_appendChars(&pupl->name, ptmp, (s_uint32)(pend-ptmp));
    } 
    ptmp = pend;
    if(*ptmp == '\"') ptmp++;
    if(*ptmp == ';') ptmp++;
    // pull out filename if it is on this line.
    while(*ptmp != '\n') {
      if(iswhitespace(*ptmp)) ptmp++;
      else break;   
    }
    if(*ptmp != '\n') 
    { // there's something else on this line...
      if(strncmp(ptmp, "filename=", 9) == 0) 
      {  // filename= appears.  extract name.
        ptmp+=9;
        if(*ptmp == '\"' ) { 
          pend = ++ptmp; // remove starting quote
          while(*pend != '\"') pend++;
          sc_string_assignChars(&pupl->filename, ptmp, (s_uint32)(pend-ptmp));
          if(pupl->filename.str_ == 0) sc_string_assign(&pupl->filename, "");
        } else {
          pend = ptmp; 
          while(!iswhitespace(*pend)) pend++;
          sc_string_assignChars(&pupl->filename, ptmp, (s_uint32)(pend-ptmp));
        }
      }
    }
  
    ptmp = str_advl(ptmp);
    if(pupl->filename.str_ != NULL)
    {  // pull out content-type if it appears   
      if(strncmp(ptmp, "Content-Type:", 13) == 0) 
      {     
        ptmp = str_advw(ptmp+13);
        pend = ptmp;
        while(!iswhitespace(*pend)) pend++;
        sc_string_assignChars(&pupl->content_type, ptmp, (s_uint32)(pend-ptmp));
      }
    } else ptmp = str_advl(ptmp); // skip this line

    // Find the end of the data by searching for the boundary pattern. Note
    // that we can't use string routines for this as there may be lots of 0s in
    // the data that would stop the search. call our own binary pattern search.
    pend = strstrb(pdata, alen - (s_uint32)(pdata - preq->form_args.str_),
                   preq->boundary.str_, blen);
    pupl->content_size = (s_uint32)(pend-pdata) - 2; // -2 because \r\n is tacked onto data

    if(pupl->filename.str_ != NULL) {
      pupl->pcontent = (s_uint8*)malloc(pupl->content_size);
      memcpy(pupl->pcontent, pdata, pupl->content_size);
      // okay, this upload structure has been filled out, add to list
      if(preq->pupload == 0) preq->pupload = pupl;
      else sc_httpUpload_append(preq->pupload,pupl);
      preq->upload_arg_count++;
    } else {
      // copy name and value into form_args
      sc_string_appendChars(&fargs, pupl->name.str_, pupl->name.length_);
      sc_string_appendChars(&fargs, "\0", 1);
      sc_string_appendChars(&fargs, pdata, pupl->content_size);
      sc_string_appendChars(&fargs, "\0", 1);
      preq->form_arg_count++;
      sc_httpUpload_destruct(pupl);
      free(pupl);
      pupl = 0;
    } 
    phdr = pend + blen;
    if(strncmp(phdr, "--", 2) == 0)
      break; // done!
  }
  sc_string_free(&preq->form_args);
  preq->form_args = fargs; 
  return 0;
}

sc_status sc_httpRequest_parseArgs(sc_httpRequest* preq)
{ 
  /* parse query args */
  preq->query_arg_count = parse_args_url_encoded(&preq->query_args, 
                                                 &preq->query_args);
  
  /* parse form args. these may be url-encoded or multipart-encoded. */
  if(preq->form_encoding == FORM_ENCODING_URL) {
    preq->form_arg_count = parse_args_url_encoded(&preq->form_args,
                                                  &preq->form_args);
  } else {
    parse_args_multipart_encoded(preq);
  }
  return sc_status_success;
}

int sc_httpRequest_getArgCount(const sc_httpRequest* preq)
{
  return preq->query_arg_count + preq->form_arg_count + preq->upload_arg_count;
}

const char* sc_httpRequest_getArgValue(const sc_httpRequest* preq, 
                                       const char* argname)
{
  const char* pargs = preq->query_args.str_;
  s_uint32 idx=0;

  if (!argname) return 0;
  for(idx = 0; idx < preq->query_arg_count; idx++) 
  {
    if(_stricmp(pargs, argname) == 0) {
      while(*pargs != '\0') pargs++;
      pargs++;
      return pargs;
    }
    while(*pargs != '\0') pargs++;
    pargs++;
    while(*pargs != '\0') pargs++;
    pargs++;
  }
  // try form args
  pargs = preq->form_args.str_;
  for(idx = 0; idx < preq->form_arg_count; idx++) 
  {
    if(_stricmp(pargs, argname) == 0) {
      while(*pargs != '\0') pargs++;
      pargs++;
      return pargs;
    }
    while(*pargs != '\0') pargs++;
    pargs++;
    while(*pargs != '\0') pargs++;
    pargs++;
  }
  // not found.
  return 0;
}

const char* sc_httpRequest_getArgValueByIndex(const sc_httpRequest* preq,
                                              s_uint32 idx)
{
  const char* pargs=0;
  s_uint32 count;
  if(preq == NULL) return NULL;
  count = preq->query_arg_count + preq->form_arg_count;
  if (idx >= count) return NULL;
  
  pargs = sc_httpRequest_getArgName(preq, idx);
  if(!pargs) return NULL;

  while(*pargs != '\0')
    pargs++;
  
  pargs++;
  return pargs;
}

const char* sc_httpRequest_getArgName(const sc_httpRequest* preq,
                                      s_uint32 idx)
{
  const char* pargs = 0;
  s_uint32 count;
  if(preq == NULL) return NULL;
  count = preq->query_arg_count + preq->form_arg_count;

  if(idx < preq->query_arg_count) {
    pargs = preq->query_args.str_;
    while(idx--) {
      while(*pargs != '\0') pargs++;
      pargs++;
      while(*pargs != '\0') pargs++;
      pargs++;
    }
    return pargs;
  } else if(idx < (preq->query_arg_count + preq->form_arg_count)) {
      pargs = preq->form_args.str_;
    idx -= preq->query_arg_count; 
    while(idx--) {
      while(*pargs != '\0') pargs++;
      pargs++;
      while(*pargs != '\0') pargs++;
      pargs++;
    }
    return pargs;
  } else {
    sc_httpUpload* pupl = preq->pupload;
    idx -= preq->query_arg_count + preq->form_arg_count;
    if(idx >= preq->upload_arg_count) return NULL;
    while(idx--) pupl = pupl->next;
    return pupl->name.str_;
  }
}

const sc_httpUpload*
sc_httpRequest_getArgUploadValue(const sc_httpRequest* preq, 
                                 const char* argName)
{
  sc_httpUpload* pupl = 0;
  if(preq == NULL || argName == NULL)
    return NULL;
  
  pupl = preq->pupload;
  while(pupl != NULL) {
    if(_stricmp(pupl->name.str_, argName) == 0) break;
    pupl = pupl->next;
  }
  
  return pupl;
}

const sc_httpUpload*
   sc_httpRequest_getArgUploadValueByIndex(const sc_httpRequest* preq,
                                           s_uint32 idx)
{
  int i = 0;
    s_uint32 acount=0;
  sc_httpUpload* pupl = preq->pupload;

  if(preq == NULL) return NULL;
  acount = preq->query_arg_count + preq->form_arg_count;
  if(idx < acount) return NULL; // wrong argument type
  idx -= acount;
  if(idx >= preq->upload_arg_count) return 0; // index too large
  
  pupl = preq->pupload;
  while(idx--) 
    pupl = pupl->next;  
  return pupl;
}

sc_httpRequest_argType
   sc_httpRequest_getArgType(const sc_httpRequest* preq, 
                             const char* argName)
{
  const char* pargs=0;
  sc_httpUpload* pupl =0;

  s_uint32 idx=0;
  if(!argName) return sc_httpRequest_argType_unknown;
    
  pargs = preq->query_args.str_;
  for(idx = 0; idx < preq->query_arg_count; idx++) 
  {
    if(_stricmp(pargs, argName) == 0) 
      return sc_httpRequest_argType_normal;  
    while(*pargs != '\0') pargs++;
    pargs++;
    while(*pargs != '\0') pargs++;
    pargs++;
  }
  // try form args
  pargs = preq->form_args.str_;
  for(idx = 0; idx < preq->form_arg_count; idx++) 
  {
    if(_stricmp(pargs, argName) == 0) 
      return sc_httpRequest_argType_normal;    
    while(*pargs != '\0') pargs++;
    pargs++;
    while(*pargs != '\0') pargs++;
    pargs++;
  }
  // try uploads...
    pupl = preq->pupload;
  while(pupl != NULL) {
    if(_stricmp(pupl->name.str_, argName)==0) break;
    pupl = pupl->next;
  }
  if(pupl) return sc_httpRequest_argType_upload;
  return sc_httpRequest_argType_unknown;
}

int sc_httpRequest_getArgTypeByIndex(const sc_httpRequest* preq, 
                                  s_uint32 idx)
{
  s_uint32 acount=0;
  if(preq == NULL) return sc_httpRequest_argType_unknown;
  acount = preq->query_arg_count + preq->form_arg_count;
  if(idx < 0) 
    return sc_httpRequest_argType_unknown;
  if(idx < acount) 
    return sc_httpRequest_argType_normal;
  if(idx < acount + preq->upload_arg_count)
      return sc_httpRequest_argType_upload;
  return sc_httpRequest_argType_unknown;
}

sc_status sc_httpResponse_construct(sc_httpResponse* presp)
{
  return sc_string_construct(&presp->document);
}
sc_status sc_httpResponse_destruct(sc_httpResponse* presp)
{
  return sc_string_destruct(&presp->document);
}
sc_status sc_httpResponse_clear(sc_httpResponse* presp)
{
  return sc_string_clear(&presp->document);
}
sc_status sc_httpResponse_write(sc_httpResponse* presp, const char* pstr)
{
  return sc_string_append(&presp->document, pstr);
}
sc_status sc_httpResponse_writeHeader(sc_httpResponse* presp,
                                      const sc_httpRequest* preq,
                                      int httpStatus)
{
  sc_status status = sc_status_success;
  char buf[80];
  const char* pdext=0;
  sprintf(buf,"HTTP/1.0 %d %s\r\n", httpStatus, http_status(httpStatus));
  
  sc_httpResponse_write(presp, buf);

  if(httpStatus == 200) {
    pdext = strrchr(preq->document.str_, '.');
    sprintf(buf, "Content-type: %s\r\n\r\n", http_content_type(pdext));
    sc_httpResponse_write(presp, buf);
  } else {
    sc_httpResponse_write(presp, "\r\n");
  }
  
  sm_status_return(status);
}

sc_status sc_httpServer_addDocument(sc_httpServer* psrv, sc_httpDocument* pdoc)
{
  sc_status status = sc_status_success;
  
  status = sc_httpDocumentList_pushBack(&psrv->documents, pdoc);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_httpResponse_writeBinary(sc_httpResponse* presp, 
                               const s_uint8* pbuf,
                               s_uint32 bufsize )
{
  return sc_string_appendChars(&presp->document, (const char*)pbuf, bufsize);
}


s_uint32 sc_httpUpload_getContentSize(const sc_httpUpload* pupl)
{
  return pupl->content_size;
}
const char* sc_httpUpload_getContentType(const sc_httpUpload* pupl)
{
  return pupl->content_type.str_;
}
const char* sc_httpUpload_getContentFilename(const sc_httpUpload* pupl)
{
  return pupl->filename.str_;
}
sc_status sc_httpUpload_getContent(const sc_httpUpload* pupl, s_uint8* pOutBuf)
{
  if(!pOutBuf) return -1;
  if(!pupl) return -1;
  memcpy(pOutBuf, pupl->pcontent, pupl->content_size);
  return 0;
}

sc_status sc_httpDocument_construct(sc_httpDocument* doc, const char* name,
                                    sc_httpDocumentCallback callback)
{
  sc_status status = sc_status_success;
  doc->name = NULL;
  if(name) {
    doc->name = strdup(name);
    if(doc->name == NULL)
      sm_status_returnError(status, sc_err_outOfMemory);
  }
  doc->callback = callback;
  sm_status_return(status);
}
sc_status sc_httpDocument_constructCopy(sc_httpDocument* doc,
                                        sc_httpDocument* oth)
{
  sc_status status = sc_status_success;
  doc->name = NULL;
  if(oth->name) {
    doc->name = strdup(oth->name);
    if(doc->name == NULL)
      sm_status_returnError(status, sc_err_outOfMemory);
  }
  doc->callback = oth->callback;
  sm_status_return(status);
}

sc_status sc_httpDocument_assign(sc_httpDocument* doc,
                                 sc_httpDocument* rhs)
{
  sc_status status = sc_status_success;

  if(doc->name != NULL) {
    free(doc->name);
    doc->name = NULL;
  }
  if(rhs->name) {
    doc->name = strdup(rhs->name);
    if(doc->name == NULL)
      sm_status_returnError(status, sc_err_outOfMemory);
  }
  doc->callback = rhs->callback;  

  sm_status_return(status);
}

sc_status sc_httpDocument_destruct(sc_httpDocument* doc)
{
  sc_status status = sc_status_success;
  if(doc->name != NULL)
    free(doc->name);
  sm_status_return(status);
}

const char* sc_httpRequest_getBody(const sc_httpRequest* preq)
{
  return preq->body.str_;
}
sc_httpRequestType sc_httpRequest_getRequestType(const sc_httpRequest* preq)
{
  return preq->requestType;
}

sm_listImpl(sc_httpDocument);
