#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#include <si_sha/instbrg.h>
#include <si_sha/autopkt.h>


static sc_status sc_insteonBridge_handleAsync(sc_insteonBridge* pb,
                                              const s_uint8* res);

sc_status sc_insteonBridge_serviceDevice(sc_insteonBridge* pb,
                                         sc_automationPacket* pkt);

sc_status sc_insteonBridge_processRequest(sc_insteonBridge* pb,
                                          sc_automationBridgeConnection* pc,
                                          sc_automationRequest*  req,
                                          sc_automationResponse* res);

sc_status sc_insteonBridge_construct(sc_insteonBridge* pb,
                                     const TCHAR* name,
                                     const TCHAR* devName,
                                     const TCHAR* basePath)
{
  static const sc_automationBridge_vtable vtbl = {
    (sc_automationBridge_serviceDeviceF)sc_insteonBridge_serviceDevice,
    (sc_automationBridge_processRequestF)sc_insteonBridge_processRequest
  };
  sc_status status = sc_status_success;
  sc_insteonBridgeVersion bv;
  char addr[12] = {0};

  /* the insteon PLM bridge uses a fixed baud of 19200 */
  status = sc_serialPort_construct(&pb->port_, devName,
                                   se_baudRate_19200);
  sm_status_returnIfFailed(status);

  pb->address_     = 0;
  pb->category_    = 0;
  pb->subcategory_ = 0;

  /* get actual address and type from bridge */
  status = sc_insteonBridge_getVersion(pb, &bv);
  if(sm_status_failed(status)) 
    sc_serialPort_destruct(&pb->port_);
  sm_status_returnIfFailed(status);

  pb->address_     = bv.address_;
  pb->category_    = bv.category_;
  pb->subcategory_ = bv.subcategory_;

  sprintf(addr, "%0X-%0X-%0X", (int)(bv.address_ >> 16),
          (int)((bv.address_ >> 8) & 0xFF),
          (int)(bv.address_ & 0xFF));

  /* can now initialize the base class */
  status = sc_automationBridgeBase_construct(&pb->base_, name, basePath,
                                             32, 16, &vtbl, pb);
  sm_status_returnIfFailed(status);

  /* let the user know what we found... */
  pb->base_.fdDev_ = pb->port_.fd_;
  printf("Insteon bridge %s [%s] starting on %s.\n", name, addr, devName );

  sm_status_return(status);
}

sc_status sc_insteonBridge_destruct(sc_insteonBridge* pb)
{
  sc_status status = sc_status_success;

  printf("Insteon bridge %s stopping.\n", pb->base_.name_);
  
  status = sc_automationBridgeBase_destruct(&pb->base_);
  sm_status_returnIfFailed(status);
  
  status = sc_serialPort_destruct(&pb->port_);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

static int getResponseLength(int code)
{ /* TODO: put the rest of these codes in the se_insteonCode enum */
  switch(code)
  {
    case se_insteonCode_stdInEvent: /* standard insteon message */
      return 9;
    case se_insteonCode_extInEvent: /* extended insteon message */
      return 21;
    case se_insteonCode_x10InEvent: /* not supported */
      return 2; 
    case se_insteonCode_linkStatusEvent: 
      return 8;
    case se_insteonCode_buttonEvent: 
      return 1;
    case se_insteonCode_resetLinksEvent: /* user reset link DB */
      return 0;
    case se_insteonCode_groupEvent: /* group event */
      return 4;
    case 0x57: /* link data */
      return 8;
    case 0x60: /* get version response */
      return 7;
    case 0x61: /* send group response */
      return 4;
    case 0x62: /* send message response */
      return 7;
    case 0x64: /* start linking response */
      return 3;
    case 0x65: /* cancel linking response */
      return 1;
    case 0x66: /* set device response */
      return 4;
    case 0x67: /* reset response */
      return 1;
    case 0x68: /* ack response */
      return 1;
    case se_insteonCode_getFirstLink: 
      return 1;
    case se_insteonCode_getNextLink: 
      return 1;
    case 0x6b: /* set config response */
      return 2;
    case 0x6c: /* get link data response */
      return 1;
    case 0x6d: /* LED on response */
      return 1;
    case 0x6e: /* LED off response */
      return 1;
    case 0x6f: /* manage link response */
      return 10;
    case 0x70: /* NAK response */
      return 2;
    case 0x71: /* long ACK response */
      return 3;
    case 0x72: /* sleep response */
      return 1;
    case 0x73: /* get configuration response */
      return 4;
    default: /* unknown response */
      return -1;
  };
}

static sc_status readResponse(sc_insteonBridge* pib, s_uint8* buff, s_int32* plen)
{
  sc_status status = sc_status_success;
  int len=0, br=0;

  /* every possible response has at least 2 bytes */
  status = sc_serialPort_readAll(&pib->port_, buff, 2);
  sm_status_returnIfFailed(status);

  if(buff[0] != 0x02)
    sm_status_returnError(status, sc_status_failure);

  len = getResponseLength(buff[1]);
  if(len < 0)
    sm_status_returnError(status, sc_status_failure);
  if(len > 0) {
    status = sc_serialPort_readAll(&pib->port_, buff+2, len);
    sm_status_returnIfFailed(status);
  }
  
  *plen = len + 2;
  
  sm_status_return(status);
}


sc_status sc_insteonBridge_getVersion(sc_insteonBridge* pib,
                                      sc_insteonBridgeVersion* pbv)
{ /* Gets bridge version information */
  sc_status status = sc_status_success;
  static const s_uint8 cmd[2] = { 0x02, 0x60 };
  s_uint8 res[24] = { 0 };
  s_int32 bw=0, br=0, done=0;

  status = sc_serialPort_write(&pib->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(!done) {
    status = readResponse(pib, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != 0x60) {
      status = sc_insteonBridge_handleAsync(pib, res);
      sm_status_returnIfFailed(status);      
    } else {
      pbv->address_         = atou32(res[2],res[3],res[4]);
      pbv->category_        = res[5];
      pbv->subcategory_     = res[6];
      pbv->firmwareVersion_ = res[7];
      if(res[8] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }
  }
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_resetLinks(sc_insteonBridge* pb)
{
  sc_status status = sc_status_success;
  static const s_uint8 cmd[2] = { 0x02, 0x67 };
  s_uint8 res[24];
  int br=0, bw=0, done=0;

  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(!done) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != 0x67) {
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);
    } else {
      if(res[2] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }
  }
  
  sm_status_return(status);
}

static sc_status nullLinkEntry(sc_insteonLinkEntry* ple)
{ /* initializes a NULL insteonLinkEntry */
  sc_status status = sc_status_success;

  ple->address_ = 0;
  ple->category_ = 0;
  ple->subcategory_ = 0;
  ple->group_ = 0;
  ple->flags_ = 0;
  ple->isEOL  = 0xFF;
  ple->firmwareVersion_ = 0;
  ple->reserved_[0] = 0;
  ple->reserved_[1] = 0;
  
  sm_status_return(status);  
}

static sc_status parseLinkEntry(sc_insteonLinkEntry* ple, const s_uint8* buff)
{ /* parses insteonLinkEntry struct from data in buff */
  sc_status status = sc_status_success;

  ple->flags_           = buff[2];
  ple->group_           = buff[3];
  ple->address_         = atou32(buff[4],buff[5],buff[6]);
  ple->category_        = buff[7];
  ple->subcategory_     = buff[8];
  ple->firmwareVersion_ = buff[9];
  ple->isEOL            = 0x00;
  ple->reserved_[0]     = 0;
  ple->reserved_[1]     = 0;  
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_getFirstLinkEntry(sc_insteonBridge* pib,
                                             sc_insteonLinkEntry* ple)
{
  sc_status status = sc_status_success;
  static const s_uint8 cmd[2] = { 0x02, se_insteonCode_getFirstLink };
  s_uint8 res[24];
  int br=0, bw=0, state=0;

   /* write out first link command. */
  status = sc_serialPort_write(&pib->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(state != 2) {
    status = readResponse(pib, res, &br);
    sm_status_returnIfFailed(status);

    if(state == 0) {
      if(res[1] == se_insteonCode_getFirstLink) {
        if(res[2] != 0x06) {
          /* empty list ... this is not an error */
          status = nullLinkEntry(ple);
          sm_status_return(status);
        }
        state = 1;
        continue;
      }
    } else if (state == 1) {
      if(res[1] == se_insteonCode_linkData) {
        status = parseLinkEntry(ple, res);
        sm_status_returnIfFailed(status);        
        state = 2;
        continue;
      }
    }
    
    /* wasn't a response for me ... queue it off or drop it */
    status = sc_insteonBridge_handleAsync(pib, res);
    sm_status_returnIfFailed(status);    
  }
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_getNextLink(sc_insteonBridge* pb,
                                       sc_insteonLinkEntry* ple)
{
  sc_status status = sc_status_success;
  static const s_uint8 cmd[2] = { 0x02, se_insteonCode_getNextLink };
  s_uint8 res[24];
  int br=0, bw=0, state=0;

   /* write out next link command. */
  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(state != 2) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(state == 0) {
      if(res[1] == se_insteonCode_getNextLink) {
        if(res[2] != 0x06) {
          /* end of list */
          status = nullLinkEntry(ple);
          sm_status_return(status);
        }
        state = 1;
        continue;
      }
    } else if (state == 1) {
      if(res[1] == se_insteonCode_linkData) {
        status = parseLinkEntry(ple, res);
        sm_status_returnIfFailed(status);        
        state = 2;
        continue;
      }
    }

    /* wasn't a response for me ... queue it off or drop it */
    status = sc_insteonBridge_handleAsync(pb, res);
    sm_status_returnIfFailed(status);    
  }
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_getConfiguration(sc_insteonBridge* pb,
                                            s_uint8* pcfg)
{
  sc_status status = sc_status_success;
  static const s_uint8 cmd[2] = { 0x02, 0x73 };
  s_uint8 res[24];
  int br=0, bw=0, done=0;

  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(!done) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != 0x73) {
      /* not a response to my message. */
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);      
    } else {
      *pcfg = res[2];
      if(res[5] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }    
  }  
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_setConfiguration(sc_insteonBridge* pb, s_uint8 cfg)
{
  sc_status status = sc_status_success;
  /* unsupported */  
  sm_status_return(status);
}

sc_status sc_insteonBridge_startLinking(sc_insteonBridge* pb)
{
  sc_status status = sc_status_success;
  static const s_uint8 cmd[4] = { 0x02, 0x64, 0x03, 0x01 };
  s_uint8 res[24];
  int br=0, bw=0, done=0;

  /* write out startLinking command */
  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(!done) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != 0x64) {
      /* not a response to my message. */
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);      
    } else {
      if(res[4] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }    
  }  

  sm_status_return(status);
}

sc_status sc_insteonBridge_cancelLinking(sc_insteonBridge* pb)
{
  sc_status status = sc_status_success;
  static const s_uint8 cmd[2] = { 0x02, 0x65 };
  s_uint8 res[24];
  int br=0, bw=0, done=0;

  /* write out cancelLinking command */
  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(!done) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != 0x65) {
      /* not a response to my message */
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);      
    } else {
      if(res[2] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }    
  }  

  sm_status_return(status);
}

sc_status sc_insteonBridge_sendGroupMessage(sc_insteonBridge* pb,
                                          s_uint8 groupId,
                                          s_uint8 cmd1, s_uint8 cmd2)
{
  sc_status status = sc_status_success;
  s_uint8 cmd[5] = { 0x02, 0x61 };
  s_uint8 res[24] = { 0 };
  s_int32 bw=0, br=0, done=0;

  cmd[2] = groupId;
  cmd[3] = cmd1;
  cmd[4] = cmd2;
  
  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);
  
  while(!done) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != 0x61) {
      /* not a response to my message */
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);      
    } else {
      if(res[5] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }    
  }  
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_sendStdMessage(sc_insteonBridge* pb,
                                          s_uint32 toAddress,
                                          s_uint8 cmd1, s_uint8 cmd2)
{
  sc_status status = sc_status_success;
  s_uint8 cmd[8] = { 0x02, se_insteonCode_sendMessage };
  s_uint8 res[24] = { 0 };
  s_int32 bw=0, br=0, done=0;

  /* standard direct message with 3 possible hops */
  cmd[2] = (s_uint8)((toAddress >> 16) & 0xFF);
  cmd[3] = (s_uint8)((toAddress >>  8) & 0xFF);
  cmd[4] = (s_uint8)(toAddress & 0xFF);
  cmd[5] = MF_HOPSLEFT(3) | MF_MAXHOPS(3);
  cmd[6] = cmd1;
  cmd[7] = cmd2;

  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(!done) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != se_insteonCode_sendMessage) {
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);
    } else {
      if(res[8] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }
  }
  sm_status_return(status);  
}

sc_status sc_insteonBridge_sendExtMessage(sc_insteonBridge* pb,
                                          s_uint32 toAddress,
                                          s_uint8 cmd1, s_uint8 cmd2,
                                          s_uint8* data, s_int32 len)
{
  sc_status status = sc_status_success;
  s_uint8 cmd[22] = { 0x02, se_insteonCode_sendMessage };
  s_uint8 res[24] = { 0 };
  s_int32 bw=0, br=0, done=0, idx;

  /* standard direct message with 3 possible hops */
  cmd[2] = (s_uint8)((toAddress >> 16) & 0xFF);
  cmd[3] = (s_uint8)((toAddress >>  8) & 0xFF);
  cmd[4] = (s_uint8)(toAddress & 0xFF);
  cmd[5] = MF_EXTM | MF_HOPSLEFT(3) | MF_MAXHOPS(3);
  cmd[6] = cmd1;
  cmd[7] = cmd2;
  
  for(idx=0; idx < len; ++idx) 
    cmd[8+idx] = data[idx];

  status = sc_serialPort_write(&pb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  while(!done) {
    status = readResponse(pb, res, &br);
    sm_status_returnIfFailed(status);

    if(res[1] != 0x62) {
      /* not a response to my message */
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);
    } else {
      if(res[22] != 0x06)
        sm_status_returnError(status, sc_status_failure);
      done = 1;
    }
  }
  sm_status_return(status);  
}

sc_status sc_insteonBridge_serviceDevice(sc_insteonBridge* pb,
                                         sc_automationPacket* pkt)
{ /* device socket is readable */
  sc_status status = sc_status_success;
  char res[32] = {0};
  int  rlen = sizeof(res);
  
  /* This is for async events */  
  status = readResponse(pb, res, &rlen);
  sm_status_returnIfFailed(status);
  /* figure out what type of event, and queue it off */
  status = sc_insteonBridge_handleAsync(pb, res);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_waitAcq(sc_insteonBridge* pb,
                                   s_uint32 addr,
                                   sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;
  char res[32] = {0};
  int  rlen = sizeof(res), done = 0;
  s_uint32 from, to, flags;

  while(!done) {
    
    status = readResponse(pb, res, &rlen);
    sm_status_returnIfFailed(status);

    from  = atou32(res[2],res[3],res[4]);
    to    = atou32(res[5],res[6],res[7]);
    flags = (res[8] >> 5) & 3;
    
    done = (flags == 1) && (from == addr);
    if(!done) {
      /* figure out what type of event, and queue it off */
      status = sc_insteonBridge_handleAsync(pb, res);
      sm_status_returnIfFailed(status);
    }    
  }

  status = sc_automationPacket_pushUInt8 (pkt, res[10]); /* cmd2 */
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt8 (pkt, res[9]);  /* cmd1 */
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt8 (pkt, res[8]);  /* flags */
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt32(pkt, to);     /* to */
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt32(pkt, from);   /* from */
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt8 (pkt, se_insteonCode_stdInEvent);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt8 (pkt, 0x02);
  sm_status_returnIfFailed(status);      
  status = sc_automationPacket_pushUInt16(pkt, 0x0001);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_discover(sc_insteonBridge* pb,
                                    sc_automationBridgeConnection* pc,
                                    sc_automationResponse* res)
{
  sc_status status = sc_status_success;
  sc_insteonLinkEntry le;
  sc_automationPacketPtrh pptrh;
  sc_automationPacket* pkt;

  status = sc_automationPacketPtrh_construct(&pptrh, NULL);
  sm_status_returnIfFailed(status);
  
  status = sc_automationBridgeBase_newPacket(&pb->base_, &pptrh);
  sm_status_returnIfFailed(status);
  pkt = sc_automationPacketPtrh_get(&pptrh);

  /* read out & serialize each link entry back as we go */
  status = sc_insteonBridge_getFirstLinkEntry(pb, &le);
  sm_status_returnIfFailed(status);

  while(!le.isEOL) {
    status = sc_automationPacket_pushUInt32(pkt, le.address_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_pushUInt8 (pkt, le.flags_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_pushUInt8 (pkt, le.group_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_pushUInt8 (pkt, le.category_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_pushUInt8 (pkt, le.subcategory_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_pushUInt8 (pkt, le.firmwareVersion_);
    sm_status_returnIfFailed(status);    
    status = sc_automationPacket_pushUInt16(pkt, 0x0002);
    sm_status_returnIfFailed(status);      

    status = sc_automationPacket_serializeToFd(pkt, pc->fd_);
    sm_status_returnIfFailed(status);

    status = sc_automationBridgeBase_newPacket(&pb->base_, &pptrh);
    sm_status_returnIfFailed(status);
    pkt = sc_automationPacketPtrh_get(&pptrh);
    
    status = sc_insteonBridge_getNextLink(pb, &le);
    sm_status_returnIfFailed(status);
  }

  status = sc_automationPacket_pushUInt16(pkt, 0x0000);
  sm_status_returnIfFailed(status);
  
  status = sc_automationPacket_serializeToFd(pkt, pc->fd_);
  sm_status_returnIfFailed(status);

  status = sc_automationPacketPtrh_destruct(&pptrh);
  sm_status_returnIfFailed(status);  

  sm_status_return(status);  
}

sc_status sc_insteonBridge_processRequest(sc_insteonBridge* pb,
                                          sc_automationBridgeConnection* pc,
                                          sc_automationRequest*  req,
                                          sc_automationResponse* res)
{
  sc_status status = sc_status_success;
  s_uint32 to;
  s_uint16 rq;
  s_uint8  c0, c1, gr, ext[14];
  sc_insteonBridgeVersion vs;

  /* this is always called with the bridge mutex held */  
  status = sc_automationPacket_popUInt16(req, &rq);
  sm_status_returnIfFailed(status);

  switch(rq) {
    case RC_DISCOVER:
      status = sc_insteonBridge_discover(pb, pc, res);
      sm_status_returnIfFailed(status);
      break;
    case RC_STDMESSAGE:
      /* read out the address, & cmd bytes */
      status = sc_automationPacket_popUInt32(req, &to);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c0);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c1);
      sm_status_returnIfFailed(status);
      status = sc_insteonBridge_sendStdMessage(pb, to, c0, c1);
      sm_status_returnIfFailed(status);           
      status = sc_automationPacket_pushUInt16(res, 0);
      sm_status_returnIfFailed(status);
      break;
    case RC_EXTMESSAGE:
      /* read out the address, cmd bytes, & ext data  */
      status = sc_automationPacket_popUInt32(req, &to);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c0);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c1);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popBinary(req, ext, 14);
      sm_status_returnIfFailed(status);
      status = sc_insteonBridge_sendExtMessage(pb, to, c0, c1, ext, 14);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt16(res, 0);
      sm_status_returnIfFailed(status);            
      break;
    case RC_GRPMESSAGE:
      /* read out the group, & cmd bytes */
      status = sc_automationPacket_popUInt8(req, &gr);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c0);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c1);
      sm_status_returnIfFailed(status);      
      status = sc_insteonBridge_sendGroupMessage(pb, gr, c0, c1);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt16(res, 0);
      sm_status_returnIfFailed(status);      
      break;
    case RC_GETVERSION:
      /* do a get version and return ver struct */
      status = sc_insteonBridge_getVersion(pb, &vs);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(res, vs.reserved_);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(res, vs.firmwareVersion_);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(res, vs.subcategory_);      
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(res, vs.category_);
      sm_status_returnIfFailed(status);            
      status = sc_automationPacket_pushUInt32(res, vs.address_);
      sm_status_returnIfFailed(status);      
      status = sc_automationPacket_pushUInt16(res, 0);
      sm_status_returnIfFailed(status);      
      break;
    case RC_STDMSGWACK:
      /* read out the address, & cmd bytes */
      status = sc_automationPacket_popUInt32(req, &to);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c0);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_popUInt8(req, &c1);
      sm_status_returnIfFailed(status);
      status = sc_insteonBridge_sendStdMessage(pb, to, c0, c1);
      sm_status_returnIfFailed(status);
      status = sc_insteonBridge_waitAcq(pb, to, res);
      sm_status_returnIfFailed(status);      
      break;
    case RC_RESETLINKS:
      /* reset link database on bridge controller */
      status = sc_insteonBridge_resetLinks(pb);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt16(res, 0);
      sm_status_returnIfFailed(status);
    default:
      break;
  }
    
  sm_status_return(status);  
}


static sc_status sc_insteonBridge_handleAsync(sc_insteonBridge* pb, const s_uint8* res)
{
  sc_status status = sc_status_success;
  s_uint32 from=0, to=0;
  sc_automationPacketPtrh pptrh;
  sc_automationPacket* pkt;

  status = sc_automationPacketPtrh_construct(&pptrh, NULL);
  sm_status_returnIfFailed(status);

  status = sc_automationBridgeBase_newPacket(&pb->base_, &pptrh);
  sm_status_returnIfFailed(status);
  pkt = sc_automationPacketPtrh_get(&pptrh);
  
  /* translate res out */
  switch(res[1]) {
    case se_insteonCode_stdInEvent: /* standard insteon message */
      from = atou32(res[2],res[3],res[4]);
      to   = atou32(res[5],res[6],res[7]);
      status = sc_automationPacket_pushUInt8(pkt, res[10]); /* cmd2 */
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(pkt, res[9]);  /* cmd1 */
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(pkt, res[8]);  /* flags */
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt32(pkt, to);     /* to */
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt32(pkt, from);   /* from */
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(pkt, se_insteonCode_stdInEvent);
      sm_status_returnIfFailed(status);
      break;
    case se_insteonCode_extInEvent: /* extended insteon message */
      status = sc_automationPacket_pushUInt8(pkt, se_insteonCode_extInEvent);
      sm_status_returnIfFailed(status);
      break;
    case se_insteonCode_x10InEvent: /* not supported */
      status = sc_automationPacket_pushUInt8(pkt, se_insteonCode_x10InEvent);
      sm_status_returnIfFailed(status);
      break; 
    case se_insteonCode_linkStatusEvent: 
      status = sc_automationPacket_pushUInt8(pkt, se_insteonCode_linkStatusEvent);
      sm_status_returnIfFailed(status);
      break;
    case se_insteonCode_buttonEvent:
      status = sc_automationPacket_pushUInt8(pkt, res[2]);
      sm_status_returnIfFailed(status);
      status = sc_automationPacket_pushUInt8(pkt, se_insteonCode_buttonEvent);
      sm_status_returnIfFailed(status);
      break;
    case se_insteonCode_resetLinksEvent: /* user reset link DB */
      status = sc_automationPacket_pushUInt8(pkt, se_insteonCode_resetLinksEvent);
      sm_status_returnIfFailed(status);
      break;
    case se_insteonCode_groupEvent: /* group event */
      status = sc_automationPacket_pushUInt8(pkt, se_insteonCode_groupEvent);
      sm_status_returnIfFailed(status);
      break;
    default:
      sm_status_return(status);
  }

  status = sc_automationPacket_pushUInt8(pkt, 0x02);
  sm_status_returnIfFailed(status);      
  status = sc_automationPacket_pushUInt16(pkt, 0x0001); /* async+success */
  sm_status_returnIfFailed(status);

  /* message was out of order.  Put it on OOO queue, and return direct
   * reponse first, then OOO messages. */
  status = sc_automationBridgeBase_queueAsync(&pb->base_, &pptrh);
  sm_status_returnIfFailed(status);

  status = sc_automationPacketPtrh_destruct(&pptrh);
  sm_status_returnIfFailed(status);  
  
  sm_status_return(status);
}
