#!/usr/bin/perl

use Getopt::Long;

my $cmd = $0;
my $usage = "\nusage: $cmd --update <deviceName> --add <networkType> --device=<devicePath> [--args=startArgs]";
$usage .= "\n  deviceName  : the name of an existing bridge device.";
$usage .= "\n  networkType : the network type, e.g. 'Insteon' or 'ZigBee'";
$usage .= "\n  devicePath  : the path to the device file, e.g. /dev/ttyUSB0";
$usage .= "\n  startArgs   : startup args for the bridge, e.g. \"panId=0xcafe;\"\n";

# process arguments
my $ident = '';
my $ntype = '';
my $dpath = '';
my $bargs = '';
my $altr  = '';
my $next  = '';

my @networks = qw(Insteon ZigBee);

$result = GetOptions("update:s"  => \$ident,
                     "add:s"     => \$ntype,
                     "device=s"  => \$dpath,
                     "args:s"    => \$bargs,
                     "next"      => \$next,
                     "altroot:s" => \$altr);
if(!$result) {
  die $usage;
}
if($ident && $ntype) {
  die "cannot specify both update and add in the same invocation.\n";  
}
if( !($ident || $ntype) ) {
  die "must specify either --update or --add actions.\n";
}
if($ident) {
  unless ($ident =~ /^[a-zA-Z0-9_]+$/) {
    die "'$ident' is not a valid identifier.";
  }
}
if($ntype) {
  my $found = '';
  foreach $net (@networks) {
    if($ntype =~ /^$net$/i) {
      $found = $net;
      last;
    }
  }
  die "$ntype is not a supported network type. Must be one of: @networks." if(!$found);
  $ntype = $found;
}


# fix bug with name stickyness

sub checkDevicePath {
  if( ! -e $dpath) {
    print "warning: $dpath does not currently exist.\n";
  } else {
    # file exists, check a few other items
    if( ! (-r $dpath)) {
      print "warning: $dpath is not readable by the current user.\nYou may need to be added to the 'dialout' group.";
    }
    if( ! (-w $dpath)) {
      print "warning: $dpath is not writable by the current user.\nYou may need to be added to the 'dialout' group";
    }
    if( ! ( (-p $dpath) || (-c $dpath) ) ) {
      print "warning: $dpath does not look like a device.\n";
    }
  }
}


sub findSqlite3 {
  my $sql3 = "$altr/var/lib/lase/bin/sqlite3";
  if( ! -e $sql3) {
    $sql3 = "../../../ext/bin/sqlite3";
  }
  if( ! -e $sql3) {
    $sql3 = "../../../var/lib/lase/bin/sqlite3";
  }
  if( ! -e $sql3) {
    $sql3 = "./ext/bin/sqlite3";
  }
  if( ! -e $sql3) {
    die "cannot find sqlite3";
  }
  return $sql3;
}

my $sqlite3 = findSqlite3();
my $db = "$altr/var/lib/lase/db/lase.db";

sub deviceIdFromIdent {
  my $did='', $id = shift @_;
  chomp($did = `$sqlite3 $db "select id from devices where identifier='$id';"`);
  return $did;
}
sub deviceIdFromAddress {
  my $did='', $addr = shift @_;
  chomp($did = `$sqlite3 $db "select id from devices where address='$addr';"`);
  return $did;
}
sub bridgeIdFromDeviceId {
  my $bid='', $did = shift @_;
  chomp($bid = `$sqlite3 $db "select id from bridges where deviceId='$did';"`);
  return $bid;
}
sub getUnusedDefaultIdentifier {
  my @lines = (), $classname = lc(shift @_);
  chomp(@lines=`$sqlite3 $db "select identifier from devices where identifier LIKE '$classname%';"`);
  my $max = -1;  
  foreach my $line (@lines) {
    if($line =~ /$classname([\d]+)/i) {
      $max = $1 if($1 > $max);      
    }
  }
  return $classname . ($max+1);
}

if($ident) {
  my $did = deviceIdFromIdent($ident);
  if($did) {
    my $bid = bridgeIdFromDeviceId($did);
    die "Error: device '$ident' is not a bridge. Try re-adding." if(!$bid);
    if($dpath) {
      checkDevicePath();
      if ($bargs) {
        $did = `$sqlite3 $db "update bridges set deviceId=$did, devPath='$dpath', args='$bargs' where id=$bid;"`;
      } else {
        $did = `$sqlite3 $db "update bridges set deviceId=$did, devPath='$dpath' where id=$bid;"`;
      }      
    } else {
      $did = `$sqlite3 $db "update bridges set deviceId=$did, args='$bargs' where id=$bid;"`;
    }
    print("Bridge $ident updated.\n");
  } else {
    die "Cannot find device $ident to update.";
  }
}

if ($ntype) {
  checkDevicePath();
  my $brgtool = "$altr/var/lib/lase/bin/brgdetect", @lines = ();
  chomp(@lines = `$brgtool $ntype $dpath`);
  foreach $line (@lines) {
    if($line =~ /^>>\s*([^,]+),\s*([^\s]+)\s*$/) {
      $addr = $1;
      $dtyp = $2;
      last;
    }
  }

  print("Found $ntype bridge: type=$dtyp, address=$addr\n");

  # first check that this bridge is not already present in devices
  my $did = deviceIdFromAddress($addr);
  if(!$did) {
    # doesn't exist, give it a default identifier
    my $defid = getUnusedDefaultIdentifier($ntype);
    $did = `$sqlite3 $db "insert into devices (address,type,identifier) values ('$addr','$dtyp','$defid');"`;
    $did = deviceIdFromAddress($addr);
    print("Bridge $defid added.\n");
  }
  my $bid = bridgeIdFromDeviceId($did);
  if(!$bid) {
    if(!$bargs) {
      $did = `$sqlite3 $db "insert into bridges (deviceId,devPath) values ($did,'$dpath');"`;
    } else {
      $did = `$sqlite3 $db "insert into bridges (deviceId,devPath,args) values ($did,'$dpath','$bargs');"`;
    }
  } else {
    die "Bridge already installed.  Use --update to modify its configuration.\n";
  }
}
