

/*
 * Function that returns a new object with p as its prototype.
 * Ref: 'JavaScript The definitive guide' (Flanagan).
 */
function heir(p) {
  function h() {}
  h.prototype = p;
  return new h();
}
function show(o) {
  for(var k in o) {
    print(k);
  }
}

/*
 * Base class for all device types.  Provides a means to identify
 * devices, and for derived devices to reach their bridge. 
 */
function Device(dname, daddr, dtype, ntype, brg) {
  
  var vname        = dname;  
  var vaddress     = daddr;
  var vdeviceType  = dtype;
  var vnetworkType = ntype;
  var vbridge      = brg;
  var vevents      = new Object();

  this.__defineGetter__("name",        function() { return vname; });
  this.__defineGetter__("address",     function() { return vaddress; });
  this.__defineGetter__("networkType", function() { return vnetworkType; });
  this.__defineGetter__("deviceType",  function() { return vdeviceType; });
  this.__defineGetter__("bridge",      function() { return vbridge; });
  this.__defineGetter__("events",      function() { return vevents; });
  this._rename = function(newName) { vname = newName; }
}
Device.classname = "Device";

/* allows derived devices to fire an event to listeners */
Device.prototype._fireEvent = function(evt)
{
  var event = (typeof(evt) == 'string') ? this.events[evt] : evt;
  var args = [];
  for(var idx=1; idx < arguments.length; ++idx)
    args.push(arguments[idx]);
  for(var li in event.listeners) {
    var el = event.listeners[li];
    if(el instanceof EventListener) {
      el.invoke(args);
    }
  }
}

/*
 * Properties listed here will not be persisted across runs of the
 * framework (typically because their values are constants or determined
 * automatically).
 */
Device.nonserialized = ["name", "address", "networkType", "deviceType",
                        "bridge", "events" ];
Device.nondescriptive = [ "bridge", "events" ];

/* General enum providing a common On/Off values in all devices. */
Device.OnOffEnum = { On: 1, Off: 0, Unknown: -1 };

/*
 * Provides basic object persistence to an archive. This function is an 
 * adaptation for stored persistence across multiple runs, as opposed to 
 * for communications purposes, where a general description of the object is
 * desired.
 */
Device.prototype._toArchive = function() {
  function getNonserialized(o) {
    var ns = [];
    var ctor = o.constructor;
    while(ctor) {
      if('nonserialized' in ctor) 
        ns = ns.concat(ctor.nonserialized);      
      ctor = ctor.superclass;
    }
    return ns;
  }  
  var nonserialized = getNonserialized(this);
  
  var archive = {};
  var __funcs = { count: 0 };
  for(var k in this) {
    if(nonserialized.indexOf(k) == -1) {
      var val = this[k];
      if(typeof(val) == 'function') {
        if('__eventTarget' in val) {
          var fs = val.toString().replace(/\n/g, ' ');
          __funcs[k] = fs.replace(/\s\s+/g,' ');
          __funcs.count++;
        }
      } else {
        archive[k] = val;
      }
    }
  }
  if(__funcs.count > 0) {
    delete __funcs.count; // don't need this in archive
    archive.__funcs = __funcs;
  }
  return archive;
}

Device.prototype._load = function(archive) {
  for(var k in archive) {
    if(k == "__funcs") {
      var funcs = archive[k];      
      for(var f in funcs) {
        this[f] = eval('(' + funcs[f] + ')');
      }
    } else {
      this[k] = archive[k];
    }
  }
}
Device.prototype._save = function() {
  var arch = null;
  try {
    this.toJSON = this._toArchive;
    arch = JSON.stringify(this);
  } catch(e) {
    this.toJSON = null;
    throw e;
  }
  this.toJSON = null;
  return arch;
}
/*
 * Provides a general object description suitable for returning as a
 * result. Used where a complete description of an object is needed,
 * but the functionality of the original is not (i.e., this is not an
 * RPC proxy).
 */
Device.prototype._toDescriptor = function() {
  function getNondescriptive(o) {
    var nd = [];
    var ctor = o.constructor;
    while(ctor) {
      if('nondescriptive' in ctor) 
        nd = nd.concat(ctor.nondescriptive);
      ctor = ctor.superclass;
    }
    return nd;
  }  
  var nondescriptive = getNondescriptive(this);

  var descriptor = {};
  for(var k in this) {
    if(nondescriptive.indexOf(k) == -1) {
      var val = this[k];
      descriptor[k] = val;
    }
  }
  return descriptor;
}
Device.prototype._describe = function() {
  var desc = null;
  try {
    this.toJSON = this._toDescriptor;
    desc = JSON.stringify(this);
  } catch(e) {
    this.toJSON = null;
    throw e;
  }
  this.toJSON = null;
  return desc;  
}

/*
 * Misc. classes supporting events.  
 */
function EventListener(targetObject, memberFunc) {
  this.invoke = function(args) { memberFunc.apply(targetObject, args); }
  this.__defineGetter__("target", function() { return targetObject; });
  this.__defineGetter__("method", function() { return memberFunc; });
}
EventListener.classname="EventListener";

function EventContainer() {
  this.listeners = [];
}
EventContainer.classname = "EventContainer";

/*
 * Subscribes object 'target' to the event.  When the event occurs,
 * target.method(event); will be invoked.
 */
EventContainer.prototype.subscribe = function(target, method) {
  this.listeners.push(new EventListener(target,method));
  method.__eventTarget = true;
}

/*
 * Unsubscribes object 'target' from the event. 'target' must be an
 * object reference.
 */
EventContainer.prototype.unsubscribe = function(target) {
  var idx = 0;
  for(var l in this.listeners) {
    if(l.target === target) 
      break;
    ++idx;
  }
  if(idx < this.listeners.length) 
    this.listeners = this.listeners.slice(idx,idx+1);
}

/*
 * A simple change event structure that is suitable for a wide array of
 * events.
 */
function ChangeEvent(srcObj,oldVal,newVal) {
  this.__defineGetter__("sourceObject", function() { return srcObj; });
  this.__defineGetter__("oldValue",     function() { return oldVal; });
  this.__defineGetter__("newValue",     function() { return newVal; });  
}
ChangeEvent.classname = "ChangeEvent";


/*
 * Base class for Insteon devices.
 */
function InsteonDevice(dname, daddr, dtype, ntype, brg) {
  arguments.callee.superclass.call(this, dname, daddr, dtype, ntype, brg);
}
InsteonDevice.prototype   = heir(Device.prototype);
InsteonDevice.prototype.constructor = InsteonDevice;
InsteonDevice.classname   = "InsteonDevice";
InsteonDevice.superclass  = Device;

InsteonDevice.prototype._decodeStdInsteonMsg = function(pkt) {
  var data, code, pos;
  pos = pkt.tell();
  data = pkt.popUInt16();
  data = pkt.popUInt8();
  code = pkt.popUInt8();
  if(code == 0x50) {
    /* std insteon event */
    data = pkt.popUInt32();
    data = pkt.popUInt32();
    data = pkt.popUInt8();
    var c1 = pkt.popUInt8();
    var c2 = pkt.popUInt8();
    return { cmd1: c1, cmd2: c2 };
  } else {
    pkt.seek(pos);
    return null;
  }
}

/*
 *  Support for Insteon IoLinc device.  This device is an actuator that
 *  has a general purpose relay switching 5V loads.  It has some other
 *  features (an analog sensing input) which are not supported here. 
 */
function IoLinc(dname, daddr, dtype, ntype, brg) {
  arguments.callee.superclass.call(this, dname, daddr, dtype, ntype, brg);
  this.events.onChange = new EventContainer();
  var relayState = Device.OnOffEnum.Unknown;
  this.__defineGetter__("state", function() {
    if(relayState == Device.OnOffEnum.Unknown)
      this._getRelayState();
    return relayState; });
  this.__defineSetter__("state", function(v) { this._setRelayState(v); });
  this._privSetRelay = function(v) {
    if(v != relayState) {
      var oldval = relayState;
      relayState = v;
      if(oldval != Device.OnOffEnum.Unknown)
        this._fireEvent(this.events.onChange, new ChangeEvent(this, oldval, relayState));
    }
  }
  this._getRelayState();
}
IoLinc.prototype   = heir(InsteonDevice.prototype);
IoLinc.prototype.constructor = IoLinc;
IoLinc.classname   = "IoLinc";
IoLinc.superclass  = InsteonDevice;
IoLinc.nonserialized = [ "state" ];

IoLinc.prototype._setRelayState = function(v) {
  if(v == Device.OnOffEnum.Off)
    this.bridge.sendStdMessage(this.address, 0x13, 0x01);
  else
    this.bridge.sendStdMessage(this.address, 0x11, 0x01);
}
IoLinc.prototype._getRelayState = function() {
  var pkt = this.bridge.sendStdWaitAck(this.address, 0x19, 0x00);
  this._relayMessage(pkt);
}

/*
 * Message relayed in pkt from the network.  This message will likely
 * need to be re-issued in the JavaScript object framework, so user
 * defined behaviors can be realized.
 */
IoLinc.prototype._relayMessage = function(pkt) {
  // see if the packet contains a std insteon 'on' or 'off' message
  var msg = this._decodeStdInsteonMsg(pkt);
  if(msg) {
    if(msg.cmd1 == 0x13)      // Off
      this._privSetRelay(Device.OnOffEnum.Off);
    else if(msg.cmd1 == 0x11) // On
      this._privSetRelay(Device.OnOffEnum.On);
    else if(msg.cmd1 == 0x0) {
      if(msg.cmd2 == 0)
        this._privSetRelay(Device.OnOffEnum.Off);
      else
        this._privSetRelay(Device.OnOffEnum.On);
    }
  }  
}


/*
 *  Support for Insteon LampLinc device.  This device is an actuator that
 *  can switch a resistive 300W load (such as a lamp) on or off. 
 */
function LampLinc(dname, daddr, dtype, ntype, brg) {
  arguments.callee.superclass.call(this, dname, daddr, dtype, ntype, brg);
  this.events.onChange = new EventContainer();
  var lampState = Device.OnOffEnum.Unknown;
  this.__defineGetter__("state", function() {
    if(lampState == Device.OnOffEnum.Unknown)
      this._getLampState();
    return lampState; });
  this.__defineSetter__("state", function(v) { this._setLampState(v); });
  this._privSetLamp = function(v) {
    if(v != lampState) {
      var oldval = lampState;
      lampState = v;
      if(oldval != Device.OnOffEnum.Unknown)
        this._fireEvent(this.events.onChange, new ChangeEvent(this, oldval, lampState));
    }
  }
  this._getLampState();
}
LampLinc.prototype   = heir(InsteonDevice.prototype);
LampLinc.prototype.constructor = LampLinc;
LampLinc.classname   = "LampLinc";
LampLinc.superclass  = InsteonDevice;
LampLinc.nonserialized = [ "state" ];

LampLinc.prototype._setLampState = function(v) {
  if(v == Device.OnOffEnum.Off)
    this.bridge.sendStdMessage(this.address, 0x13, 0x00);
  else
    this.bridge.sendStdMessage(this.address, 0x11, 0xFE);
}
LampLinc.prototype._getLampState = function() {
  var pkt = this.bridge.sendStdWaitAck(this.address, 0x19, 0x00);
  this._relayMessage(pkt);
}

/*
 * Message relayed in pkt from the network.  This message is re-issued.
 */
LampLinc.prototype._relayMessage = function(pkt) {
  // see if the packet contains a std insteon 'on' or 'off' message
  var msg = this._decodeStdInsteonMsg(pkt);
  if(msg) {
    if(msg.cmd1 == 0x13)      // Off
      this._privSetLamp(Device.OnOffEnum.Off);
    else if(msg.cmd1 == 0x11) // On
      this._privSetLamp(Device.OnOffEnum.On);
    else if(msg.cmd1 == 0x0) {
      if(msg.cmd2 == 0)
        this._privSetLamp(Device.OnOffEnum.Off);
      else
        this._privSetLamp(Device.OnOffEnum.On);
    }
  }  
}


/*
 * Support for Insteon Motion detectors.  This device is
 * Device.OnOffEnum.On when motion is sensed, and Off otherwise.  The
 * device has a quirk that unlike other insteon devices, it will *not*
 * respond to requests for it's current value, so at first boot of the
 * network, the state of the motion sensor is not actually known until
 * it changes.
 */
function MotionSensor(dname, daddr, dtype, ntype, brg)
{
  arguments.callee.superclass.call(this, dname, daddr, dtype, ntype, brg);
  // declare event 'onChange'
  this.events.onChange = new EventContainer();
  // add motion getter
  var motionState = Device.OnOffEnum.Unknown;
  this._getMotionState = function() {
    if(motionState == Device.OnOffEnum.Unknown) {
      this.bridge.sendStdMessage(this.address, 0x19, 0x00);
    }
    return motionState;
  };
  this._setMotionState = function(v) {
    if(v != motionState) {
      var oldval = motionState;
      motionState = v;
      this._fireEvent(this.events.onChange, new ChangeEvent(this, oldval, motionState));
    }
  };
}
MotionSensor.prototype   = heir(InsteonDevice.prototype);
MotionSensor.prototype.constructor = MotionSensor;
MotionSensor.classname   = "MotionSensor";
MotionSensor.superclass  = InsteonDevice;
MotionSensor.nonserialized = [ "state" ];

MotionSensor.prototype.__defineGetter__("state", function() { return this._getMotionState(); });

/*
 * Message relayed in pkt from the network.  This message will likely
 * need to be re-issued in the JavaScript object framework, so user
 * defined behaviors can be realized.
 */
MotionSensor.prototype._relayMessage = function(pkt) {
  // see if the packet contains a std insteon 'on' or 'off' message
  var msg = this._decodeStdInsteonMsg(pkt);
  if(msg) {
    if(msg.cmd1 == 0x13) // Off
      this._setMotionState(Device.OnOffEnum.Off);
    else if(msg.cmd1 == 0x11) // On
      this._setMotionState(Device.OnOffEnum.On);
    else if(msg.cmd1 == 0x19) {
      if(msg.cmd2 == 0)
        this._setMotionState(Device.OnOffEnum.Off);
      else
        this._setMotionState(Device.OnOffEnum.On);        
    }
  }
}

this.save = function() {
  function keyOfIn(member,obj) {
    for(var k in obj) {
      if(obj[k] === member)
        return k;
    }
    return null;
  }
  for(var k in this) {
    if(this[k] instanceof Device) {
      var str = this[k]._save();
      if(str.length > 2) {
        this._archiveDevice(k, str);
      }
      /* clear out any previously saved links */
      this._clearLinks(this[k]);
      /* save current links */
      for(var e in this[k].events) {
        var ec = this[k].events[e];
        if( (ec instanceof EventContainer) && (ec.listeners.length > 0)) {
          var srcIdent = k;
          var srcEvent = e;
          var lsnrs = ec.listeners;
          for(var i=0; i < lsnrs.length; ++i) {
            var lsnr = lsnrs[i];
            dstIdent = lsnr.target.name;
            dstMethd = keyOfIn(lsnr.method, lsnr.target);
            this._archiveLink(srcIdent,srcEvent,dstIdent,dstMethd);
          }
        }
      }
    }
  }
}

this.rename = function(obj, newName) {
  if(typeof(newName) != 'string') {
    throw new Error("new object name must be a string.");
  }
  if(this[newName] != null) {
    throw new Error("An object named '" + newName + "' already exists.");
  }
  if(newName.match(/^[a-zA-Z0-9_]+$/)) { /* newName is a valid identifier */
    this._rename(obj,newName);
    if('name' in obj)
      obj._rename(newName);
  } else throw new Error("'" + newName + "' is not a valid identifier.");
}
