#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <si_sys/defs.h>
#include <si_std/string.h>
#include <si_sha/lase.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>


sc_status sc_app_readInputLine(sc_string* pl, int fd, s_int32* pDone);

#define INVALID_SOCKET (-1)

/** Use and 8MB context.  Rarely need to garbage collect this way.
 */
#define JS_CONTEXT_SIZE (8*1024*1024)


/* could get fancier than this, but no time */
const char* getRootDir(const char* name) {
  if(strstr(name, "lucs-lase-tlase") != 0) /* name when built, not installed */
    return "."; /* run built version from $(ROOT_PROJECT)/src/lase */
  else return "/var/lib/lase"; /* installed location */
}

int main(int argc, char*argv[])
{
  sc_status status = sc_status_success;
  sc_lase lase;
  sc_scriptRequestPtrh req0, req1;
  sc_scriptRequest *preq0, *preq1;
  int quit=0, tmp=0, nfds, rv, outFd;
  int servSock = INVALID_SOCKET, cliSock = INVALID_SOCKET;
  struct sockaddr_un saddr;
  fd_set rfds;
  sc_string namePath;
  mode_t mode;
  const char* rootDir = getRootDir(argv[0]);

  status = sc_string_construct(&namePath);
  if(sm_status_failed(status)) {
    sc_string_destruct(&namePath);
    exit(99);
  }
  
  /* create the lase service socket */
  servSock = socket(AF_UNIX, SOCK_STREAM, 0);
  if(servSock == INVALID_SOCKET) {
    sc_string_destruct(&namePath);
    sm_status_returnError(status, sc_status_failure);
  }

  memset(&saddr, 0, sizeof(saddr));
  saddr.sun_family = AF_UNIX;

  status = sc_string_append(&namePath, rootDir);
  sm_status_returnIfFailed(status);
  status = sc_string_append(&namePath, "/lase.sock");
  sm_status_returnIfFailed(status);
  
  /* unlink any previous socket entry with my name*/
  unlink(namePath.str_);
  strncpy(saddr.sun_path, namePath.str_, sizeof(saddr.sun_path)-1);

  status = sc_string_destruct(&namePath);
  sm_status_returnIfFailed(status);

  /* name the service socket */  
  rv = bind(servSock, (struct sockaddr*)&saddr, sizeof(saddr));
  if(rv == -1) {
    close(servSock);
    fprintf(stderr, "failed to bind to address: %s\n", saddr.sun_path);
    exit(99);
  }
  /* put it in listen mode */
  rv = listen(servSock, 5);
  if(rv == -1) {
    close(servSock);
    fprintf(stderr, "failed to listen at bound address: %s\n", saddr.sun_path);
    exit(99);
  }

  /* open up permissions on this socket.  This isn't ideal, but for
   * now, I'm trying to make it easier for others to use the program */
  mode = S_IRUSR | S_IWUSR | S_IXUSR |
         S_IRGRP | S_IWGRP | S_IXGRP |
         S_IROTH | S_IWOTH | S_IXOTH;  
  chmod(saddr.sun_path, mode);

  /* okay now start lase environment */  
  status = sc_lase_construct(&lase, JS_CONTEXT_SIZE, rootDir);
  sm_status_returnIfFailed(status);

  status = sc_lase_start(&lase);
  sm_status_returnIfFailed(status);  

  status = sc_scriptRequestPtrh_construct(&req0, NULL);
  sm_status_returnIfFailed(status);

  status = sc_scriptRequestPtrh_construct(&req1, NULL);
  sm_status_returnIfFailed(status);

  /* allocate script requests for local use */
  status = sc_lase_allocRequest(&lase, &req0);
  sm_status_returnIfFailed(status);
  preq0 = sc_scriptRequestPtrh_get(&req0);
  
  status = sc_lase_allocRequest(&lase, &req1);
  sm_status_returnIfFailed(status);  
  preq1 = sc_scriptRequestPtrh_get(&req1);
  
  nfds = 1 + sm_max(STDIN_FILENO, servSock);

  while(!quit)
  {    
    FD_ZERO(&rfds);
    FD_SET(STDIN_FILENO, &rfds);
    FD_SET(servSock, &rfds);
    if(cliSock != INVALID_SOCKET)
      FD_SET(cliSock, &rfds);
    
    outFd = INVALID_SOCKET;
    
    rv = select(nfds, &rfds, NULL, NULL, NULL);
    if(rv > 0)
    { /* read & execute scripts from stdin or sock */
      if(FD_ISSET(STDIN_FILENO, &rfds)) {    /* stdin has data */
        status = sc_app_readInputLine(&preq0->scriptIn, STDIN_FILENO, &tmp);
        sm_status_returnIfFailed(status);
        outFd = STDOUT_FILENO;
      } else if(FD_ISSET(servSock, &rfds)) { /* new client connecting */
        if(cliSock) 
          close(cliSock);
        cliSock = accept(servSock, NULL, NULL);
        /* fix up nfds to include new client socket */
        if(cliSock != INVALID_SOCKET) 
          nfds = sm_max(nfds, 1+cliSock);
      } else if(FD_ISSET(cliSock, &rfds)) {  /* client socket has data */
        /* read script from external client */
        status = sc_app_readInputLine(&preq0->scriptIn, cliSock, &tmp);
        sm_status_returnIfFailed(status);
        outFd = cliSock;
      }
      quit = quit || (tmp < 0);

      if( (outFd >= 0) && (tmp > 0) )
      { /* complete script has been read */
        sc_string_trimRight(&preq0->scriptIn);
        
        if(strcmp(preq0->scriptIn.str_, "exit") == 0)
          break; /* don't attempt to execute 'exit', just do it. */
        status = sc_lase_executeRequest(&lase, &req0);
        sm_status_returnIfFailed(status);

        /* check for and report error */
        if(preq0->scriptOut.length_ > 0) {
          /* output script result on outFd */
          write(outFd, preq0->scriptOut.str_, preq0->scriptOut.length_);
          write(outFd, "\n", 1);
        }
        /* clear request */
        status = sc_scriptRequest_clear(preq0);
        sm_status_returnIfFailed(status);
      }
    } else break; /* signal */
  }

  close(servSock);
  if(cliSock != INVALID_SOCKET)
    close(cliSock);
  
  sc_lase_stop(&lase);
  sc_lase_destruct(&lase);  
  sc_scriptRequestPtrh_destruct(&req0);
  sc_scriptRequestPtrh_destruct(&req1);
  
  sm_status_return(status);
}


sc_status sc_app_readInputLine(sc_string* pl, int fd, s_int32* pDone)
{
  sc_status status = sc_status_success;
  int rb=1, done=0;
  char ch;

  while(rb) {
    rb = read(fd, &ch, 1);
    if(rb < 0) {
      done = -1; /* error */
      break;
    } else if(rb == 0) {
      done = 1;
      break;
    }
    switch(ch) {
      case 0x00:   /* null character */
      case 0x04:   /* ^D, EOF in unix */
      case 0x1a:   /* ^Z, EOF in windows */
        done = 1; 
        rb   = 0;
        break;
      case '\n':   /* end of line, exit loop but input not done */
        done = 1; 
        rb = 0;    /* note: fall through */
      default:
        sc_string_appendChars(pl, &ch, 1);
        break;
    }
  }

  if(pDone)
    *pDone = done; /* got a complete script line, as indicated by EOF, */
                   /* NULL character, or new line character*/

  sm_status_return(status);
}














#if 0
/*
 * Routines for reading multi-line scripts
 * This needed more attention to work well, set aside for now
 */
sc_status sc_lase_readInputScript(sc_lase* pl, int fd, s_int32* pEOF)
{
  sc_status status = sc_status_success;
  int rb=1, done=0;
  s_int32 eoi = 0;

  status = sc_string_clear(&pl->ln);
  sm_status_returnIfFailed(status);

  while(!eoi) {
    status = sc_lase_readInputLine(pl, fd, &eoi);
    sm_status_returnIfFailed(status);
  }

  status = sc_string_trimRight(&pl->ln);
  sm_status_returnIfFailed(status);

  if(pEOF)
    *pEOF = (eoi < 0) ? 1 : 0;

  sm_status_return(status);
}

sc_status sc_lase_loadScript(sc_lase* pl, const TCHAR* filename)
{
  sc_status status = sc_status_success;

  status = loadScript(pl->jscx, filename);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status sc_lase_executeInputScript(sc_lase* pl, s_int32* pExit)
{
  sc_status status = sc_status_success;
  const char* script = NULL;
  int lnum = 0;
  JSBool    rv;
  jsval     rval;

  script = pl->ln.str_;

  if(strcmp(script, "exit") == 0) {
    if(pExit) *pExit = 1;
    sm_status_return(status);
  }

  /* this should actually get queued off to lase thread */
  rv = JS_EvaluateScript(pl->jscx, pl->glob, script, strlen(script),
                         "script", lnum, &rval);
  if(!rv) {
    printf("JS_EvaluateScript returned false.\n");    
  } else {
    // printf("****************************************\n");
  }

  sm_status_return(status);
}
#endif
