#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <si_std/string.h>
#include <si_sha/adevmgr.h>
#include <si_sha/sreqpool.h>
#include <si_sha/lase.h>

#include <sqlite3.h>

sc_status sc_automationDeviceMgr_construct(sc_automationDeviceMgr* pom,
                                           sc_lase* lase)
{
  sc_status status = sc_status_success;
  int rc = 0;
  sc_string dbPath;
  char* errMsg = NULL;

  pom->jsEngine = lase;
  pom->db = NULL;

  /* build up database path from base path */
  status = sc_string_constructFrom(&dbPath, lase->basePath.str_);
  sm_status_returnIfFailed(status);
  status = sc_string_append(&dbPath, "/db/lase.db");
  sm_status_returnIfFailed(status);

  /* open device database */
  rc = sqlite3_open(dbPath.str_, &pom->db);
  if(rc != 0) {
    fprintf(stderr, "Failed to open database \"%s\" error:%s\n", dbPath.str_,
            sqlite3_errmsg(pom->db));
    sqlite3_close(pom->db);
    sc_string_destruct(&dbPath);
    sm_status_returnError(status, sc_status_failure);
  }

  /* init database flags */
  rc = sqlite3_exec(pom->db, "update devices set created='FALSE';", NULL, 0, &errMsg);
  if(rc != 0) {
    fprintf(stderr, "Failed updating devices \"%s\" error:%s\n", dbPath.str_,
            errMsg);
    sqlite3_free(errMsg);
    sqlite3_close(pom->db);
    sc_string_destruct(&dbPath);
    sm_status_returnError(status, sc_status_failure);
  }  
  rc = sqlite3_exec(pom->db, "update scripts set loaded='FALSE';", NULL, 0, &errMsg);
  if(rc != 0) {
    fprintf(stderr, "Failed updating scripts \"%s\" error:%s\n", dbPath.str_,
            errMsg);
    sqlite3_free(errMsg);
    sqlite3_close(pom->db);
    sc_string_destruct(&dbPath);
    sm_status_returnError(status, sc_status_failure);
  }
  
  status = sc_string_destruct(&dbPath);
  sm_status_returnIfFailed(status);
  
  /* construct the cache lists */
  status = sc_nameToNameList_construct(&pom->inameFromDevAddr, 8, 32);
  sm_status_returnIfFailed(status);  
  status = sc_nameToNameList_construct(&pom->classFromDevType, 8, 32);
  sm_status_returnIfFailed(status);
  status = sc_nameToNameList_construct(&pom->instcFromDevAddr, 8, 64);
  sm_status_returnIfFailed(status);  

  /* TODO: switch cache to hashtables! */
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_destruct(sc_automationDeviceMgr* pom)
{
  sc_status status = sc_status_success;

  /* free caches */
  status = sc_nameToNameList_destruct(&pom->inameFromDevAddr);
  sm_status_returnIfFailed(status);  
  status = sc_nameToNameList_destruct(&pom->classFromDevType);
  sm_status_returnIfFailed(status);
  status = sc_nameToNameList_destruct(&pom->instcFromDevAddr);
  sm_status_returnIfFailed(status);

  sqlite3_close(pom->db);
  
  sm_status_return(status);
}

static int nameCallback(void* uarg, int argc, char* argv[], char* names[])
{ /* Extract name from first column */
  char* buff = (char*) uarg;
  if(argc != 1) {
    buff[0] = 0;
  } else {
    strcpy(buff, argv[0]);
  }
  return 0;
}

sc_status sc_automationDeviceMgr_getInstanceNameFromDeviceAddress(
                                          sc_automationDeviceMgr* pom,
                                          const TCHAR* deviceAddr,
                                          const TCHAR** instanceName)
{
  static const char* fmt = "select identifier from devices where address='%s';";
  sc_status status = sc_status_success;
  sc_nameToNameListIterator it;
  sc_nameToName* pn2n = NULL;
  int rc;
  char buff[128], *errMsg=0;
  char inName[128];

  /* check for a cache hit */
  status = sc_nameToNameList_getBegin(&pom->inameFromDevAddr, &it);
  sm_status_returnIfFailed(status);
  
  for(; !sc_nameToNameListIterator_isEnd(&it);
         sc_nameToNameListIterator_next(&it))
  {
    pn2n = sc_nameToNameListIterator_get(&it);
    if(strcmp(pn2n->first, deviceAddr) == 0) {
      *instanceName = pn2n->second;
      /* TODO: move this to the front of the cache list */
      sm_status_return(status);
    }
  }

  /* look up device name in database */
  sprintf(buff, fmt, deviceAddr);
  inName[0] = 0;
  rc = sqlite3_exec(pom->db, buff, nameCallback, inName, &errMsg);
  if(rc != 0) {
    fprintf(stderr, "instanceName lookup failed: %s\n", errMsg);
    sqlite3_free(errMsg);
    sm_status_returnError(status, sc_status_failure);
  }
  if(strlen(inName) == 0) {
    *instanceName = NULL;
  } else {
    sc_nameToName* pf = 0;
    sc_nameToName n2n;

    /* add it to the cache */
    n2n.first  = deviceAddr;
    n2n.second = inName;
    status = sc_nameToNameList_pushFront(&pom->inameFromDevAddr, &n2n);
    sm_status_returnIfFailed(status);
    status = sc_nameToNameList_front(&pom->inameFromDevAddr, &pf);
    sm_status_returnIfFailed(status);

    *instanceName = pf->second;
  }
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_getClassFromDeviceType(
                                          sc_automationDeviceMgr* pom,
                                          const TCHAR* deviceType,
                                          const TCHAR** className)
{
  static const char* fmt = "select name from classes where dtype='%s';";
  sc_status status = sc_status_success;
  sc_nameToNameListIterator it;
  sc_nameToName* pn2n = NULL;
  int rc;
  char buff[128], *errMsg=0;
  char clName[128];

  /* check cache first */
  status = sc_nameToNameList_getBegin(&pom->classFromDevType, &it);
  sm_status_returnIfFailed(status);

  for(; !sc_nameToNameListIterator_isEnd(&it);
         sc_nameToNameListIterator_next(&it))
  {
    pn2n = sc_nameToNameListIterator_get(&it);
    if(strcmp(pn2n->first, deviceType) == 0) {
      *className = pn2n->second;
      /* move this to the front of the cache list as an 'age' */      
      sm_status_return(status);
    }
  }

  /* not in cache, lookup in database */
  sprintf(buff, fmt, deviceType);
  clName[0] = 0;
  rc = sqlite3_exec(pom->db, buff, nameCallback, clName, &errMsg);
  if(rc != 0) {
    fprintf(stderr, "className lookup failed: %s\n", errMsg);
    sqlite3_free(errMsg);
    sm_status_returnError(status, sc_status_failure);
  }
  if(strlen(clName) == 0) {
    *className = NULL;
  } else {
    sc_nameToName* pf = 0;
    sc_nameToName n2n;

    /* add it to the cache */
    n2n.first  = deviceType;
    n2n.second = clName;
    status = sc_nameToNameList_pushFront(&pom->classFromDevType, &n2n);
    sm_status_returnIfFailed(status);
    status = sc_nameToNameList_front(&pom->classFromDevType, &pf);
    sm_status_returnIfFailed(status);
    
    *className = pf->second;
  }
  sm_status_return(status);
}

int deviceCreated(void* arg, int argc, char* argv[], char* names[])
{ /* extracts name from first column */
  char** pstr = (char**)arg;
  if(argc == 1) {
    *pstr = strdup(argv[0]);
  }
  return 0;
}

sc_status sc_automationDeviceMgr_getInstanceFromDeviceAddress(
                                          sc_automationDeviceMgr* pom,
                                          const TCHAR* deviceAddr,
                                          const TCHAR** instanceName)
{
  static const char* qf = "select identifier from devices where address='%s' and created='TRUE';";
  sc_status status = sc_status_success;
  sc_nameToNameListIterator it;
  sc_nameToName* pn2n = NULL;
  char qbuff[256], *emsg = NULL;
  int rv;

  status = sc_nameToNameList_getBegin(&pom->instcFromDevAddr, &it);
  sm_status_returnIfFailed(status);

  for(; !sc_nameToNameListIterator_isEnd(&it);
      sc_nameToNameListIterator_next(&it))
  {
    pn2n = sc_nameToNameListIterator_get(&it);
    if(strcmp(pn2n->first, deviceAddr) == 0) {
      *instanceName = pn2n->second;
      sm_status_return(status);
    }      
  }

  /* fall back on database */
  *instanceName = NULL;
  sprintf(qbuff, qf, deviceAddr);
  rv = sqlite3_exec(pom->db, qbuff, deviceCreated, (void*)instanceName, &emsg);
  if(rv != 0) {
    fprintf(stderr, "error querying device table: %s\n", emsg);
    sqlite3_free(emsg);
    sm_status_returnError(status, sc_status_failure);
  }
  if(*instanceName) { /* it was created, cache it */
    sc_nameToName entry;
    
    entry.first  = deviceAddr;
    entry.second = *instanceName;    
    status = sc_nameToNameList_pushBack(&pom->instcFromDevAddr, &entry);
    sm_status_returnIfFailed(status);
    
    status = sc_nameToNameList_back(&pom->instcFromDevAddr, &pn2n);
    sm_status_returnIfFailed(status);
    
    free((void*)*instanceName);
    *instanceName = pn2n->second;    
  }
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_invalidateDeviceCache(
  sc_automationDeviceMgr* pom,
  const TCHAR* instanceName)
{ /* Clear the given instance name from the caches. For renaming. */
  sc_status status = sc_status_success;
  sc_nameToNameListIterator it;
  sc_nameToName* pn2n = NULL;

  /* find any hit in instcFromDevAddr cache */
  status = sc_nameToNameList_getBegin(&pom->instcFromDevAddr, &it);
  sm_status_returnIfFailed(status);

  for(; !sc_nameToNameListIterator_isEnd(&it);
         sc_nameToNameListIterator_next(&it))
  {
    pn2n = sc_nameToNameListIterator_get(&it);
    if(strcmp(pn2n->second, instanceName) == 0) {
      sc_nameToNameList_erase(&pom->instcFromDevAddr, &it);   
      break;
    }      
  }
  
  /* find any hit in inameFromDevAddr cache */
  status = sc_nameToNameList_getBegin(&pom->inameFromDevAddr, &it);
  sm_status_returnIfFailed(status);

  for(; !sc_nameToNameListIterator_isEnd(&it);
         sc_nameToNameListIterator_next(&it))
  {
    pn2n = sc_nameToNameListIterator_get(&it);
    if(strcmp(pn2n->second, instanceName) == 0) {
      sc_nameToNameList_erase(&pom->inameFromDevAddr, &it);
      break;
    }      
  }
  
  sm_status_return(status);
}


typedef struct nintag {
  s_uint32 classNameLength;
  s_uint32 idx;  
} nin_t;

static int allNumbers(const char* c) {
  char ch = *c++;
  while(ch) {
    if(ch < '0' || ch > '9')
      return 0;
    ch = *c++;
  }
  return 1;
}

int newInstName(void* arg, int argc, char* argv[], char* names[])
{ /* makes a new instance name by extracting digits from end of name */
  /* and taking the max */
  nin_t* pnin = (nin_t*)arg;
  char *pnum;
  /* point pnum after classname */
  pnum = argv[0] + pnin->classNameLength; 
  if( allNumbers(pnum) ) {
    /* only a possible conflict if all numbers */
    s_uint32 idx = (s_uint32)atoi(pnum);
    /* take the max */
    pnin->idx = sm_max(pnin->idx, idx);
  }  
  return 0;
}

sc_status sc_automationDeviceMgr_makeInstanceName(sc_automationDeviceMgr* pom,
                                                  const TCHAR* className,
                                                  const TCHAR* addr,
                                                  const TCHAR** pinstName)
{ /* makes an instance name based on the classname */
  static const char* q = "select identifier from devices where identifier LIKE '%s%%';";
  sc_status status = sc_status_success;
  char qbuff[128], in[72], *emsg = NULL;
  nin_t nin;
  int rv;

  /* change first char of classname to lowercase */
  strcpy(in, className);
  in[0] = (TCHAR)tolower(in[0]);

  /* format query for conflicts */
  sprintf(qbuff, q, in);
  nin.classNameLength = strlen(in);
  nin.idx = 0;

  /* Query database to get any similar names and check for conflicts */
  rv = sqlite3_exec(pom->db, qbuff, newInstName, &nin, &emsg);
  if(rv != 0) {
    fprintf(stderr, "failed to query devices, error: %s\n", emsg);
    sqlite3_free(emsg); 
    sm_status_returnError(status, sc_status_failure);
  }

  /* print new name & dup it out */
  sprintf(in, "%s%d", in, nin.idx+1);
  *pinstName = strdup(in);
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_deviceCreated(sc_automationDeviceMgr* pom,
                                               const char* daddr,
                                               const char* ident)
{
  static const char* fmt = "update devices set created='TRUE' where address='%s';";
  sc_status status = sc_status_success;
  sc_nameToName entry;
  char sql[256], *errMsg = NULL;
  int rc;
  
  entry.first  = daddr;
  entry.second = ident;    
  status = sc_nameToNameList_pushBack(&pom->instcFromDevAddr, &entry);
  sm_status_returnIfFailed(status);

  /* update flag in db */
  sprintf(sql, fmt, daddr);
  rc = sqlite3_exec(pom->db, sql, NULL, NULL, &errMsg);
  if(rc != 0) {
    fprintf(stderr, "failed to update devices, error: %s\n", errMsg);
    sqlite3_free(errMsg);
    sm_status_returnError(status, sc_status_failure);
  }

  sm_status_return(status);
}

extern 
int _stricmp(const char* s1, 
             const char* s2);

static const char* _fieldByName(const char* field,
                               int argc, char* argv[], char* names[])
{
  int idx;
  for (idx=0; idx < argc; ++idx) {
    if(_stricmp(names[idx], field) == 0) {
      return argv[idx];
    }
  }
  return NULL;
}

static int archiveCallback(void* a, int argc, char* argv[], char* names[])
{
  sc_string* arch = (sc_string*)a;
  sc_string_assign(arch, _fieldByName("archive",argc,argv,names));  
  return 0;
}

static int devIdCallback(void* a, int argc, char* argv[], char* names[])
{
  s_int32* pdid = (s_int32*)a;  
  *pdid = atoi(_fieldByName("id",argc,argv,names));
  return 0;
}
int getDeviceId(sqlite3* db, const char* ident)
{
  static const char* query = "select id from devices where identifier='%s';";
  char qbuff[128];
  s_int32 rv, did = 0;
  char* emsg = NULL;
  
  sprintf(qbuff, query, ident);
  rv = sqlite3_exec(db, qbuff, devIdCallback, &did, &emsg);
  if(rv != 0) {
    fprintf(stderr, "failed query '%s' : %s\n", qbuff, emsg);
    sqlite3_free(emsg);
    return 0;
  }
  return did;
}

typedef struct rlcbInfo {
  sc_string* script;
  const char* devIdent;
  s_bool isSrc;
} rlcbInfo;

int restoreLinkCallback(void* arg, int argc, char* argv[], char* names[])
{
  static const char* fmt = "%s.events.%s.subscribe(%s,%s.%s);\n";
  char sbuff[512];
  const char *srcIdent, *srcEvent, *dstIdent, *dstMethd;
  rlcbInfo* info = (rlcbInfo*)arg;
  int created = 0;

  if(info->isSrc) {
    srcIdent = info->devIdent;
    srcEvent = _fieldByName("srcEvent",   argc,argv,names);
    dstIdent = _fieldByName("identifier", argc,argv,names);
    dstMethd = _fieldByName("dstMethod",  argc,argv,names);
    created  = _stricmp(_fieldByName("created",argc,argv,names), "TRUE") == 0;
  } else {
    srcIdent = _fieldByName("identifier", argc,argv,names);
    srcEvent = _fieldByName("srcEvent",   argc,argv,names);
    dstIdent = info->devIdent;
    dstMethd = _fieldByName("dstMethod",  argc,argv,names);
    created  = _stricmp(_fieldByName("created",argc,argv,names), "TRUE") == 0;
  }

  if(created) {
    sprintf(sbuff, fmt, srcIdent, srcEvent, dstIdent, dstIdent, dstMethd);
    sc_string_append(info->script, sbuff);
  }

  return 0;
}

sc_status sc_automationDeviceMgr_restoreDeviceLinks(sc_automationDeviceMgr* pom,
                                                    s_int32 did,
                                                    const char* devIdent)
{
  static const char* sq = "select l.*, d.identifier, d.created from links l join devices d on d.id=l.dstDid where l.srcDid=%d;";
  static const char* dq = "select l.*, d.identifier, d.created from links l join devices d on d.id=l.srcDid where l.dstDid=%d;";
  sc_status status = sc_status_success;
  s_int32 rv;
  char qbuff[128], *emsg = NULL;
  rlcbInfo info;
  sc_string script;
  sc_scriptRequestPtrh srp;
  sc_scriptRequest* rp;

  status = sc_string_construct(&script);
  sm_status_returnIfFailed(status);

  info.devIdent = devIdent;
  info.script   = &script;
  info.isSrc    = seTrue;

  /* build up subscribes where devIdent was the src */
  sprintf(qbuff, sq, did);
  rv = sqlite3_exec(pom->db, qbuff, restoreLinkCallback, &info, &emsg);
  if(rv != 0) {
    fprintf(stderr, "failed query '%s', error: %s\n", qbuff, emsg);
    sqlite3_free(emsg);
    sm_status_returnError(status, sc_status_failure);
  }

  /* build up subscribes where devIdent was the dst */
  info.isSrc    = seFalse;
  sprintf(qbuff, dq, did);
  rv = sqlite3_exec(pom->db, qbuff, restoreLinkCallback, &info, &emsg);
  if(rv != 0) {
    fprintf(stderr, "failed query '%s', error: %s\n", qbuff, emsg);
    sqlite3_free(emsg);
    sm_status_returnError(status, sc_status_failure);
  }

  if(script.length_ > 0) {
    /* okay, we now have one big script of all link restoring 'subscribe'
     * calls.  Let's queue this off to the js engine */
    status = sc_scriptRequestPtrh_construct(&srp, NULL);
    sm_status_returnIfFailed(status);

    status = sc_lase_allocRequest(pom->jsEngine, &srp);
    sm_status_returnIfFailed(status);
    rp = sc_scriptRequestPtrh_get(&srp);

    status = sc_string_assign(&rp->scriptIn, script.str_);
    sm_status_returnIfFailed(status);

    status = sc_string_destruct(&script);
    sm_status_returnIfFailed(status);

    status = sc_lase_submitRequest(pom->jsEngine, &srp);
    sm_status_returnIfFailed(status);

    status = sc_scriptRequestPtrh_destruct(&srp);
    sm_status_returnIfFailed(status);
  } else { /* no links to restore */
    status = sc_string_destruct(&script);
    sm_status_returnIfFailed(status);
  }
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_restoreDeviceState(sc_automationDeviceMgr* pom,
                                                    const char* ident)
{
  static const char* qArchFmt = "select archive from archives where deviceId=%d;";
  sc_status status = sc_status_success;
  char qbuff[128];
  char* emsg = NULL;
  s_int32 rv, did;
  sc_string archive;
  sc_scriptRequestPtrh srp;
  sc_scriptRequest* rp;
  
  did = getDeviceId(pom->db, ident);
  if(did == 0)
    sm_status_returnError(status, sc_status_failure);
  
  /* okay, now find any stored archive for device by deviceId */
  sc_string_construct(&archive);
  sprintf(qbuff, qArchFmt, did);
  rv = sqlite3_exec(pom->db, qbuff, archiveCallback, &archive, &emsg);
  if(rv != 0) {
    fprintf(stderr, "failed query '%s' : %s\n", qbuff, emsg);
    sqlite3_free(emsg);
    sc_string_destruct(&archive);
    sm_status_returnError(status, sc_status_failure);
  }

  if(archive.length_ > 0) {
    sprintf(qbuff, "%s._load(eval('(", ident);
    sc_string_prepend(&archive, qbuff);
    sc_string_append(&archive, ")'));");

    status = sc_scriptRequestPtrh_construct(&srp, NULL);
    sm_status_returnIfFailed(status);

    status = sc_lase_allocRequest(pom->jsEngine, &srp);
    sm_status_returnIfFailed(status);
    rp = sc_scriptRequestPtrh_get(&srp);

    status = sc_string_assign(&rp->scriptIn, archive.str_);
    sm_status_returnIfFailed(status);

    status = sc_string_destruct(&archive);
    sm_status_returnIfFailed(status);

    status = sc_lase_submitRequest(pom->jsEngine, &srp);
    sm_status_returnIfFailed(status);

    status = sc_scriptRequestPtrh_destruct(&srp);
    sm_status_returnIfFailed(status);
  } else {
    status = sc_string_destruct(&archive);
    sm_status_returnIfFailed(status);    
  }
  
  status = sc_automationDeviceMgr_restoreDeviceLinks(pom, did, ident);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);  
}


sc_status sc_automationDeviceMgr_saveArchive(sc_automationDeviceMgr* pom,
                                             const TCHAR* ident,
                                             const TCHAR* archive)
{
  static const char* df = "delete from archives where deviceId=%d;";
  static const char* af = "insert into archives (deviceId,archive) values (%d,'%s');";
  sc_status status = sc_status_success;
  s_int32 did, len, rv;
  char qbuff[64], *abuff = NULL, *emsg = NULL;

  did = getDeviceId(pom->db, ident);
  if(did == 0)
    sm_status_returnError(status, sc_status_failure);

  /* deletes any existing archive. */
  sprintf(qbuff, df, did);
  sqlite3_exec(pom->db, qbuff, NULL, NULL, NULL);

  len = strlen(af) + strlen(archive);
  abuff = malloc(len+16);
  if(abuff == NULL) {
    sm_status_returnError(status, sc_err_outOfMemory);
  }
  
  sprintf(abuff, af, did, archive);
  rv = sqlite3_exec(pom->db, abuff, NULL, NULL, &emsg);
  free(abuff); abuff = NULL;
  if(rv != 0) {
    fprintf(stderr, "failed adding archive, error: %s\n", emsg);
    sqlite3_free(emsg);
    sm_status_returnError(status, sc_status_failure);    
  }
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_clearLinks(sc_automationDeviceMgr* pom,
                                            const TCHAR* srcIdent)
{
  static const char* dq = "delete from links where srcDid=%d;";
  sc_status status = sc_status_success;
  char qbuff[64];
  s_int32 srcId;

  srcId = getDeviceId(pom->db, srcIdent);
  sprintf(qbuff, dq, srcId);
  sqlite3_exec(pom->db, qbuff, NULL, NULL, NULL);
  
  sm_status_return(status);
}
                                            

sc_status sc_automationDeviceMgr_saveLink(sc_automationDeviceMgr* pom,
                                          const TCHAR* srcIdent,
                                          const TCHAR* srcEvent,
                                          const TCHAR* dstIdent,
                                          const TCHAR* dstMethd)
{
  static const char* q = "insert into links (srcDid,dstDid,srcEvent,dstMethod) values (%d,%d,\"%s\",\"%s\")";
  sc_status status = sc_status_success;
  s_int32 srcId, dstId, rv, len;
  char* qbuff = NULL, *emsg = NULL;

  srcId = getDeviceId(pom->db, srcIdent);
  dstId = getDeviceId(pom->db, dstIdent);
  
  len = strlen(q) + strlen(dstMethd) + strlen(srcEvent) + 32;
  qbuff = malloc(len);
  if(qbuff == NULL) {
    sm_status_returnError(status, sc_err_outOfMemory);
  }
  
  sprintf(qbuff, q, srcId,dstId,srcEvent,dstMethd);

  rv = sqlite3_exec(pom->db, qbuff, NULL, NULL, &emsg);
  
  free(qbuff);
  
  if(rv != 0) {
    sm_status_returnError(status, sc_status_failure);
  }
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_addDeviceEntry(sc_automationDeviceMgr* pom,
                                                const TCHAR* dtyp,
                                                const TCHAR* addr,
                                                const TCHAR* idnt)
{
  static const char* qf = "insert into devices (address,type,identifier) "
                          "values ('%s','%s','%s');";
  sc_status status = sc_status_success;
  char qbuff[256], *emsg = NULL;
  int rv;

  sprintf(qbuff, qf, addr, dtyp, idnt);
  rv = sqlite3_exec(pom->db, qbuff, NULL, NULL, &emsg);
  if(rv != 0) {
    fprintf(stderr, "failed to add device entry: %s\n", emsg);
    sqlite3_free(emsg);
    sm_status_returnError(status, sc_status_failure);
  }

  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_renameDevice(sc_automationDeviceMgr* pom,
                                              const TCHAR* curName,
                                              const TCHAR* newName)
{
  static const char* scr = "this.%s = %s; delete this.%s;";
  static const char* qf = "update devices set identifier='%s' where identifier='%s';";
  sc_status status = sc_status_success;
  sc_scriptRequestPtrh srp;
  sc_scriptRequest* rp = NULL;
  char qbuff[256], *emsg = NULL;
  int length = strlen(scr), rv;

  status = sc_scriptRequestPtrh_construct(&srp, NULL);
  sm_status_returnIfFailed(status);

  status = sc_lase_allocRequest(pom->jsEngine, &srp);
  sm_status_returnIfFailed(status);
  rp = sc_scriptRequestPtrh_get(&srp);

  length += strlen(curName) * 2 + strlen(newName);

  status = sc_string_setMinCapacity(&rp->scriptIn, length);
  if(sm_status_failed(status)) {
    sc_scriptRequestPtrh_destruct(&srp);
    sm_status_return(status);
  }

  sprintf(rp->scriptIn.str_, scr, newName, curName, curName);
  status = sc_lase_executeRequest(pom->jsEngine, &srp);
  if(sm_status_failed(status)) {
    sc_scriptRequestPtrh_destruct(&srp);
    sm_status_return(status);
  }

  status = sc_scriptRequestPtrh_destruct(&srp);
  sm_status_returnIfFailed(status);

  sprintf(qbuff, qf, newName, curName);
  rv = sqlite3_exec(pom->db, qbuff, NULL, NULL, &emsg);
  if(rv != 0) {
    fprintf(stderr, "error: updating devices - %s\n", emsg);
    sqlite3_free(emsg);
    sm_status_returnError(status, sc_status_failure);
  }

  /* last, but not least, invalidate cache entry */
  status = sc_automationDeviceMgr_invalidateDeviceCache(pom, curName);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_createDevice(sc_automationDeviceMgr* pom,
                                              const TCHAR* brgName,
                                              sc_automationDevice* pd)
{
  static const TCHAR* scriptFmt = "this.%s = new %s(\"%s\", \"%s\", \"%s\", \"%s\", %s);\n";
  sc_status status = sc_status_success;
  sc_scriptRequestPtrh srp;
  sc_scriptRequest* rp = NULL;
  const TCHAR *className = NULL, *dtyp, *ntyp, *addr, *instName = NULL;
  int len, nameAllocd = 0;

  dtyp = pd->type_;
  ntyp = pd->netType_;
  addr = pd->address_;

  /* lookup classname from device type */
  status = sc_automationDeviceMgr_getClassFromDeviceType(pom, dtyp, &className);
  sm_status_returnIfFailed(status);
  if(className == NULL) {
    sm_status_returnError(status, sc_status_failure);
  }

  /* Lookup name for this device by address */
  status = sc_automationDeviceMgr_getInstanceNameFromDeviceAddress(pom, addr, &instName);
  sm_status_returnIfFailed(status);
  if(instName == NULL) {  /* device hasn't been seen yet, no name */
    status = sc_automationDeviceMgr_makeInstanceName(pom, className, addr, &instName);
    sm_status_returnIfFailed(status);

    /* add this device entry to make sure the device names remain
     * stable on discovery. */
    status = sc_automationDeviceMgr_addDeviceEntry(pom, dtyp, addr, instName);
    sm_status_returnIfFailed(status);
    
    nameAllocd = 1;
  }

  status = sc_scriptRequestPtrh_construct(&srp, NULL);
  sm_status_returnIfFailed(status);

  /* allocate script request */
  status = sc_lase_allocRequest(pom->jsEngine, &srp);
  sm_status_returnIfFailed(status);
  rp = sc_scriptRequestPtrh_get(&srp);

  /* compute length of script */
  len = strlen(scriptFmt) + strlen(className) + strlen(addr) +
        strlen(dtyp) + strlen(ntyp) + strlen(brgName) + strlen(instName) * 2 ;

  /* make sure there's enough storage in script */
  status = sc_string_setMinCapacity(&rp->scriptIn, len+8);
  sm_status_returnIfFailed(status);

  /* sprintf script */
  sprintf(rp->scriptIn.str_, scriptFmt,
          instName, className, instName,
          addr, dtyp, ntyp, brgName);

  /* submit script for execution */
  status = sc_lase_submitRequest(pom->jsEngine, &srp);
  sm_status_returnIfFailed(status);

  /* used for fast lookup */
  status = sc_automationDeviceMgr_deviceCreated(pom, addr, instName);
  sm_status_returnIfFailed(status);

  /* restore device data (including event target functions added by user) */
  status = sc_automationDeviceMgr_restoreDeviceState(pom, instName);
  sm_status_returnIfFailed(status);

  status = sc_scriptRequestPtrh_destruct(&srp);
  sm_status_returnIfFailed(status);

  if(nameAllocd) {
    free((void*)instName);
  }
  
  sm_status_return(status);
}

sc_status sc_automationDeviceMgr_relayMessage(sc_automationDeviceMgr* pom,
                                              const TCHAR* deviceAddress,
                                              sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;
  sc_scriptRequestPtrh srp;
  sc_scriptRequest* rp = NULL;
  const TCHAR *instanceName = NULL;

  status = sc_automationDeviceMgr_getInstanceFromDeviceAddress(pom,
                                             deviceAddress, &instanceName);
  sm_status_returnIfFailed(status);
  if(instanceName == NULL) 
    sm_status_return(status); /* no instance, don't need to relay */

  status = sc_scriptRequestPtrh_construct(&srp, NULL);
  sm_status_returnIfFailed(status);

  /* alloc request for message relay */
  status = sc_lase_allocRequest(pom->jsEngine, &srp);
  sm_status_returnIfFailed(status);
  rp = sc_scriptRequestPtrh_get(&srp);

  rp->flags = se_scriptRequest_RelayMessage;
  rp->pkt = pkt;
  rp->obj = instanceName;

  /* submit is fine in this case, message was async anyway */
  status = sc_lase_submitRequest(pom->jsEngine, &srp);
  sm_status_returnIfFailed(status);

  status = sc_scriptRequestPtrh_destruct(&srp);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}


sc_status sc_nameToName_constructCopy(sc_nameToName* pn2n, sc_nameToName* poth)
{
  sc_status status = sc_status_success;

  pn2n->first  = strdup(poth->first);
  pn2n->second = strdup(poth->second);
  
  sm_status_return(status);
}

sc_status sc_nameToName_assign(sc_nameToName* pn2n, sc_nameToName* poth)
{ /* assignment operator for nameToName struct  */
  sc_status status = sc_status_success;

  if(pn2n->first)
    free((void*)pn2n->first);
  if(pn2n->second)
    free((void*)pn2n->second);
  
  pn2n->first  = strdup(poth->first);
  pn2n->second = strdup(poth->second);  
  
  sm_status_return(status);
}
sc_status sc_nameToName_destruct(sc_nameToName* pn2n)
{ /* destruct nameToName struct  */
  sc_status status = sc_status_success;
  if(pn2n->first)
    free((void*)pn2n->first);
  if(pn2n->second)
    free((void*)pn2n->second);
  sm_status_return(status);
}


sm_listImpl(sc_nameToName);
