#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <si_sha/autopkt.h>
#include <pthread.h>

#include <assert.h>

sc_status sc_automationPacket_construct(sc_automationPacket* ppkt,
                                        s_int32 capacity,
                                        s_uint8* pbuffer)
{
  sc_status status = sc_status_success;

  if(pbuffer == NULL)
  { /* ever use this case? */
    pbuffer = malloc(capacity);
    if(pbuffer == NULL)
      sm_status_returnError(status, sc_err_outOfMemory);
    ppkt->priv_.own_ = seTrue;
  } else ppkt->priv_.own_ = seFalse;

  ppkt->priv_.buffer_   = pbuffer;
  ppkt->priv_.top_      = pbuffer + capacity;
  ppkt->priv_.capacity_ = capacity;
  ppkt->priv_.size_     = 0;

  /* just initialize rep to NULL, caller has to set this up */
  status = sc_rep_construct(&ppkt->_rep, ppkt, NULL, NULL);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status sc_automationPacket_destruct(sc_automationPacket* ppkt)
{
  sc_status status = sc_status_success;  
  if(ppkt->priv_.own_)
    free(ppkt->priv_.buffer_);
  sm_status_return(status);
}

sc_status sc_automationPacket_pushBinary(sc_automationPacket* ppkt,
                                         const s_uint8* pdata,
                                         s_int32 nbytes)
{
  sc_status status = sc_status_success;

  if( (ppkt->priv_.size_ + nbytes) > ppkt->priv_.capacity_)
    sm_status_returnError(status, sc_err_outOfMemory); /* packets don't grow. */

  /* stack grows down */
  ppkt->priv_.size_ += nbytes;
  ppkt->priv_.top_  -= nbytes; 

  memcpy(ppkt->priv_.top_, pdata, nbytes);

  sm_status_return(status);
}

/* macro to generate push methods. */
#define sm_packetPush(S) \
  sc_status sc_automationPacket_pushUInt##S(sc_automationPacket* ppkt,\
                                            s_uint##S data)\
  {\
    sc_status status = sc_status_success;\
    if( (ppkt->priv_.size_ + (int)sizeof(data)) > ppkt->priv_.capacity_)\
      sm_status_returnError(status, sc_err_outOfMemory); \
    ppkt->priv_.size_ += sizeof(data);\
    ppkt->priv_.top_  -= sizeof(data);\
    *((s_uint##S*)ppkt->priv_.top_) = data;\
    sm_status_return(status);\
  }\
  sc_status sc_automationPacket_pushInt##S(sc_automationPacket* ppkt,\
                                           s_int##S data)\
  {\
    sc_status status = sc_status_success;\
    if( (ppkt->priv_.size_ + (int)sizeof(data)) > ppkt->priv_.capacity_)\
      sm_status_returnError(status, sc_err_outOfMemory); \
    ppkt->priv_.size_ += sizeof(data);\
    ppkt->priv_.top_  -= sizeof(data);\
    *((s_int##S *)ppkt->priv_.top_) = data;\
    sm_status_return(status);\
  }

sm_packetPush(8);
sm_packetPush(16);
sm_packetPush(32);
sm_packetPush(64);


sc_status sc_automationPacket_popBinary(sc_automationPacket* ppkt,
                                        s_uint8* pdata,
                                        s_int32 nbytes)
{
  sc_status status = sc_status_success;

  if(ppkt->priv_.size_ < nbytes)
    sm_status_returnError(status, sc_err_badArgument);

  memcpy(pdata, ppkt->priv_.top_, nbytes);

  ppkt->priv_.size_ -= nbytes;
  ppkt->priv_.top_  += nbytes; 

  sm_status_return(status);
}

/* macro to generate pop methods. */
#define sm_packetPop(S)\
  sc_status sc_automationPacket_popUInt##S(sc_automationPacket* ppkt,\
                                           s_uint##S* pdata) \
  {\
    sc_status status = sc_status_success;\
    \
    if(ppkt->priv_.size_ < sizeof(*pdata)) \
      sm_status_returnError(status, sc_status_failure);\
    \
    ppkt->priv_.size_ -= sizeof(*pdata);\
    *pdata = *((s_uint##S*)ppkt->priv_.top_);\
    ppkt->priv_.top_ += sizeof(*pdata);\
    sm_status_return(status);\
  }\
  sc_status sc_automationPacket_popInt##S(sc_automationPacket* ppkt, s_int##S* pdata) \
  {\
    sc_status status = sc_status_success;\
    \
    if(ppkt->priv_.size_ < sizeof(*pdata)) \
      sm_status_returnError(status, sc_status_failure);\
    \
    ppkt->priv_.size_ -= sizeof(*pdata);\
    *pdata = *((s_int##S*)ppkt->priv_.top_);\
    ppkt->priv_.top_ += sizeof(*pdata);\
    sm_status_return(status);\
  }

sm_packetPop(8);
sm_packetPop(16);
sm_packetPop(32);
sm_packetPop(64);

sc_status sc_automationPacket_getSize(sc_automationPacket* ppkt, s_int32* psz)
{
  sc_status status = sc_status_success;
  *psz = ppkt->priv_.size_;
  sm_status_return(status);  
}

sc_status sc_automationPacket_clear(sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;

  /* sets packet to empty without changing capacity */
  pkt->priv_.top_ = pkt->priv_.buffer_ + pkt->priv_.capacity_;
  pkt->priv_.size_ = 0;

  sm_status_return(status);
}

sc_status sc_automationPacket_tell(sc_automationPacket* pkt, s_uint32* ppos)
{
  sc_status status = sc_status_success;

  /* Return the current size of the data. top_ can always be computed from
   * this. */
  *ppos = pkt->priv_.size_;
  
  sm_status_return(status);  
}

sc_status sc_automationPacket_setPos(sc_automationPacket* pkt, s_uint32 pos)
{
  sc_status status = sc_status_success;

  if(pos > pkt->priv_.capacity_) 
    sm_status_returnError(status, sc_status_failure);  

  /* pos is the size of the data ... compute top_ position from this */
  pkt->priv_.top_ = pkt->priv_.buffer_ + pkt->priv_.capacity_ - pos;
  pkt->priv_.size_ = pos;
  
  sm_status_return(status);  
}


sc_status sc_automationPacket_serializeToFd(sc_automationPacket* pkt,
                                            int fd)
{
  sc_status status = sc_status_success;

  /* writes the payload size and then payload */
  write(fd, &pkt->priv_.size_, sizeof(pkt->priv_.size_));
  write(fd, pkt->priv_.top_, pkt->priv_.size_);
  
  sm_status_return(status);  
}
sc_status sc_automationPacket_deserializeFromFd(sc_automationPacket* pkt,
                                                int fd)
{
  sc_status status = sc_status_success;
  s_int32 sz;
  
  /* read payload size and then payload */
  read(fd, &sz, sizeof(pkt->priv_.size_));
  assert(sz < pkt->priv_.capacity_);
  /* set internal stack to point to top of payload */
  pkt->priv_.size_ = sz;
  pkt->priv_.top_ = pkt->priv_.buffer_ + pkt->priv_.capacity_ - sz;
  read(fd, pkt->priv_.top_, pkt->priv_.size_);

  sm_status_return(status);  
}



sc_status sc_automationPacket_returnToPool(sc_automationPacket* pkt,
                                           sc_automationPacketPool* ppool)
{
  sc_status status = sc_status_success;
  sc_automationPacketPtrh ptrh;

  /* empties the packet so it's 'fresh' on next allocation */
  status = sc_automationPacket_clear(pkt);
  sm_status_returnIfFailed(status);

  status = sc_automationPacketPtrh_construct(&ptrh, &pkt->_rep);
  sm_status_returnIfFailed(status);

  /* adds it back to free pool */
  status = sc_automationPacketPool_free(ppool, &ptrh);  
  sm_status_returnIfFailed(status);  

  /* pool now refs it */
  status = sc_automationPacketPtrh_destruct(&ptrh);
  sm_status_returnIfFailed(status);  
  
  sm_status_return(status);  
}


/*
 * Creates 'nodes' new packets for the packet pool
 * NOTE: pool mux must be held by caller.
 * 
 */
static 
sc_status sc_automationPacketPool_newNodes(sc_automationPacketPool* pool,
                                           s_int32 nodes)
{
  /* vtable sets packets to be returned to pool on delete */
  static const sc_rep_vtable vtbl = {
    sc_rep_addRef_default,
    sc_rep_relRef_default,
    (sc_rep_delete)sc_automationPacket_returnToPool};
  sc_status status = sc_status_success;
  sc_automationPacketPtrh ptrh;
  sc_automationPacket* ppkt = NULL;
  s_int32 nidx;
  
  for(nidx = 0; nidx < nodes; ++nidx)
  {
    /* note that allocator creates storage for packet object & data all
     * in a single allocation */
    status = sc_blockAllocator_allocate(&pool->al_, (void**)&ppkt);
    sm_status_returnIfFailed(status);

    /* set data buffer to point to just after packet object. */
    status = sc_automationPacket_construct(ppkt, pool->packetSize_,
                                          (s_uint8*)(ppkt+1));
    if(sm_status_failed(status)) {
      sc_blockAllocator_free(&pool->al_, ppkt);
      sm_status_return(status);
    }

    /* on free, return packet to this pool */
    ppkt->_rep.pown = pool;   /* pool is owner */
    ppkt->_rep.vtbl = &vtbl;  /* vtbl set 'returnToPool' for delete */
    
    status = sc_automationPacketPtrh_construct(&ptrh, &ppkt->_rep);
    if(sm_status_failed(status)) {
      sc_automationPacket_destruct(ppkt);
      sc_blockAllocator_free(&pool->al_, ppkt);
      sm_status_return(status);
    }

    status = sc_automationPacketPtrhList_pushBack(&pool->pool_, &ptrh);
    if(sm_status_failed(status)) {
      sc_automationPacketPtrh_destruct(&ptrh);
      sm_status_return(status);
    }
    
    /* list now references packet, can destruct temp ptrh */
    sc_automationPacketPtrh_destruct(&ptrh);
  }

  sm_status_return(status);
}

/*
 * Allocate a packet from the pool, growing pool if needed
 */
sc_status sc_automationPacketPool_allocate(sc_automationPacketPool* ppool,
                                           sc_automationPacketPtrh* pptrh)
{
  sc_status status = sc_status_success;
  s_uint32 pc;
    
  pthread_mutex_lock(&ppool->mux_);

  status = sc_automationPacketPtrh_assign(pptrh, NULL);
  sm_status_returnIfFailed(status);
  
  status = sc_automationPacketPtrhList_getSize(&ppool->pool_, &pc);
  if(sm_status_failed(status)) {
    pthread_mutex_unlock(&ppool->mux_);
    sm_status_return(status);
  }
    

  if(pc == 0) { /* pool list is empty */
    if(ppool->currCapacity_ < ppool->maxCapacity_) { 
      s_int32 allocCount = ppool->currCapacity_; /* binary growth */
      
      /* limit to pool maxCapacity */
      if((allocCount + ppool->currCapacity_) > ppool->maxCapacity_)
        allocCount = ppool->maxCapacity_ - ppool->currCapacity_;
      
      /* create allocCount new nodes. */
      status = sc_automationPacketPool_newNodes(ppool, allocCount);
      if(sm_status_failed(status)) {
        pthread_mutex_unlock(&ppool->mux_);
        sm_status_return(status);
      }
      ppool->currCapacity_ += allocCount; /* update capacity */
      
      /* This can be a warning that packets are leaking somewhere. */
      printf("Packet pool grown: now %d packets.\n", ppool->currCapacity_);
    } else {
      /* user said not to grow any further than this */
      pthread_mutex_unlock(&ppool->mux_);
      sm_status_returnError(status, sc_err_outOfMemory);
    }
  }

  /* allocate and return the packet */
  status = sc_automationPacketPtrhList_popBack(&ppool->pool_, pptrh);
  if(sm_status_failed(status)) {
    pthread_mutex_unlock(&ppool->mux_);
    sm_status_return(status);
  }

  pthread_mutex_unlock(&ppool->mux_);
  
  sm_status_return(status);
}


sc_status sc_automationPacketPool_free(sc_automationPacketPool* ppool,
                                       sc_automationPacketPtrh* pptrh)
{
  sc_status status = sc_status_success;

  pthread_mutex_lock(&ppool->mux_);

  /* push the packet in pptrh back into the free pool */
  status = sc_automationPacketPtrhList_pushBack(&ppool->pool_, pptrh);
  if(sm_status_failed(status)) {
    pthread_mutex_unlock(&ppool->mux_);
    sm_status_return(status);
  }
  
  pthread_mutex_unlock(&ppool->mux_);
  
  sm_status_return(status);
}

sc_status sc_automationPacketPool_construct(sc_automationPacketPool* ppool,
                                            s_int32 pktSize,
                                            s_int32 initialCapacity,
                                            s_int32 maxCapacity)
{
  sc_status status = sc_status_success;
  pthread_mutexattr_t mattr;

  /* create mutex to protect pool */
  pthread_mutexattr_init(&mattr);
  pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE_NP);
  pthread_mutex_init(&ppool->mux_, &mattr);
  
  ppool->packetSize_  = pktSize;
  ppool->maxCapacity_ = maxCapacity;

  /* Create free list */
  status = sc_automationPacketPtrhList_construct(&ppool->pool_,
                                                 initialCapacity,
                                                 maxCapacity);
  sm_status_returnIfFailed(status);

  /* Create a block allocator for blocks */
  status = sc_blockAllocator_construct(&ppool->al_,
                                       pktSize+sizeof(sc_automationPacket),
                                       initialCapacity, maxCapacity);
  sm_status_returnIfFailed(status);
  
  if(initialCapacity) { /* create initialCapacity free nodes */
    status = sc_automationPacketPool_newNodes(ppool, initialCapacity);
    if(sm_status_failed(status)) {
      sc_automationPacketPtrhList_destruct(&ppool->pool_);
      sm_status_return(status);
    }
    ppool->currCapacity_ = initialCapacity;
  }
  
  sm_status_return(status);
}

sc_status sc_automationPacket_delete(sc_automationPacket* pkt,
                                     sc_blockAllocator* owner)
{
  sc_status status = sc_status_success;
  
  status = sc_automationPacket_destruct(pkt);
  sm_status_returnIfFailed(status);

  /* back to allocator it came from */
  status = sc_blockAllocator_free(owner, pkt);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status sc_automationPacketPool_destruct(sc_automationPacketPool* ppool)
{
  static const sc_rep_vtable vtbl = {
        sc_rep_addRef_default,
        sc_rep_relRef_default,
        (sc_rep_delete)sc_automationPacket_delete };    
  sc_status status = sc_status_success;
  sc_automationPacketPtrhListIterator it;
          
  /* reset the vtables in packets to free them back to heap instead of
   * packet pool. */
  status = sc_automationPacketPtrhList_getBegin(&ppool->pool_, &it);
  sm_status_returnIfFailed(status);
  
  for(; !sc_automationPacketPtrhListIterator_isEnd(&it);
         sc_automationPacketPtrhListIterator_next(&it))
  {
    sc_automationPacketPtrh* ptrh = sc_automationPacketPtrhListIterator_get(&it);
    sc_automationPacket* pkt = sc_automationPacketPtrh_get(ptrh);
    pkt->_rep.vtbl = &vtbl;
    pkt->_rep.pown = &ppool->al_;
  }

  /* Clear the free list.  Packets released to allocator at this point */
  status = sc_automationPacketPtrhList_clear(&ppool->pool_);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrhList_destruct(&ppool->pool_);
  sm_status_returnIfFailed(status);
  
  /* and back to heap at this point */
  status = sc_blockAllocator_destruct(&ppool->al_);
  sm_status_returnIfFailed(status);

  pthread_mutex_destroy(&ppool->mux_);

  sm_status_return(status);
}

/* instantiation of ptrh and list packet classes */
sm_ptrhImpl(sc_automationPacket);
sm_listImpl(sc_automationPacketPtrh);
