#include <stdlib.h>
#include <stdio.h>

#include <si_sha/sreqpool.h>


sc_status sc_scriptRequest_construct(sc_scriptRequest* psr)
{
  sc_status status = sc_status_success;

  status = sc_string_construct(&psr->scriptIn);
  sm_status_returnIfFailed(status);

  status = sc_string_construct(&psr->scriptOut);
  sm_status_returnIfFailed(status);

  status = sc_sem_construct(&psr->complete, 0, 1);
  sm_status_returnIfFailed(status);

  psr->flags = 0;
  psr->obj = NULL;
  psr->pkt = NULL;

  status = sc_rep_construct(&psr->_rep, psr, NULL, NULL);
  sm_status_returnIfFailed(status);

  psr->scriptStatus = sc_status_success;  
  
  sm_status_return(status);
}

sc_status sc_scriptRequest_destruct(sc_scriptRequest* psr)
{
  sc_status status = sc_status_success;

  status = sc_string_destruct(&psr->scriptIn);
  sm_status_returnIfFailed(status);
  status = sc_string_destruct(&psr->scriptOut);
  sm_status_returnIfFailed(status);
  status = sc_sem_destruct(&psr->complete);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_scriptRequest_clear(sc_scriptRequest* psr)
{
  sc_status status = sc_status_success;

  status = sc_string_clear(&psr->scriptIn);
  sm_status_returnIfFailed(status);
  status = sc_string_clear(&psr->scriptOut);
  sm_status_returnIfFailed(status);
  
  psr->flags = 0;
  psr->obj = NULL;
  psr->pkt = NULL;

  psr->scriptStatus = sc_status_success;

  status = sc_sem_trywait(&psr->complete, NULL);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_scriptRequest_returnToPool(sc_scriptRequest* req,
                                        sc_scriptRequestPool* ppool)
{
  sc_status status = sc_status_success;
  sc_scriptRequestPtrh ptrh;

  status = sc_scriptRequest_clear(req);
  sm_status_returnIfFailed(status);

  status = sc_scriptRequestPtrh_construct(&ptrh, &req->_rep);
  sm_status_returnIfFailed(status);

  status = sc_scriptRequestPool_free(ppool, &ptrh);  
  sm_status_returnIfFailed(status);

  /* pool now references it */
  status = sc_scriptRequestPtrh_destruct(&ptrh);
  sm_status_returnIfFailed(status);  

  sm_status_return(status);  
}


/*
 * Creates 'nodes' new packets for the packet pool
 * NOTE: pool mux must be held by caller.
 * 
 */
static 
sc_status sc_scriptRequestPool_newNodes(sc_scriptRequestPool* pool,
                                        s_int32 nodes)
{
  static const sc_rep_vtable vtbl = {
    sc_rep_addRef_default,
    sc_rep_relRef_default,
    (sc_rep_delete)sc_scriptRequest_returnToPool
  };
  sc_status status = sc_status_success;
  sc_scriptRequestPtrh ptrh;
  sc_scriptRequest* preq = NULL;
  s_int32 nidx;

  for(nidx = 0; nidx < nodes; ++nidx)
  {
    status = sc_blockAllocator_allocate(&pool->al_, (void**)&preq);
    sm_status_returnIfFailed(status);

    status = sc_scriptRequest_construct(preq);
    if(sm_status_failed(status)) {
      sc_blockAllocator_free(&pool->al_, preq);
      sm_status_return(status);
    }

    /* on free, return packet to this pool */
    preq->_rep.pown = pool;
    preq->_rep.vtbl = &vtbl;

    status = sc_scriptRequestPtrh_construct(&ptrh, &preq->_rep);
    if(sm_status_failed(status)) {
      sc_scriptRequest_destruct(preq);
      sc_blockAllocator_free(&pool->al_, preq);
      sm_status_return(status);
    }

    status = sc_scriptRequestPtrhList_pushBack(&pool->pool_, &ptrh);
    if(sm_status_failed(status)) {
      sc_scriptRequestPtrh_destruct(&ptrh);
      sm_status_return(status);
    }

    /* list now references request, can destruct temp ptrh */
    sc_scriptRequestPtrh_destruct(&ptrh);
  }

  sm_status_return(status);
}

/*
 * Allocate a packet from the pool, growing pool if needed
 */
sc_status sc_scriptRequestPool_allocate(sc_scriptRequestPool* ppool,
                                        sc_scriptRequestPtrh* pptrh)
{
  sc_status status = sc_status_success;
  s_uint32 pc;


  pthread_mutex_lock(&ppool->mux_);

  status = sc_scriptRequestPtrh_assign(pptrh, NULL);
  sm_status_returnIfFailed(status);
  
  status = sc_scriptRequestPtrhList_getSize(&ppool->pool_, &pc);
  if(sm_status_failed(status)) {
    pthread_mutex_unlock(&ppool->mux_);
    sm_status_return(status);
  }


  if(pc == 0) { /* pool list is empty */
    if(ppool->currCapacity_ < ppool->maxCapacity_) { 
      s_int32 allocCount = ppool->currCapacity_; /* binary growth */
      if((allocCount + ppool->currCapacity_) > ppool->maxCapacity_)
        allocCount = ppool->maxCapacity_ - ppool->currCapacity_;
      status = sc_scriptRequestPool_newNodes(ppool, allocCount);
      if(sm_status_failed(status)) {
        pthread_mutex_unlock(&ppool->mux_);
        sm_status_return(status);
      }
      ppool->currCapacity_ += allocCount;
      printf("Script request pool grown: now %d packets.\n", ppool->currCapacity_);
    } else {
      pthread_mutex_unlock(&ppool->mux_);
      sm_status_returnError(status, sc_err_outOfMemory);
    }
  }

  status = sc_scriptRequestPtrhList_popBack(&ppool->pool_, pptrh);
  if(sm_status_failed(status)) {
    pthread_mutex_unlock(&ppool->mux_);
    sm_status_return(status);
  }

  pthread_mutex_unlock(&ppool->mux_);

  sm_status_return(status);
}


sc_status sc_scriptRequestPool_free(sc_scriptRequestPool* ppool,
                                    sc_scriptRequestPtrh* pptrh)
{
  sc_status status = sc_status_success;

  pthread_mutex_lock(&ppool->mux_);

  status = sc_scriptRequestPtrhList_pushBack(&ppool->pool_, pptrh);
  if(sm_status_failed(status)) {
    pthread_mutex_unlock(&ppool->mux_);
    sm_status_return(status);
  }

  pthread_mutex_unlock(&ppool->mux_);

  sm_status_return(status);
}

sc_status sc_scriptRequestPool_construct(sc_scriptRequestPool* ppool,
                                            s_int32 initialCapacity,
                                            s_int32 maxCapacity)
{
  sc_status status = sc_status_success;
  pthread_mutexattr_t mattr;

  pthread_mutexattr_init(&mattr);
  pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE_NP);
  pthread_mutex_init(&ppool->mux_, &mattr);

  ppool->maxCapacity_ = maxCapacity;

  status = sc_scriptRequestPtrhList_construct(&ppool->pool_,
                                              initialCapacity,
                                              maxCapacity);
  sm_status_returnIfFailed(status);

  status = sc_blockAllocator_construct(&ppool->al_,
                                       sizeof(sc_scriptRequest),
                                       initialCapacity, maxCapacity);
  sm_status_returnIfFailed(status);

  if(initialCapacity) {
    status = sc_scriptRequestPool_newNodes(ppool, initialCapacity);
    if(sm_status_failed(status)) {
      sc_scriptRequestPtrhList_destruct(&ppool->pool_);
      sm_status_return(status);
    }
    ppool->currCapacity_ = initialCapacity;
  }

  sm_status_return(status);
}

sc_status sc_scriptRequest_delete(sc_scriptRequest* req,
                                  sc_blockAllocator* owner)
{
  sc_status status = sc_status_success;

  status = sc_scriptRequest_destruct(req);
  sm_status_returnIfFailed(status);

  status = sc_blockAllocator_free(owner, req);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status sc_scriptRequestPool_destruct(sc_scriptRequestPool* ppool)
{
  static const sc_rep_vtable vtbl = {
    sc_rep_addRef_default,
    sc_rep_relRef_default,
    (sc_rep_delete)sc_scriptRequest_delete
  };
  
  sc_status status = sc_status_success;
  sc_scriptRequestPtrhListIterator it;

  /* reset the vtables in packets to free them to heap */
  status = sc_scriptRequestPtrhList_getBegin(&ppool->pool_, &it);
  sm_status_returnIfFailed(status);

  for(; !sc_scriptRequestPtrhListIterator_isEnd(&it);
         sc_scriptRequestPtrhListIterator_next(&it))
  {
    sc_scriptRequestPtrh* ptrh = sc_scriptRequestPtrhListIterator_get(&it);
    sc_scriptRequest* req = sc_scriptRequestPtrh_get(ptrh);
    req->_rep.vtbl = &vtbl;
    req->_rep.pown = &ppool->al_;
  }

  status = sc_scriptRequestPtrhList_clear(&ppool->pool_);
  sm_status_returnIfFailed(status);
  status = sc_scriptRequestPtrhList_destruct(&ppool->pool_);
  sm_status_returnIfFailed(status);

  status = sc_blockAllocator_destruct(&ppool->al_);
  sm_status_returnIfFailed(status);

  pthread_mutex_destroy(&ppool->mux_);

  sm_status_return(status);
}

sm_ptrhImpl(sc_scriptRequest);
sm_listImpl(sc_scriptRequestPtrh);
