#include <stdlib.h>
#include <string.h>


#include <si_std/asyncbrg.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <si_std/asyncbrg.h>

/* this file describes an async bridge that fires events asychronously while
 * taking commands to execute synchronously though some interface
 * commands.  This is a fairly complex task, and is common to nearly all
 * automation bridges that we will develop.
 */


/*  sc_asyncBridgeBase
 * event thread:  reads the event queue, and invokes event handlers for
 * each event serially.
 *
 * command thread: waits on incoming serial data and synchronous command
 * requests from the synchronous interface.
 * 
 */

/* http://atmel.com/dyn/resources/prod_documents/BitCloud_ATAVRRZRAVEN_1_8_0.zip */
   


sc_status sc_asyncBridgeBase_construct(sc_asyncBridgeBase* pb,
                                   const sc_asyncBridge_vtable* vtbl,
                                   void* instance)
{
  sc_status status = sc_status_success;

  pb->instance_ = instance;
  pb->vtbl_ = vtbl;
  pb->evtThread_ = (pthread_t)-1;
  pb->svcThread_ = (pthread_t)-1;
  pb->state_ = se_bridgeStateStopped;
  
  pthread_mutex_init(&pb->evtMux_, NULL);
  
  /* note that anonymous semphore are not supported under OS X. A named
   * sem would have to be used there. The interface is unfortunately
   * different in some annoying ways, so OS X support is omitted here */
  // sem_init(&pb->evtRdy_, 0, 0);  

  sm_status_return(status);
}

static sc_status sc_asyncBridge_serviceThread(sc_asyncBridgeBase* pb)
{
  sc_status status = sc_status_success;
  fd_set rfds;
  struct timeval tv;
  int rv, nfds;

  nfds = (pb->fdDev_ > pb->fdCmdS_) ? pb->fdDev_ : pb->fdCmdS_;
  nfds++;
  
  while(1)
  {
    FD_ZERO(&rfds);
    FS_SET(pb->fdDev_,  &rfds);
    FS_SET(pb->fdCmdS_, &rfds);
    tv.tv_sec = 5;
    tv.tv_usec = 0;
    
    rv = select(nfds, &rfds, NULL, NULL, &tv);
    if(rv > 0)
    {
      if(pb->state_ == se_bridgeStateStopping)
        break;
      if(FD_ISSET(pb->fdDev_, &rfds)) {
        /* device is readable */
        status = pb->vtbl_->readDevice(pb->instance_);
        sm_status_returnIfFailed(status);
      }
      if(FD_ISSET(pb->fdCmdS_, &rfds)) {
        /* command is readable */
        status = pb->vtbl_->readRequest(pb->instance_);
        sm_status_returnIfFailed(status);
      }
    }
    if(pb->state_ == se_bridgeStateStopping)
      break;    
  }
  
  sm_status_return(status);
}
static sc_status sc_asyncBridge_eventThread(sc_asyncBridgeBase* pb)
{
  sc_status status = sc_status_success;

  while(1)
  {
    sem_wait(&pb->evtRdy_);
    if(pb->state_ == se_bridgeStateStopping)
      break;

    status = pb->vtbl_->fireEvent(pb->instance_);
    sm_status_returnIfFailed(status);
    
  }
  sm_status_return(status);
}

static void* serviceThreadEntry(void* arg)
{ 
  sc_asyncBridgeBase* pb = (sc_asyncBridgeBase*)arg;
  sc_status status = sc_status_success;

  status = sc_asyncBridge_serviceThread(pb);
  
  return NULL;
}

static void* eventThreadEntry(void* arg)
{
  sc_asyncBridgeBase* pb = (sc_asyncBridgeBase*)arg;
  sc_status status = sc_status_success;

  status = sc_asyncBridge_eventThread(pb);
  
  return NULL;
}

sc_status sc_asyncBridgeBase_start(sc_asyncBridgeBase* pb)
{
  sc_status status = sc_status_success;
  int rv, sockets[2] = { -1, -1};
  
  if(pb->state_ != se_bridgeStateStopped)
    sm_status_returnError(status, sc_status_failure);

  rv = socketpair(AF_UNIX, SOCK_STREAM, 0, sockets);
  if(rv != 0)
    sm_status_returnError(status, sc_status_failure);

  pb->fdCmdC_ = sockets[0];
  pb->fdCmdS_ = sockets[1];
  
  pb->state_ = se_bridgeStateStarting;
  
  pthread_create(&pb->evtThread_, NULL, eventThreadEntry, pb);
  pthread_create(&pb->svcThread_, NULL, serviceThreadEntry, pb);

  /* this is not very rigorous ... */
  pb->state_ = se_bridgeStateStarted;  
  
  sm_status_return(status);
}

sc_status sc_asyncBridgeBase_stop(sc_asyncBridgeBase* pb)
{
  sc_status status = sc_status_success;
  s_uint8 ch = 0x04; /* ^D */

  if(pb->state_ != se_bridgeStateStarted)
    sm_status_return(status);
  
  pb->state_ = se_bridgeStateStopping;

  /* wake 'em up */
  write(pb->fdCmdC_, &ch, 1);
  sem_post(&pb->evtRdy_);

  pthread_join(pb->svcThread_, NULL);
  pthread_join(pb->evtThread_, NULL);

  close(pb->fdCmdC_);
  close(pb->fdCmdS_);
  
  pb->state_ = se_bridgeStateStopped;
  
  sm_status_return(status);
}

sc_status sc_asyncBridgeBase_destruct(sc_asyncBridgeBase* pb)
{
  sc_status status = sc_status_success;
  
  if(pb->state_ == se_bridgeStateStarted) {
    status = sc_asyncBridgeBase_stop(pb);
    sm_status_returnIfFailed(status);
  }

  sem_destroy(&pb->evtRdy_);
  pthread_mutex_destroy(&pb->evtMux_);
  
  sm_status_return(status);
}

