#include <stdlib.h>
#include <si_std/list.h>

struct sc_listNodeAllocator
{
  s_int32 refc;
  const sc_listNodeDescriptor* ndesc;
  sc_blockAllocator allocator;
};

#define sm_ownrPtr(pn)\
    ((sc_listNodeAllocator*)((pn)->ownr))
#define sm_valPtr(pn)\
  ((void*)((pn)+1))
#define sm_nodePtr(it)\
  (it)->priv_.node
#define sm_nodeValid(pn)\
  (pn != NULL)

#define sm_nodeRelease(n)\
    sc_listNodeAllocator_deleteNode(sm_ownrPtr(n),&(n))

#define sm_itValid(it)\
  ( (it) && (sm_nodeValid(sm_nodePtr(it))) )
#define sm_itEnd(it)\
  ( sm_nodePtr(it)->next == NULL )


sc_status sc_listNodeAllocator_construct(sc_listNodeAllocator* pown,
                                      const sc_listNodeDescriptor* ndesc,
                                      s_int32 icap, s_int32 mcap)
{
  
  pown->refc  = 1;
  pown->ndesc = ndesc;

  return sc_blockAllocator_construct(&pown->allocator,
                                     ndesc->nodeSize+sizeof(sc_listNode),
                                     icap,mcap);
}
sc_status sc_listNodeAllocator_destruct(sc_listNodeAllocator* pown)
{
  return sc_blockAllocator_destruct(&pown->allocator);
}
sc_status sc_listNodeAllocator_new(sc_listNodeAllocator** alloc,
                                     const sc_listNodeDescriptor* ndesc,
                                     s_int32 icap, s_int32 mcap)
{
  sc_status status = sc_status_success;  
  sc_listNodeAllocator* pna = NULL;

  if(!alloc || !ndesc)
    sm_status_returnError(status, sc_err_nullArgument);

  /* create new node allocator */
  pna = (sc_listNodeAllocator*) malloc(sizeof(sc_listNodeAllocator));
  sm_status_returnIfFailed(status);

  status = sc_listNodeAllocator_construct(pna, ndesc, icap, mcap);
  if(sm_status_failed(status)) {
    free(pna);
    *alloc = NULL;
    sm_status_return(status);
  }

  *alloc = pna;
  
  sm_status_return(status);
}
sc_status sc_listNodeAllocator_delete(sc_listNodeAllocator** alloc)
{
  sc_status status = sc_status_success;
  sc_listNodeAllocator* ptr = *alloc;
  
  status = sc_listNodeAllocator_destruct(ptr);
  sm_status_returnIfFailed(status);

  free(ptr);  
  sm_status_returnIfFailed(status);  
  
  sm_status_return(status);
}

#define sm_lna_ref(A)\
 ++((A)->refc)

#define sm_lna_unref(A)\
  if(--((A)->refc) == 0)\
    sc_listNodeAllocator_delete(&(A))

sc_status sc_listNodeAllocator_newNode(sc_listNodeAllocator* alloc,
                                       sc_listNode** ppn, void* nv)
{
  sc_status status = sc_status_success;
  
  status = sc_blockAllocator_allocate(&alloc->allocator, (void**)ppn);
  sm_status_returnIfFailed(status);
    
  status = alloc->ndesc->constructCopy(sm_valPtr(*ppn),nv);
  if(sm_status_failed(status)) {
    sc_blockAllocator_free(&alloc->allocator, *ppn);
    *ppn = NULL;
    sm_status_return(status);
  }

  (*ppn)->ownr = alloc;
  sm_lna_ref(alloc);

  sm_status_return(status);
}
sc_status sc_listNodeAllocator_deleteNode(sc_listNodeAllocator* alloc,
                                          sc_listNode** ppn)
{
  sc_status status = sc_status_success;

  status = alloc->ndesc->destruct(sm_valPtr(*ppn));
  sm_status_returnIfFailed(status);
  
  status = sc_blockAllocator_free(&alloc->allocator, *ppn);
  sm_status_returnIfFailed(status);

  sm_lna_unref(alloc);
  
  sm_status_return(status);
}

sc_status
   sc_list_construct(sc_list* pl, const sc_listNodeDescriptor* ndesc,
                     s_int32 icap, s_int32 mcap)
{
  sc_status status = sc_status_success;
  sc_listNodeAllocator* pnAllocator = NULL;
  
  if(ndesc == NULL || pl == NULL)
    sm_status_returnError(status, sc_err_nullArgument);

  status = sc_listNodeAllocator_new(&pnAllocator, ndesc, icap, mcap);
  sm_status_returnIfFailed(status);

  pl->nodeCount = 0;
  pl->end.next  = NULL;
  pl->end.prev  = NULL;
  pl->end.ownr  = NULL;
  pl->head      = &pl->end;
  
  pl->nodeAllocator = pnAllocator;
  
  sm_status_return(status);
}
sc_status sc_list_destruct(sc_list* pl)
{
  sc_status status = sc_status_success;

  status = sc_list_clear(pl);
  sm_status_returnIfFailed(status);
  
  sm_lna_unref(pl->nodeAllocator);
  sm_status_return(status);
}
sc_status sc_listIterator_init(sc_listIterator* pit)
{
  if(!pit)
    return sc_err_nullArgument;
  sm_nodePtr(pit) = NULL;
  return sc_status_success;
}
sc_status sc_listIterator_next(sc_listIterator* pit)
{
  if(sm_itValid(pit))
    sm_nodePtr(pit) = sm_nodePtr(pit)->next;
  else
    return sc_err_invalidIterator;
  return sc_status_success;
}
sc_status sc_listIterator_prev(sc_listIterator* pit)
{
  if(sm_itValid(pit))
    sm_nodePtr(pit) = sm_nodePtr(pit)->prev;
  else
    return sc_err_invalidIterator;
  return sc_status_success;
}
sc_status sc_listIterator_getPtr(sc_listIterator* pit, void** pptr)
{
  sc_status status = sc_status_success;
  if(!pptr)
    sm_status_returnError(status, sc_err_nullArgument);
  if(sm_itValid(pit) && !sm_itEnd(pit))
  {
    sc_listNode* pn = sm_nodePtr(pit);
    *pptr = ++pn;
  }
  else 
    sm_status_returnError(status, sc_err_invalidIterator);
  sm_status_return(status);
}
sc_status sc_listIterator_assign(sc_listIterator* pThis,
                                 sc_listIterator* pThat)
{
  sc_status status = sc_status_success;
  
  if(!pThis)
    sm_status_returnError(status, sc_err_nullArgument);

  sm_nodePtr(pThis) = (pThat) ? sm_nodePtr(pThat) : NULL;
  
  return sc_status_success;
}

sc_status sc_list_getBegin(sc_list* pl, sc_listIterator* pit)
{
  sc_status status = sc_status_success;
  sc_listIterator bit;
  if(!pl || !pit)
    sm_status_returnError(status, sc_err_nullArgument);
  bit.priv_.node = pl->head;  
  return sc_listIterator_assign(pit, &bit);
}
sc_status sc_list_getEnd(sc_list* pl, sc_listIterator* pit)
{
  sc_status status = sc_status_success;
  sc_listIterator eit;
  if(!pl || !pit)
    sm_status_returnError(status, sc_err_nullArgument);
  eit.priv_.node = &pl->end;
  return sc_listIterator_assign(pit, &eit);
}

sc_status sc_list_insert(sc_list* pl, sc_listIterator* it, void* nt)
{
  sc_status status = sc_status_success;  
  sc_listNode* pit = NULL;
  sc_listNode* pnn = NULL;

  if(!pl || !it || !nt)
    sm_status_returnError(status, sc_err_nullArgument);
  
  status = sc_listNodeAllocator_newNode(pl->nodeAllocator, &pnn, nt);
  sm_status_returnIfFailed(status);

  if(sm_itEnd(it))
  { /* push at end */
    pnn->next = &pl->end;
    pnn->prev = pl->end.prev;
    pl->end.prev = pnn;
    if(pnn->prev)
      pnn->prev->next = pnn;
    else 
      pl->head = pnn;
  }
  else
  {
    sc_listNode* pin = sm_nodePtr(it);
    pnn->next = pin;
    pnn->prev = pin->prev;
    if(pin->prev)
      pin->prev->next = pnn;
    else 
      pl->head = pnn;
    pin->prev = pnn;
  }
  pl->nodeCount++;
  
  sm_status_return(status);
}

sc_status sc_list_erase(sc_list* pl, sc_listIterator* it)
{ /* invalidates it */
  if(sm_itValid(it) && !sm_itEnd(it))
  {
    sc_listNode* pn = sm_nodePtr(it);
    pn->next->prev = pn->prev;
    if(pn->prev)
      pn->prev->next = pn->next;
    else
      pl->head = pn->next;    
    sm_nodeRelease(pn);
    sm_nodePtr(it) = NULL; /* invalidate iterator */
    pl->nodeCount--;
  }  
  return sc_status_success;  
}
sc_status sc_list_remove(sc_list* pl, sc_listIterator* it)
{ /* does not invalidate it */
  if(sm_itValid(it) && !sm_itEnd(it))
  {
    sc_listNode* pn = sm_nodePtr(it);    
    pn->next->prev = pn->prev;
    if(pn->prev)
      pn->prev->next = pn->next;
    else
      pl->head = pn->next;
    pl->nodeCount--;
  }    
  return sc_status_success;
}
sc_status sc_list_clear(sc_list* pl)
{ 
  sc_listNode* pn = NULL, *pc = NULL;

  pc = pl->head;
  pl->nodeCount = 0;
  pl->end.prev = NULL;
  pl->head = &pl->end;
  pn = pc->next;
  while(pn != NULL)
  {
    sm_nodeRelease(pc);
    pc = pn;
    pn = pn->next;    
  }  
  return sc_status_success;  
}
sc_status sc_list_getIsEmpty(sc_list* pl, s_bool* pb)
{
  sc_status status = sc_status_success;
  if(!pb || !pl)
    sm_status_returnError(status, sc_err_nullArgument);
  *pb = (pl->nodeCount == 0) ? seTrue : seFalse;  
  sm_status_return(status);
}

sc_status sc_list_pushBack(sc_list* pl, void* nv)
{
  sc_status status = sc_status_success;
  sc_listNode* pnn = NULL;
  
  if(!pl || !nv)
    sm_status_returnError(status, sc_err_nullArgument);

  status = sc_listNodeAllocator_newNode(pl->nodeAllocator, &pnn, nv);
  sm_status_returnIfFailed(status);

  pnn->next = &pl->end;
  pnn->prev =  pl->end.prev;
  pl->end.prev = pnn;
  if(pnn->prev)
    pnn->prev->next = pnn;
  else
    pl->head = pnn;
  pl->nodeCount++;
  
  sm_status_return(status);  
}
sc_status sc_list_pushFront(sc_list* pl, void* nv)
{
  sc_status status = sc_status_success;
  sc_listNode* pnn = NULL;

  if(!pl || !nv)
    sm_status_returnError(status, sc_err_nullArgument);

  status = sc_listNodeAllocator_newNode(pl->nodeAllocator, &pnn, nv);
  sm_status_returnIfFailed(status);

  pnn->next = pl->head;
  pl->head = pnn;
  pnn->prev = NULL;
  pnn->next->prev = pnn;
  
  pl->nodeCount++;

  sm_status_return(status);  
}
sc_status sc_list_popBack(sc_list* pl, void* pv)
{
  sc_status status = sc_status_success;
  sc_listNode* pn = NULL;

  pn = pl->end.prev;
  if(pn)
  {
    if(pv) {
      status = sm_ownrPtr(pn)->ndesc->assign(pv,sm_valPtr(pn));
      sm_status_returnIfFailed(status);
    }
    if(pn->prev)
      pn->prev->next = pn->next;
    else
      pl->head = pn->next;      
    pl->end.prev = pn->prev;
    sm_nodeRelease(pn);
    pl->nodeCount--;
  }
  else /* TODO: better error */
    sm_status_returnError(status, sc_status_failure);
  
  sm_status_return(status);
}
sc_status sc_list_popFront(sc_list* pl, void* pv)
{
  sc_status status = sc_status_success;
  sc_listNode* pn = NULL;
  
  pn = pl->head;
  if(pn->next)
  { /* it's not the end */
    if(pv) {
      status = sm_ownrPtr(pn)->ndesc->assign(pv,sm_valPtr(pn));
      sm_status_returnIfFailed(status);
    }
    pn->next->prev = NULL;
    pl->head = pn->next;
    sm_nodeRelease(pn);
    pl->nodeCount--;
  }
  else /* TODO: better error */
    sm_status_returnError(status, sc_status_failure);
  
  sm_status_return(status);
}

sc_status sc_list_front(sc_list* pl, void** ppv)
{
  sc_status status = sc_status_success;
  
  if(pl->head->next)
    *ppv = sm_valPtr(pl->head);
  else
    sm_status_returnError(status, sc_status_failure);    
  
  sm_status_return(status);
}
sc_status sc_list_back(sc_list* pl, void** ppv)
{
  sc_status status = sc_status_success;

  if(pl->end.prev) 
    *ppv = sm_valPtr(pl->end.prev);
  else
    sm_status_returnError(status, sc_status_failure);    
  
  sm_status_return(status);
}
sc_status sc_list_getSize(sc_list* pl, s_uint32* psz)
{
  sc_status status = sc_status_success;

  *psz = pl->nodeCount;
  
  sm_status_return(status);
}



