#include <stdlib.h>

#include <si_sys/defs.h>
#include <si_sys/blkalloc.h>


/* static helper functions */
static sc_status allocParse(sc_blockAllocator* pba, void* ptr, s_int32 nblks);
static sc_status sc_blockAllocator_grow(sc_blockAllocator* pba, s_int32 minsz);


sc_status sc_blockAllocator_construct(sc_blockAllocator* pba,
                                      s_int32 blockSize,
                                      s_int32 initialCapacity,
                                      s_int32 maxCapacity)
{
  sc_status status = sc_status_success;

  /* blockSize must be able to contain descriptor */
  if(blockSize < sizeof(sc_blockDesc))
    blockSize = sizeof(sc_blockDesc);
  
  pba->blockSize   = blockSize;
  pba->maxCapacity = maxCapacity;
  pba->buffers     = NULL;
  pba->blocks      = NULL;
  pba->nblocks     = 0;
  
  if(initialCapacity) {
    status = sc_blockAllocator_grow(pba, initialCapacity);
    sm_status_returnIfFailed(status);
  }
  
  sm_status_return(status);
}

sc_status sc_blockAllocator_destruct(sc_blockAllocator* pba)
{
  sc_status status = sc_status_success;

  /* note: this very simple implementation requires that all
   * blocks have been returned (freed) by allocator destruct */
  
  sc_bufferDesc* pbuff = pba->buffers;
  while(pbuff != NULL) {
    sc_bufferDesc* pnext = pbuff->next;
    free(pbuff);
    pbuff = pnext;
  }
  
  sm_status_return(status);
}

sc_status sc_blockAllocator_allocate(sc_blockAllocator* pba, void** ppblock)
{
  sc_status status = sc_status_success;
  
  if(pba->blocks != NULL) {
    /* Simply dish out the one on top. Done! */
    *ppblock = pba->blocks;
    pba->blocks = pba->blocks->next;
  } else {
    s_int32 ngrow = pba->nblocks; /* binary growth (aka size doubling) */

    /* don't exceed maxCapacity */
    if(pba->nblocks + ngrow > pba->maxCapacity)
      ngrow = pba->maxCapacity - pba->nblocks;
    if(ngrow == 0)
      sm_status_returnError(status, sc_err_outOfMemory);
    
    status = sc_blockAllocator_grow(pba, ngrow);
    sm_status_returnIfFailed(status);
    status = sc_blockAllocator_allocate(pba, ppblock);
    sm_status_returnIfFailed(status);
  }
  sm_status_return(status);
}

sc_status sc_blockAllocator_free(sc_blockAllocator* pba, void* pblock)
{
  sc_status status = sc_status_success;

  /* push it back on top.  Re-using recently used buffers early can increase
   * cache performance in some small systems */
  
  sc_blockDesc* pblk = (sc_blockDesc*)pblock;
  pblk->next         = pba->blocks;
  pba->blocks        = pblk;
  
  sm_status_return(status);
}

static sc_status sc_blockAllocator_grow(sc_blockAllocator* pba, s_int32 newBlocks)
{
  sc_status status = sc_status_success;
  sc_bufferDesc* pbuff = NULL;

  /* add buffer descriptor to front to avoid extra malloc */
  pbuff = malloc( (pba->blockSize * newBlocks) + sizeof(sc_bufferDesc) );

  if(pbuff == NULL)
    return sc_err_outOfMemory;

  /* track the new buffer of blocks for later free */
  pbuff->next = pba->buffers;
  pba->buffers = pbuff;

  /* parse the contiguous buffer into blocks */
  status = allocParse(pba, (void*)(pbuff+1), newBlocks); 
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status allocParse(sc_blockAllocator* pba, void* ptr, s_int32 nblks)
{
  sc_status status = sc_status_success;
  s_int32 bsz = pba->blockSize, idx = nblks;
  sc_blockDesc* pblk = (sc_blockDesc*)ptr;

  /* walk the contiguous buffer, chaining each bsz size block to the next */  
  while(--idx) {
    sc_blockDesc* next = (sc_blockDesc*)(((s_uint8*)pblk)+bsz);
    pblk->next = next;
    pblk = next;
  }

  /* stick them at the front of blocks list */
  pblk->next    = pba->blocks;
  pba->blocks   = (sc_blockDesc*)ptr;
  pba->nblocks += nblks;

  sm_status_return(status);
}
