#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#include <si_sys/defs.h>
#include <si_sha/zigbee.h>
#include <si_sha/rzraven.h>
#include <si_sha/zigbdev.h>
#include <si_std/string.h>

#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>


sc_status sc_zigbeeBridgeDevice_construct(sc_zigbeeBridgeDevice* pbd,
                                          sc_automationDeviceMgr* pdm,
                                          const TCHAR* name,
                                          const TCHAR* bridgePath,
                                          s_uint8 channel,
                                          s_uint16 panId
                                         )
{
  static const sc_automationBridgeDevice_vtable vtbl = {
    (sc_automationBridgeDevice_processEventF)sc_zigbeeBridgeDevice_processEvent,
    (sc_automationBridgeDevice_processRequestF)sc_zigbeeBridgeDevice_processRequest,
    (sc_automationBridgeDevice_newPacketF)sc_zigbeeBridgeDevice_newPacket,
    (sc_automationBridgeDevice_discoverF)sc_zigbeeBridgeDevice_discover
  };
  sc_status status = sc_status_success;
  sc_automationPacketPtrh rqptrh, rsptrh;
  sc_automationPacket *req, *res;
  sc_string namePath;
  s_uint64 vaddr;
  s_uint32 vdfwr, vtype;
  s_uint16 result;
  struct sockaddr_un addr;
  int rv;
  char daddr[32];
  char dtype[16];

  /* construct pool and allocate a couple of packets */
  status = sc_automationPacketPool_construct(&pbd->pool_, 128, 16, INT_MAX);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacketPtrh_construct(&rqptrh, NULL);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_construct(&rsptrh, NULL);
  sm_status_returnIfFailed(status);

  status = sc_automationPacketPool_allocate(&pbd->pool_, &rqptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPool_allocate(&pbd->pool_, &rsptrh);
  sm_status_returnIfFailed(status);
  req = sc_automationPacketPtrh_get(&rqptrh);
  res = sc_automationPacketPtrh_get(&rsptrh);

  /* build up bridge name path */
  sc_string_construct(&namePath);
  sc_string_append(&namePath, bridgePath);
  sc_string_appendChars(&namePath, "/", 1);
  sc_string_append(&namePath, name);

  /* connect to bridge at bridgePath/name */
  pbd->base_.fd_ = socket(AF_UNIX, SOCK_STREAM, 0);

  memset(&addr, 0, sizeof(addr));
  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, namePath.str_, sizeof(addr.sun_path)-1);

  rv = connect(pbd->base_.fd_, (struct sockaddr*)&addr, sizeof(addr));
  if(rv != 0) {
    sc_automationPacketPtrh_destruct(&rsptrh);
    sc_automationPacketPtrh_destruct(&rqptrh);
    sc_string_destruct(&namePath);
    sc_automationPacketPool_destruct(&pbd->pool_);
    sm_status_returnError(status, sc_status_failure);
  }
  /* rv == 0 => success */

  status = sc_automationPacket_pushUInt16(req, ZC_GETVERSION);
  sm_status_returnIfFailed(status);

  /* send a message to get the address */
  status = sc_zigbeeBridgeDevice_processRequest(pbd, req, res);
  sm_status_returnIfFailed(status);

  status = sc_automationPacket_popUInt16(res, &result);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_popUInt32(res, &vdfwr);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_popUInt32(res, &vtype);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_popUInt64(res, &vaddr);
  sm_status_returnIfFailed(status);  

  status = sc_automationPacket_clear(req);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_clear(res);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt16(req, panId);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_pushUInt8(req, channel);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt16(req, ZC_FORMNETWORK);
  sm_status_returnIfFailed(status);
  status = sc_zigbeeBridgeDevice_processRequest(pbd, req, res);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_popUInt16(res, &result);
  sm_status_returnIfFailed(status);  
  
  zb_aprint(daddr, vaddr);
  zb_tprint(dtype, vtype); 

  status = sc_automationBridgeDevice_construct(&pbd->base_, pdm, name,
                                               dtype, daddr, "ZigBee",
                                               &vtbl, pbd);
  sm_status_returnIfFailed(status);  

  status = sc_automationPacketPtrh_destruct(&rqptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_destruct(&rsptrh);
  sm_status_returnIfFailed(status);
  status = sc_string_destruct(&namePath);
  sm_status_returnIfFailed(status);

  printf("ZigBee bridge device [IEEE address %s] connected.\n", daddr);

  sm_status_return(status);
}

sc_status sc_zigbeeBridgeDevice_newPacket(sc_zigbeeBridgeDevice* pb,
                                          sc_automationPacketPtrh* pptrh)
{
  return sc_automationPacketPool_allocate(&pb->pool_, pptrh);
}

sc_status sc_zigbeeBridgeDevice_processEvent(sc_zigbeeBridgeDevice* brg,
                                             sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;
  int fd;
  s_uint8 tmp, code, len, sig, sqn, cid;
  s_uint32 ppos;
  s_uint64 from;
  s_uint16 result, shortAddr, dst, src;
  char buff[32];

  fd = brg->base_.fd_;

  status = sc_automationPacket_deserializeFromFd(pkt, fd);
  sm_status_returnIfFailed(status);

  sc_automationPacket_tell(pkt, &ppos);
  sc_automationPacket_popUInt16(pkt, &result);  /* 0x0001 on success */
  sc_automationPacket_popUInt8 (pkt, &code);    /* event code */

  switch(code) {
    case EVENT_NWK_JOIN: 
      sc_automationPacket_popUInt8(pkt, &tmp); /* length */
      sc_automationPacket_popUInt16(pkt, &shortAddr); /* 16 bit net address */
      sc_automationPacket_popUInt64(pkt, &from);      /* 64 bit IEEE address */
      // sc_automationPacket_setPos(pkt, ppos);
      zb_aprint(buff, from);
      
      /* sc_automationDeviceMgr_relayMessage(brg->base_.pdm_,
                                          fromAddress, pkt); */
      
      printf("Event: ZigBee Join from %s, using %02X%02X\n",
              buff, (int)(shortAddr >> 8), (int)(shortAddr & 0xFF));
      break;
    case EVENT_NWK_LEAVE: 
      sc_automationPacket_popUInt8(pkt, &tmp); /* length */
      sc_automationPacket_popUInt64 (pkt, &from); /* flags */
      zb_aprint(buff, from);
      printf("Event: ZigBee Leave from %s\n", buff);
      break;
    case EVENT_NWK_DATA:
      sc_automationPacket_popUInt8 (pkt, &len); 
      sc_automationPacket_popUInt16(pkt, &dst); 
      sc_automationPacket_popUInt16(pkt, &src);
      sc_automationPacket_popUInt8 (pkt, &sig);
      sc_automationPacket_popUInt8 (pkt, &len); 
      printf("Event: Data Recvd from %04X, signal %02X, length %d\n",
             (int)src, (int)sig, (int)len);
      sc_automationPacket_popUInt8 (pkt, &sqn);
      sc_automationPacket_popUInt8 (pkt, &cid);
      if( (sqn == 0x01) && (cid == 0x06)) {
        puts(">> Text Message: ");
        sc_automationPacket_popUInt8 (pkt, &len);
        while(len--) {
          sc_automationPacket_popUInt8 (pkt, &tmp);
          putchar(tmp);
        }
        putchar('\n');
      } else {
        while(len--) {
          sc_automationPacket_popUInt8(pkt, &tmp);
          printf("0x%02X ('%c')\n", (int)tmp, tmp);
        }
      }
      break;
    default:
      printf("Event: %X\n", code);
      break;      
  }
  sm_status_return(status);
}

sc_status sc_zigbeeBridgeDevice_destruct(sc_zigbeeBridgeDevice* pbd)
{
  sc_status status = sc_status_success;

  printf("ZigBee bridge device %s, address %s, disconnecting.\n",
         pbd->base_.base_.name_, pbd->base_.base_.address_);

  status = sc_automationBridgeDevice_destruct(&pbd->base_);
  sm_status_returnIfFailed(status);

  status = sc_automationPacketPool_destruct(&pbd->pool_);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status sc_zigbeeBridgeDevice_processRequest(sc_zigbeeBridgeDevice* brg,
                                                sc_automationRequest* req,
                                                sc_automationResponse* res)
{
  sc_status status = sc_status_success;
  fd_set rfds;
  int nfds, rv, fd;
  struct timeval tv;

  tv.tv_sec = 10;
  tv.tv_usec = 0;

  fd = brg->base_.fd_;
  nfds = fd + 1;

  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);

  /* serialize request out */
  status = sc_automationPacket_serializeToFd(req, fd);
  sm_status_returnIfFailed(status);

  /* wait for response */
  rv = select(nfds, &rfds, NULL, NULL, &tv);
  if(rv != 1) {
    sm_status_returnError(status, sc_status_failure);
  }

  /* deserialize response from fd */
  status = sc_automationPacket_deserializeFromFd(res, fd);
  sm_status_returnIfFailed(status);

  /* TODO: check for request/response match, queue as event if not */
  /*  ^^ Such a re-ordering is now done by the bridge */

  sm_status_return(status);
}

sc_status sc_zigbeeBridgeDevice_discover(sc_zigbeeBridgeDevice* pbd)
{
  sc_status status = sc_status_success;
  sm_status_return(status);
}
