#include <stdlib.h>
#include <stdio.h>


#include <si_sha/zigbrg.h>
#include <si_sha/rzraven.h>


static sc_status sc_zigbeeBridge_getAddress(sc_zigbeeBridge* pzb);
static sc_status sc_zigbeeBridge_getVersion(sc_zigbeeBridge* pzb);
static sc_status sc_zigbeeBridge_signOn(sc_zigbeeBridge* pzb);
static sc_status readResponse(sc_zigbeeBridge* pzb, s_uint8* buff, s_int32* ln);
static sc_status sc_zigbeeBridge_handleAsync(sc_zigbeeBridge* pb,
                                             const s_uint8* res,
                                             s_int32 len);


sc_status sc_zigbeeBridge_serviceDevice(sc_zigbeeBridge* pb,
                                        sc_automationPacket* pkt);

sc_status sc_zigbeeBridge_processRequest(sc_zigbeeBridge* pb,
                                         sc_automationBridgeConnection* pc,
                                         sc_automationRequest*  req,
                                         sc_automationResponse* res);

sc_status sc_zigbeeBridge_construct(sc_zigbeeBridge* pb,
                                    const TCHAR* name,
                                    const TCHAR* devName,
                                    const TCHAR* basePath)
{
  static const sc_automationBridge_vtable vtbl = {
    (sc_automationBridge_serviceDeviceF)sc_zigbeeBridge_serviceDevice,
    (sc_automationBridge_processRequestF)sc_zigbeeBridge_processRequest
  };
  sc_status status = sc_status_success;
  char addr[32];
  
  status = sc_serialPort_construct(&pb->port_, devName,
                                   se_baudRate_115200);
  sm_status_returnIfFailed(status);

  /* turn all leds off */
  status = sc_zigbeeBridge_setLeds(pb, 0);
  sm_status_returnIfFailed(status);
  
  status = sc_zigbeeBridge_signOn(pb);
  sm_status_returnIfFailed(status);
  
  pb->address_ = 0;
  pb->version_ = 0;
  pb->type_ = ZB_RAVENBRGT;
  
  /* grab address & version from the raven bridge */
  status = sc_zigbeeBridge_getAddress(pb);
  sm_status_returnIfFailed(status);
  status = sc_zigbeeBridge_getVersion(pb);
  sm_status_returnIfFailed(status);

  status = sc_automationBridgeBase_construct(&pb->base_, name, basePath, 128, 16, &vtbl, pb);
  sm_status_returnIfFailed(status);

  zb_aprint(addr, pb->address_);
  pb->base_.fdDev_ = pb->port_.fd_;
  
  printf("ZigBee bridge %s [%s] starting on %s.\n", name, addr, devName );
  
  sm_status_return(status);
}

sc_status sc_zigbeeBridge_destruct(sc_zigbeeBridge* pb)
{
  sc_status status = sc_status_success;
  
  printf("ZigBee bridge %s stopping.\n", pb->base_.name_);

  status = sc_automationBridgeBase_destruct(&pb->base_);
  sm_status_returnIfFailed(status);

  status = sc_serialPort_destruct(&pb->port_);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}


sc_status sc_zigbeeBridge_serviceDevice(sc_zigbeeBridge* pzb,
                                        sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;
  s_uint8 buff[128];
  s_int32 len = 0;

  /* read out the response */
  status = readResponse(pzb, buff, &len);
  sm_status_returnIfFailed(status);

  /* return it in packet */
  status = sc_zigbeeBridge_handleAsync(pzb, buff, len);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

static sc_status readAddress(sc_zigbeeBridge* pzb, s_uint8* buff)
{
  sc_status status = sc_status_success;
  s_uint8 ch;

  status = sc_serialPort_readAll(&pzb->port_, &ch, 1);
  sm_status_returnIfFailed(status);

  if(ch == RESP_GET_PARAMETER) {
    /* get length as required */
    status = sc_serialPort_readAll(&pzb->port_, &ch, 1);
    sm_status_returnIfFailed(status);
    
    status = sc_serialPort_readAll(&pzb->port_, buff, 8);
    sm_status_returnIfFailed(status);
  } else {
    sm_status_returnError(status, sc_status_failure);
  }
  
  sm_status_return(status);
}
static sc_status readVersion(sc_zigbeeBridge* pzb, s_uint8* buff)
{
  sc_status status = sc_status_success;
  s_uint8 ch;

  status = sc_serialPort_readAll(&pzb->port_, &ch, 1);
  sm_status_returnIfFailed(status);

  if(ch == RESP_GET_PARAMETER) {
    /* get length as required */
    status = sc_serialPort_readAll(&pzb->port_, &ch, 1);
    sm_status_returnIfFailed(status);

    status = sc_serialPort_readAll(&pzb->port_, buff, 2);
    sm_status_returnIfFailed(status);
  } else {
    sm_status_returnError(status, sc_status_failure);
  }

  sm_status_return(status);
}

static sc_status sc_zigbeeBridge_getAddress(sc_zigbeeBridge* pzb)
{
  sc_status status = sc_status_success;  
  static const s_uint8 cmd[3] = { 2, CMD_GET_PARAMETER, PARAM_IEEE_ADDRESS };
  s_int32 bw=0;
  s_uint64 addr;

  status = sc_serialPort_write(&pzb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  status = readAddress(pzb, (s_uint8*)&addr);
  sm_status_returnIfFailed(status);

  pzb->address_ = addr;
  
  sm_status_return(status);
}

static sc_status sc_zigbeeBridge_getVersion(sc_zigbeeBridge* pzb)
{
  sc_status status = sc_status_success;  
  static const s_uint8 cmd[3] = { 2, CMD_GET_PARAMETER, PARAM_SW_REV };
  s_int32 bw=0;
  s_uint8 buff[2];

  status = sc_serialPort_write(&pzb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  status = readVersion(pzb, buff);
  sm_status_returnIfFailed(status);
 
  pzb->version_ = (buff[0] << 8) + buff[1];

  sm_status_return(status);
}

static sc_status sc_zigbeeBridge_signOn(sc_zigbeeBridge* pzb)
{
  sc_status status = sc_status_success;  
  static const s_uint8 cmd[2] = { 1, CMD_SIGN_ON };
  s_int32 bw=0;
  s_uint8 code, len, ch;
  int idx=0;

  status = sc_serialPort_write(&pzb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  status = sc_serialPort_readAll(&pzb->port_, &code, 1);
  sm_status_returnIfFailed(status);

  if(code != RESP_SIGN_ON)
    sm_status_returnError(status, sc_status_failure);

  status = sc_serialPort_readAll(&pzb->port_, &len, 1);
  sm_status_returnIfFailed(status);

  for(idx = 0; idx < len; ++idx)
  {
    status = sc_serialPort_readAll(&pzb->port_, &ch, 1);
    sm_status_returnIfFailed(status);
    putchar(ch);
  }
  putchar('\n');
  
  sm_status_return(status);
}

sc_status sc_zigbeeBridge_getVersionInfo(sc_zigbeeBridge* pzb,
                                         sc_zigbeeBridgeVersionInfo* bvi)
{
  sc_status status = sc_status_success;  

  bvi->address_ = pzb->address_;
  bvi->type_ = pzb->type_;
  
  sm_status_return(status);
}

sc_status sc_zigbeeBridge_setLeds(sc_zigbeeBridge* pzb, s_uint8 mask)
{
  sc_status status = sc_status_success;  
  s_uint8 cmd[3] = { 2, CMD_SET_LEDS };
  s_int32 bw=0;
  s_uint8 code, len, ch;
  int idx=0;

  cmd[2] = mask;
  
  status = sc_serialPort_write(&pzb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  status = sc_serialPort_readAll(&pzb->port_, &code, 1);
  sm_status_returnIfFailed(status);

  if(code != RESP_SUCCESS) {
    printf("raven response failed: %d\n", (int)code);
  }

  sm_status_return(status);
}

sc_status sc_zigbeeBridge_setMode(sc_zigbeeBridge* pzb, s_uint8 mode)
{ 
  sc_status status = sc_status_success;  
  s_uint8 cmd[3] = { 2, CMD_SET_MODE };
  s_int32 bw = 0;
  s_uint8 code;

  cmd[2] = mode;

  status = sc_serialPort_write(&pzb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  status = sc_serialPort_readAll(&pzb->port_, &code, 1);
  sm_status_returnIfFailed(status);

  if(code != RESP_SUCCESS) {
    printf("raven response failed: %d\n", (int)code);
  }

  sm_status_return(status);
}

#define IS_ERROR(b)        ((b & 0xF0) == 0x90)
#define IS_SUCCESS(b)      ((b & 0xF0) == 0x80)
#define IS_RESPONSE(b)     (IS_ERROR(b) || IS_SUCCESS(b))
#define IS_INDICATION(b)   ((b & 0xF0) == 0x50)
#define IS_CONFIRMATION(b) ((b & 0xF0) == 0x60)

#define IS_SIMPLE_RESPONSE(b) (b == RESP_SUCCESS || IS_ERROR(b))

static sc_status readResponse(sc_zigbeeBridge* pzb,
                              s_uint8* buff,
                              s_int32* plen)
{
  sc_status status = sc_status_success;
  int len=0, br=0;
  s_uint8 code;

  /* read code */
  status = sc_serialPort_readAll(&pzb->port_, &code, 1);
  sm_status_returnIfFailed(status);

  buff[0] = code;
  
  if(IS_SIMPLE_RESPONSE(code)) {
    *plen = 1;
    sm_status_return(status);
  } else if(IS_CONFIRMATION(code)) {
    /* read the simple response code and return */
    status = sc_serialPort_readAll(&pzb->port_, buff+1, 1);
    sm_status_returnIfFailed(status);
    *plen = 2;    
    sm_status_return(status);
  }
  
  /* otherwise, it has a payload length byte */
  status = sc_serialPort_readAll(&pzb->port_, buff+1, 1);
  sm_status_returnIfFailed(status);
  
  len = buff[1];
  if(IS_INDICATION(code))
    len -= 2;
  if(len < 0)
    sm_status_returnError(status, sc_status_failure);
  if(len > 0) {
    status = sc_serialPort_readAll(&pzb->port_, buff+2, len);
    sm_status_returnIfFailed(status);
  }

  *plen = len + 2;
  
  sm_status_return(status);
}

sc_status sc_zigbeeBridge_formNetwork(sc_zigbeeBridge* pzb, s_uint8 channel, s_uint16 panId)
{ 
  sc_status status = sc_status_success;  
  s_uint8 cmd[5] = { 4, CMD_NLME_FORMATION_REQUEST };
  s_int32 bw = 0;
  s_uint8 code;

  cmd[2] = channel;
  cmd[3] = (s_uint8)(panId & 0xFF);
  cmd[4] = (s_uint8)((panId >> 8) & 0xFF);

  status = sc_serialPort_write(&pzb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  status = sc_serialPort_readAll(&pzb->port_, &code, 1);
  sm_status_returnIfFailed(status);

  if(code != RESP_SUCCESS) {
    printf("raven response failed: %d\n", (int)code);
  } else {
    
  }

  sm_status_return(status);
}

sc_status sc_zigbeeBridge_permitJoining(sc_zigbeeBridge* pzb, s_uint8 permitted)
{ 
  sc_status status = sc_status_success;  
  s_uint8 cmd[3] = { 2, CMD_NLME_PERMIT_JOIN_REQUEST };
  s_int32 bw = 0;
  s_uint8 code;

  cmd[2] = permitted;

  status = sc_serialPort_write(&pzb->port_, cmd, sizeof(cmd), &bw);
  sm_status_returnIfFailed(status);

  status = sc_serialPort_readAll(&pzb->port_, &code, 1);
  sm_status_returnIfFailed(status);

  if(code != RESP_SUCCESS) {
    printf("raven response failed: %d\n", (int)code);
  }

  sm_status_return(status);
}

static sc_status sc_zigbeeBridge_handleAsync(sc_zigbeeBridge* pb, const s_uint8* res,
                                             s_int32 len)
{
  sc_status status = sc_status_success;
  s_uint16 shortAddr, dstAddr, srcAddr;
  s_int32 dataLen;
  s_uint64 longAddr;
  sc_automationPacketPtrh pptrh;
  sc_automationPacket* pkt;

  status = sc_automationPacketPtrh_construct(&pptrh, NULL);
  sm_status_returnIfFailed(status);

  status = sc_automationBridgeBase_newPacket(&pb->base_, &pptrh);
  sm_status_returnIfFailed(status);
  pkt = sc_automationPacketPtrh_get(&pptrh);

  if(IS_CONFIRMATION(res[0])) {
    status = sc_automationPacket_pushUInt8(pkt, res[1]);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_pushUInt8(pkt, res[0]);
    sm_status_returnIfFailed(status);    
  } else if (IS_INDICATION(res[0])) {
    switch(res[0]) {
      case EVENT_NWK_DATA:
        dstAddr = ((res[3] << 8) + res[2]);
        srcAddr = ((res[5] << 8) + res[4]);
        dataLen = res[7];
        status = sc_automationPacket_pushBinary(pkt, &res[8], dataLen);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt8 (pkt, res[7]); /* length */
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt8 (pkt, res[6]); /* signal */
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt16(pkt, srcAddr);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt16(pkt, dstAddr);
        sm_status_returnIfFailed(status);        
        status = sc_automationPacket_pushUInt8 (pkt, res[1]);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt8 (pkt, res[0]);
        sm_status_returnIfFailed(status);        
        break;
      case EVENT_NWK_JOIN:
        /* assumes little endien */
        shortAddr = ((res[3] << 8) + res[2]);
        longAddr = ( (((s_uint64)res[11]) << 56) + (((s_uint64)res[10]) << 48) +
                     (((s_uint64)res[ 9]) << 40) + (((s_uint64)res[ 8]) << 32) +
                     (((s_uint64)res[ 7]) << 24) + (((s_uint64)res[ 6]) << 16) +
                     (((s_uint64)res[ 5]) <<  8) + (((s_uint64)res[ 4])) );
        status = sc_automationPacket_pushUInt64(pkt, longAddr);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt16(pkt, shortAddr);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt8(pkt, res[1]);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt8(pkt, res[0]);
        sm_status_returnIfFailed(status);        
        break;
      case EVENT_NWK_LEAVE:
        longAddr = ( (((s_uint64)res[9]) << 56) + (((s_uint64)res[8]) << 48) +
                     (((s_uint64)res[7]) << 40) + (((s_uint64)res[6]) << 32) +
                     (((s_uint64)res[5]) << 24) + (((s_uint64)res[4]) << 16) +
                     (((s_uint64)res[3]) <<  8) + (((s_uint64)res[2])) );
        status = sc_automationPacket_pushUInt64(pkt, longAddr);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt8(pkt, res[1]);
        sm_status_returnIfFailed(status);
        status = sc_automationPacket_pushUInt8(pkt, res[0]);
        sm_status_returnIfFailed(status);
        break;
    }
  } else {
    /* shouldn't get here ... */
  }
  
  status = sc_automationPacket_pushUInt16(pkt, 0x0001); /* async + success */
  sm_status_returnIfFailed(status);
  status = sc_automationBridgeBase_queueAsync(&pb->base_, &pptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_destruct(&pptrh);
  sm_status_returnIfFailed(status);  

  sm_status_return(status);
}


sc_status sc_zigbeeBridge_processRequest(sc_zigbeeBridge* pzb,
                                         sc_automationBridgeConnection* pc,
                                         sc_automationRequest*  req,
                                         sc_automationResponse* res)
{
  sc_status status = sc_status_success;
  s_uint16 code, panId;
  s_uint8 channel;

  status =  sc_automationPacket_popUInt16(req, &code);
  sm_status_returnIfFailed(status);

  switch(code) {
    case ZC_GETVERSION:
      status =  sc_automationPacket_pushUInt64(res, pzb->address_);
      sm_status_returnIfFailed(status);
      status =  sc_automationPacket_pushUInt32(res, pzb->type_);
      sm_status_returnIfFailed(status);
      status =  sc_automationPacket_pushUInt32(res, pzb->version_);
      sm_status_returnIfFailed(status);
      break;
    case ZC_FORMNETWORK:
      status =  sc_automationPacket_popUInt8(req, &channel);
      sm_status_returnIfFailed(status);
      status =  sc_automationPacket_popUInt16(req, &panId);
      sm_status_returnIfFailed(status);
      status = sc_zigbeeBridge_setMode(pzb, CMD_MODE_NWK);
      sm_status_returnIfFailed(status); 
      status = sc_zigbeeBridge_formNetwork(pzb, channel, panId);
      sm_status_returnIfFailed(status);
      status = sc_zigbeeBridge_permitJoining(pzb, 1);
      sm_status_returnIfFailed(status);      
      break;
    case ZC_STDMESSAGE:
      /* this is semi-complex */
      break;
    case ZC_DISCOVER:
      /* this is rather complicated */
      break;
    default:
      break;
  }

  status =  sc_automationPacket_pushUInt16(res, 0);
  sm_status_returnIfFailed(status);  

  sm_status_return(status);
}
